/* Front-end tree definitions for GNU compiler.
   Copyright (C) 1989 Free Software Foundation, Inc.

This file is part of GNU CC.

GNU CC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU CC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU CC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

#ifdef GPC

#include "gpc-lang.h"

#ifdef __STDC__
# define ASSERT(x, str) \
  if (! x) \
   printf("Assert failure at %s line %d : %s\n",__FILE__,__LINE__,str),abort();
#else
/* Thanks Doug! */
# define ASSERT(x, str) \
  if (! x) printf("Assert failure : %s\n", str),abort();
#endif

#define GPC_MAIN     "gpc_main"
#define INPUT_FILE   "Input"
#define OUTPUT_FILE  "Output"

#define RID_VARPARM	RID_UNUSED1

/* Define the type of Pascal we are compiling */
/* Used for reserved word recognition and possibly others... */
#define PASCAL_GNU	0	/* :-) */
#define PASCAL_EXTEND	1	/* Allow Extended Pascal */
#define PASCAL_ISO	2	/* Allow ISO Pascal */
#define PASCAL_ANSI	3	/* Allow ANSI Pascal */

/*
 * Some Pascal set constructors do not allow us to derive the
 * set size from anywhere. In such cases, the maximum allowed
 * members in the set is defined here. (Otherwise, there is no limit)
 *
 * By default the size of these sets is DEFAULT_SET_SIZE bits
 *
 * Users may change this with -fsetlimit:XX switch at compile time.
 *
 */

/* This is the size in bits */
#define DEFAULT_SET_SIZE	(8 * BITS_PER_WORD)

/* Is this really used?
 * For Pascal variant records:
 *
 *   TYPE_TAG_FIELD is a TREE_LIST node whose
 *   TREE_VALUE is the FIELD_DECL node (in records field list)
 *    that is defined as a variant tag
 *   TREE_PURPOSE is the type of the tag field. If TREE_VALUE
 *    is NULL_TREE, the tag field is specified only as a type;
 *   if TYPE_TAG_FIELD is NULL_TREE the record has no variant part.
 *
 */
#define TYPE_TAG_FIELD TYPE_MIN_VALUE

#define ORDINAL_TYPE(code) \
 ((code) == INTEGER_TYPE || (code) == CHAR_TYPE || \
  (code) == BOOLEAN_TYPE || (code) == ENUMERAL_TYPE)

#define ORDINAL_OR_REAL_TYPE(c) (ORDINAL_TYPE(c) || (c) == REAL_TYPE)

#define ORDINAL_REAL_OR_COMPLEX_TYPE(c) (ORDINAL_OR_REAL_TYPE(c) || (c) == COMPLEX_TYPE)

#define INT_REAL(c) ((c) == INTEGER_TYPE || (c) == REAL_TYPE)

#define INT_REAL_BOOL(c) ((c) == INTEGER_TYPE || (c) == REAL_TYPE || (c) == BOOLEAN_TYPE)

#define IS_ENUMERAL(c) ((c) == BOOLEAN_TYPE || (c) == ENUMERAL_TYPE)

/* This is incorrect. It should allow only BOOLEAN_TYPE */
/* Will change later */
#define IS_BOOLEAN(c) ((c) == BOOLEAN_TYPE || (c) == INTEGER_TYPE)

#define LOGICAL_RESULT_TYPE boolean_type_node

/* If c is SET_TYPE or a set CONSTRUCTOR */
#define SET_OR_CONST(c) ((c) == SET_TYPE || (c) == CONSTRUCTOR)

/* Pascal type nodes; maybe should be in tree.h */
extern tree boolean_type_node;
extern tree text_type_node;

/* Extended Pascal node for COMPLEX type */
extern tree complex_type_node;

/* this is a TEXT file type node */
#define TYPE_FILE_TEXT		TYPE_MIN_VALUE

/* This object is EXTERNAL in Pascal sense; it has been
   mentioned in the program heading. In Extended Pascal it does
   not have to be a file. 

   Use one bit of the language dependend common bits in the
   tree structure.
*/
#define PASCAL_EXTERNAL_OBJECT(object) TREE_LANG_FLAG_6 (object)

/*
 * This flag is set if the type is `packed'
 */
#define PASCAL_TYPE_PACKED(type) TYPE_LANG_FLAG_6 (type)

/*
 * Flag conformant array schema types.
 */
#define PASCAL_TYPE_CONFORMANT_INDEX(type) TYPE_LANG_FLAG_5 (type)

/*
 * When a VAR_DECL node is directly assigned to we mark it.
 * (Currently only used to find out if a value is assigned to
 *  a function via it's name or a user defined return value name)
 */
#define PASCAL_VALUE_ASSIGNED(decl) DECL_LANG_FLAG_6 (decl)

/*
 * Set this if NODE refers to a known identifier in Pascal.
 * It's used in gpc-lex.c yylex() to return correct tokens to the parser
 * when a Pascal predefined word is seen. Also it allows redefinition of
 * the known identifiers in pushdecl().
 *
 */
#define PASCAL_REDEFINABLE_DECL(decl) DECL_LANG_FLAG_7 (decl)

extern tree type_id;
extern tree const_id;
extern tree inline_id;
extern tree varparm_id;
extern tree volatile_id;

extern int current_function_assigned_value;

extern tree ptr_type_node;
extern tree value_identifier;

extern tree boolean_false_node;
extern tree boolean_true_node;
extern tree complex_zero_node;
extern tree real_zero_node;
extern tree integer_maxint_node;
extern tree null_pointer_node;
extern tree output_file_node;
extern tree input_file_node;

extern tree empty_set_node;

/* Extended Pascal nodes */
extern tree real_max_node;
extern tree real_min_node;
extern tree real_eps_node;
extern tree char_max_node;

/* not used */
extern tree char_array_index_type_node;

extern tree external_file_name_list;
void associate_external_files ();

tree make_signed_range ();

void un_initialize_block ();
void handle_formal_param_list ();

tree truthvalue_conversion ();

tree combine_strings ();
tree p_grokfields ();
tree build_rts_call ();
tree var_param_convert ();
tree build_range_type ();
tree build_pascal_array_type ();
tree build_pascal_array_ref ();
tree invert_truthvalue ();
void declare_rts_types ();
tree build_buffer_ref ();

tree error_level ();
tree object_size ();

tree construct_set_member();
tree construct_set();
tree build_set_constructor();

tree grok_packed();
void grok_directive ();

tree build_char();
tree build_pointer();
tree build_array();
tree build_file_type();
tree build_set_type();
tree build_record_variant_part ();

/* Nonzero meand always use short circuit boolean AND and OR */
extern int flag_short_circuit;

/* Nonzero means allow gpc to use C style character escapes */
extern int flag_c_escapes;

/* Nonzero means allow gpc to read C style Octal and Hex numbers */
extern int flag_c_numbers;

/* Level of keywords we recognize */
extern int flag_what_pascal;

/* Nonzero means allow nested comments */
/* (* { } *) and { (* *) } are ok */
/* this violates standard */
extern int flag_nested_comments;

extern int gpc_test1;
extern int gpc_test2;
extern int gpc_test3;
extern int gpc_test4;

void init_std_files ();

/* convert the TYPE to a range usable as ardinal type
 * if approppriate
 */
tree convert_type_to_range ();

tree maybe_call_function ();

int suspend_routine_calls ();

void resume_function_calls ();

enum tree_code expression_type_code ();

void resolve_forward_pointer ();

/* a chain of TREE_LIST nodes that contain the with_elements
 * of the currently active WITH statements
 */
extern tree with_element_chain;

/* Locates from structure/union DATUM the field NAME, and consturucts
 * an expression to refer to that field
 */
tree find_field ();

/*
 * check the declaration associated with the identifier
 */
tree check_identifier ();

tree find_real_epsilon ();

tree build_complex_type ();
tree build_complex ();

tree convert_array_to_pointer ();

/*
 * handles directives, calls start_function and returns the function
 * return type or 0
 */
tree start_pascal_function ();

/* convert set oprations to their bitwise form; for others just
 * call build_binary_op_nodefault
 */
tree build_pascal_op ();

/*
 * labels declared at the current level
 */
extern tree declared_labels;

/*
 * Build a FUNCTION_DECL node to define the DECL_CONTEXT of the
 * main program before start_function() is called to make non-local
 * gotos to main program work.
 */
extern tree declare_main_program_context ();

/*
 * Top level context of the Pascal program
 */
extern tree main_program_context;

/* Specifies the maximum number of members in any Pascal set
 * -fsetlimit:XX sets the limit to XX
 */
extern int requested_set_size;

extern tree integer_set_type_node;

/* handle set operations in expr.c */
extern int rts_set ();

/* Get a unique identifier (acting as a temporary internal variable)
 * LEAD parameter is a (char *) that is meant to declare context
 * for the compiler debuggers
 */
extern tree get_unique_identifier ();

/*
 * Returns the base type of a SET_TYPE node TYPE
 */
extern tree set_base_type ();

extern tree get_identifier_with_blank ();

#endif
