/* Gnu Pascal compiler lexical analyzer
   Copyright (C) 1989, Free Software Foundation, Inc.

This file is part of GNU GCC.

GNU GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * Author: Juki <jtv@hut.fi>
 */

#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#include "config.h"
#include "tree.h"
#include "input.h"
#include "c-tree.h"
#include "gpc-defs.h"
#include "gpc-parse.h"

#ifndef errno
extern int errno;
#endif

/* Parser variables */
extern int yydebug;
YYSTYPE	yylval;			/*  the semantic value of the */
				/*  lookahead symbol          */

int recognize_known_ids = 1;

extern double atof ();

tree lastiddecl;

/* List of types and structure classes of the current declaration */
tree current_declspecs;

char *input_filename;		/* file being read */
char *main_input_filename;	/* top-level source file */

char *token_buffer;		/* Pointer to token buffer.
				   Actual allocated length is maxtoken + 2.  */

int lineno;			/* current line number in file being read */

FILE *finput;			/* input file.
				   Normally a pipe from the preprocessor.  */

/* lexical analyzer */

/* The reserved word recognition is done with
   Doug Schmidt's gperf program in O(1) time.

   The structure resword is defined in this file.
 */
#include "gpc-gperf.c"

static int maxtoken;		/* Current nominal length of token buffer */

/* Nonzero if end-of-file has been seen on input.  */
static int end_of_file;

/* ISO Pascal does not deny redefinition of predefined type-names
   or routines (e.g. you may define a TYPE INTEGER = CHAR;)
   so the RIDPOINTERS array is not useful in here. */

static void
declare_known_id (name, type, value)
char *name;
tree type;
int value;
{
    tree id = get_identifier (name);
    tree decl;

    if (type) {
	type = build_decl (TYPE_DECL, id, type);
	TYPE_NAME (TREE_TYPE (type)) = type;
    } else {
	/* This prevents core dumps... */
	type = integer_type_node;
    }

    /* It really does not matter which DECL node it is, since
       it has the attribute flag PASCAL_REDEFINABLE_DECL(NODE) set.
       It's not used as a VAR_DECL node anyhow. */
    decl = build_decl (VAR_DECL, id, type);
    PASCAL_REDEFINABLE_DECL (decl) = 1;
    if (value) {
	DECL_INITIAL (decl) = build_int_2 (value, 0);
    }
    pushdecl(decl);
}

/* Pre-declared identifiers; i.e. not type names */
static struct known_id {
    char *name;
    int  value;
} KItable [] = {
    { "Maxint",  p_MAXINT },
    { "False",   p_FALSE },
    { "True",    p_TRUE },
    { "Input",   p_INPUT },
    { "Output",  p_OUTPUT },

    { "Rewrite", p_REWRITE },
    { "Reset", 	 p_RESET },
    { "Put", 	 p_PUT },
    { "Get", 	 p_GET },
    { "Write", 	 p_WRITE },
    { "Read", 	 p_READ },
    { "Writeln", p_WRITELN },
    { "Readln",  p_READLN },
    { "Page", 	 p_PAGE },
    { "New", 	 p_NEW },
    { "Dispose", p_DISPOSE },
    { "Abs", 	 p_ABS },
    { "Sqr", 	 p_SQR },
    { "Sin", 	 p_SIN },
    { "Cos", 	 p_COS },
    { "Exp", 	 p_EXP },
    { "Ln", 	 p_LN },
    { "Sqrt", 	 p_SQRT },
    { "Arctan",  p_ARCTAN },
    { "Trunc", 	 p_TRUNC },
    { "Round", 	 p_ROUND },
    { "Pack", 	 p_PACK },
    { "Unpack",  p_UNPACK },
    { "Ord", 	 p_ORD },
    { "Chr", 	 p_CHR },
    { "Succ", 	 p_SUCC },
    { "Pred", 	 p_PRED },
    { "Odd", 	 p_ODD },
    { "Eof", 	 p_EOF },
    { "Eoln", 	 p_EOLN },

    /* Extended Pascal required words */
    {"Otherwise",OTHERWISE },	/* Not currently handled as a RESERVED word */
    {"Maxchar",  p_MAXCHAR},
    {"Maxreal",  p_MAXREAL},
    {"Minreal",  p_MINREAL},
    {"Epsreal",  p_EPSREAL},
    /* Extended Pascal required procs & funs */
    {"Gettimestamp",p_GETTIMESTAMP},
    {"Date",	 p_DATE},
    {"Time",	 p_TIME},
    {"Halt",	 p_HALT},
    {"Extend",	 p_EXTEND},
    {"Seekwrite",p_SEEKWRITE},
    {"Seekread", p_SEEKREAD},
    {"Seekupdate",p_SEEKUPDATE},
    {"Empty",	 p_EMPTY},
    {"Update",	 p_UPDATE},
    {"Position", p_POSITION},
    {"Lastposition",p_LASTPOSITION},
    {"Re",	 p_RE},
    {"Im",	 p_IM},
    {"Cmplx",	 p_CMPLX},
    {"Card",	 p_CARD},
#ifdef NOT_YET_IN_GPC

    {"Readstr",	 p_READSTR},
    {"Writestr", p_WRITESTR},

    {"Bind",	 p_BIND},
    {"Unbind",	 p_UNBIND},
    {"Binding",	 p_BINDING},

    /* Complex type functions */
    {"Arg",	 p_ARG},
    {"Polar",	 p_POLAR},

    /* String functions */
    {"Length",	 p_LENGTH},
    {"Index",	 p_INDEX},
    {"Substr",	 p_SUBSTR},
    {"Trim",	 p_TRIM},
    {"Eq",	 p_EQ},
    {"Lt",	 p_LT},
    {"Gt",	 p_GT},
    {"Ne",	 p_NE},
    {"Le",	 p_LE},
    {"Ge",	 p_GE},
#endif /* NOT_YET_IN_GPC */

    /* More exotic fruits and birds (gpc extensions) */

    { "Inline",  INLINE },	/* influence from GCC */
    { "Asm", 	 ASM },
    { "Alignof", ALIGNOF },
    { "Break", 	 BREAK },
    { "Continue",CONTINUE },
    { "Return",  RETURN },
    { "Sizeof",  SIZEOF },

    { "Conjugate",   CONJUGATE }, /* Complex conjugate */
    { "Mark", 	 p_MARK },	/* mark the heap state */
    { "Release", p_RELEASE },	/* release heap to last mark */

    { "Default", DEFAULT },	/* Case defaults */
    { "Others",  OTHERS },

    { "Close",   p_CLOSE },	/* Close a file */
    { "Definesize", p_DEFINESIZE }, /* random access file size definition */
    { NULL,      0 }
};

/* called from decl.c after type nodes have been initialized */
declare_known_ids ()
{
    struct known_id *kptr;

    declare_rts_types ();

    /* pre-declared type names */
    declare_known_id ("Integer", integer_type_node, 0);
    declare_known_id ("Real",    double_type_node,  0);
    declare_known_id ("Boolean", boolean_type_node, 0);
    declare_known_id ("Char",    char_type_node,    0);
    declare_known_id ("Text",    text_type_node,    0);
    declare_known_id ("Complex", complex_type_node, 0);
    declare_known_id ("Void",    void_type_node,    0);

    /* pre-declared identifiers */
    for (kptr = KItable; kptr->name; kptr++)
	declare_known_id (kptr->name, NULL_TREE, kptr->value);
}

static tree line_identifier;   /* The identifier node named "line" */

void check_newline ();

void
init_lex ()
{
  extern char *malloc ();

  /* Make identifier nodes long enough for the language-specific slots.  */
  set_identifier_size (sizeof (struct lang_identifier));

  /* Start it at 0, because check_newline is called at the very beginning
     and will increment it to 1.  */
  lineno = 0;

  maxtoken = 40;
  token_buffer = malloc (maxtoken + 2);
}

void
reinit_parse_for_function ()
{
}

static int
skip_white_space (c)
register int c;
{
  register int inside;

  for (;;)
    {
      switch (c)
	{
	case '(':
	  c = getc (finput);
	  if (c != '*')
	    {
	      ungetc (c, finput);
	      return '(';
	    }

	  c = getc (finput);

	  inside = 1;
	  while (inside)
	    {
	      if (! flag_nested_comments && c == '}') 
		{
		  inside = 0;
		  c = getc (finput);
		}
	      else if (c == '*')
		{
		  while (c == '*')
		    c = getc (finput);

		  if (c != ')')
		      continue;
		  else
		    {
		      inside = 0;
		      c = getc (finput);
		    }
		}
	      else if (c == '\n')
		{
		  lineno++;
		  c = getc (finput);
		}
	      else if (c == EOF) {
		error ("unterminated comment");
		inside = 0;
	      } else
		c = getc (finput);
	    }

	  break;

	/* If flag_nested_comments is set, { is not equal to (* and
           } is not equal to *). This allows comment nesting like this
	    { (* foo (* bar *) }
	   and
	    (* {lk{{ ihmeess{ puhuko ru}tsia *)

	   But it violates the standard...
	 */
        case '{':
	  c = getc (finput);
	  while (1) {
	      if (c == '}')
		  break;
	      else if (! flag_nested_comments && c == '*') {
		  while (c == '*')
		      c = getc (finput);

		  if (c != ')')
		      continue;
		  else
		      break;
	      } else if (c == '\n')
		  lineno++;
	      else if (c == EOF) {
		  error ("unterminated comment");
		  break;
	      }
	      c = getc (finput);
	  }
	  if (c != EOF)
	    c = getc (finput);
	  break;

	case '\n':
	  check_newline ();

	case ' ':
	case '\t':
	case '\f':
	case '\r':
	case '\b':
	  c = getc (finput);
	  break;

	case '\\':
	  if (flag_c_escapes) {
	      c = getc (finput);
	      if (c == '\n')
		  lineno++;
	      else
		  error ("stray '\\' in program");
	      c = getc (finput);
	      break;
	  } else
	      return (c);

	default:
	  return (c);
	}
    }
}

/* Make the token buffer longer, preserving the data in it.
   P should point to just beyond the last valid character in the old buffer.
   The value we return is a pointer to the new buffer
   at a place corresponding to P.  */

static char *
extend_token_buffer (p)
     char *p;
{
  int offset = p - token_buffer;

  maxtoken = maxtoken * 2 + 10;
  token_buffer = (char *) realloc (token_buffer, maxtoken + 2);
  if (token_buffer == 0)
    fatal ("virtual memory exceeded");

  return token_buffer + offset;
}

/* At the beginning of a line, increment the line number
   and handle a #line directive immediately following  */

void
check_newline ()
{
  register int c;
  register int token;

  while (1)
    {
      c = getc (finput);
      lineno++;

      if (c != '#')
	{
	  /* If no #, unread the character,
	     except don't bother if it is whitespace.  */
	  if (c == ' ' || c == '\t')
	    return;
	  ungetc (c, finput);
	  return;
	}

      /* If the next four characters are #line, skip them.  */
      if (pedantic)
	  warning("ISO Pascal does not allow #directives");
      c = getc (finput);
      if (c == 'l')
	{
	  if (getc (finput) != 'i'
	      || getc (finput) != 'n'
	      || getc (finput) != 'e')
	    error ("undefined or invalid # directive");
	  c = getc (finput);
	}

      /* Skip whitespace.  */

      while (1)
	{
	  if (! (c == ' ' || c == '\t'))
	    break;
	  c = getc (finput);
	}

      /* If the # is the only nonwhite char on the line,
	 just ignore it.  Check the new newline.  */
      if (c == '\n')
	continue;

      /* Something follows the #; read a token.  */

      ungetc (c, finput);
      token = yylex ();

      if (token == UNSIGNED_INTEGER
	  && TREE_CODE (yylval.ttype) == INTEGER_CST)
	{
	  /* subtract one, because it is the following line that
	     gets the specified number */

	  int l = TREE_INT_CST_LOW (yylval.ttype) - 1;

	  token = yylex ();
	  if (token != STRING_LITERAL
	      || TREE_CODE (yylval.ttype) != STRING_CST)
	    error ("invalid #line");

	  input_filename
	    = (char *) permalloc (TREE_STRING_LENGTH (yylval.ttype) + 1);
	  strcpy (input_filename, TREE_STRING_POINTER (yylval.ttype));
	  lineno = l;
	}
      else
	error ("undefined or invalid # directive");

      /* skip the rest of this line.  */
      while ((c = getc (finput)) != '\n');
    }
}

/* RMS does not like ctype.h? */
#ifndef isalnum
#define isalnum(char) ((char >= 'a' && char <= 'z') || (char >= 'A' && char <= 'Z') || (char >= '0' && char <= '9'))
#define isdigit(char) (char >= '0' && char <= '9')
#endif isalnum

#define ENDFILE -1  /* token that represents end-of-file */

static int
readescape ()
{
  register int c = getc (finput);
  register int count, code;

  switch (c)
    {
    case 'x':
      code = 0;
      count = 0;
      while (1)
	{
	  c = getc (finput);
	  if (!(c >= 'a' && c <= 'f')
	      && !(c >= 'A' && c <= 'F')
	      && !(c >= '0' && c <= '9'))
	    {
	      ungetc (c, finput);
	      break;
	    }
	  code *= 16;
	  if (c >= 'a' && c <= 'f')
	    code += c - 'a' + 10;
	  if (c >= 'A' && c <= 'F')
	    code += c - 'A' + 10;
	  if (c >= '0' && c <= '9')
	    code += c - '0';
	  count++;
	}
      if (count == 0)
	error ("\\x used with no following hex digits");
      return code;

    case '0':  case '1':  case '2':  case '3':  case '4':
    case '5':  case '6':  case '7':
      code = 0;
      count = 0;
      while ((c <= '7') && (c >= '0') && (count++ < 3))
	{
	  code = (code * 8) + (c - '0');
	  c = getc (finput);
	}
      ungetc (c, finput);
      return code;

    case '\\': case '\'': case '"':
      return c;

    case '\n':
      lineno++;
      return -1;

    case 'n':
      return TARGET_NEWLINE;

    case 't':
      return TARGET_TAB;

    case 'r':
      return TARGET_CR;

    case 'f':
      return TARGET_FF;

    case 'b':
      return TARGET_BS;

    case 'a':
      return TARGET_BELL;

    case 'v':
      return TARGET_VT;
    }
  return c;
}

void
yyerror (string)
	char *string;
{
  char buf[200];

  strcpy (buf, string);

  /* We can't print string and character constants well
     because the token_buffer contains the result of processing escapes.  */
  if (end_of_file)
    strcat (buf, " at end of input");
  else if (token_buffer[0] == 0)
    strcat (buf, " at null character");
  else if (token_buffer[0] == '"')
    strcat (buf, " before string constant");
  else if (token_buffer[0] == '\'')
    strcat (buf, " before character constant");
  else if (token_buffer[0] < 040 || token_buffer[0] >= 0177)
    sprintf (buf + strlen (buf), " before character 0%o", token_buffer[0]);
  else
    strcat (buf, " before `%s'");

  error (buf, token_buffer);
}

/* if we scan tokens:
     INTEGER..      : we must return INTEGER now, and TWODOTS maybe next time.
     INTEGER.DIGIT  : we have a floating point number
     INTEGER.others : we have an error

   The next thing is used to avoid two ungetc() calls (only one is guaranteed
   to work) and return the token TWODOTS next time yylex() is called.
 */
static int have_DOTS = 0;

static int nextchar = -1;

/* Skips all of the white space at the current location in the input file.
   Must use and reset nextchar if it has the next character.  */

void
position_after_white_space ()
{
  register int c;

  if (nextchar != -1)
    c = nextchar, nextchar = -1;
  else
    c = getc (finput);

  ungetc (skip_white_space (c), finput);
}

static char *exp_store_loc;

static void
store_exp (what)
     int what;
{
  if (what > 9)
    store_exp (what/10);

  *exp_store_loc++ = (what%10)+'0';
}
     
/* Pack the valid floating point number starting from start
   by skipping the leading fractional zeroes if the mantissa
   is zero. Return the adjust value.
 */
static int
compress_float (start, end)
     char *start;
     char **end;
{
  int adjust  = 0;
  char *first = start;

  while (*first == '0')
    first++;
  if (*first == '.')
    adjust = -1;
  else
    return 0;	/* Nonzero mantissa */

  first++;	/* Skip '.' */

  while (*first == '0')
    {
      first++;
      adjust--;
    }
  
  if (! *first)	/* A zero */
    return 0;

  *start++ = *first++;
  *start++ = '.';

  if (! *first)
    *start++ = '0';
  else
    while (*start++ = *first++);

  *--start = '\000';
  *end = start;
  
  return adjust;
}

int
yylex ()
{
  register int c;
  register char *p;
  register int value;
  int note_underscore; /* if pedantic, gives a warning of '_' in identifiers */

  int check_known_ids = recognize_known_ids;
  recognize_known_ids = 1;

  if (nextchar >= 0)
    c = nextchar, nextchar = -1;
  else
    c = getc (finput);

  /* Effectively do c = skip_white_space (c)
     but do it faster in the usual cases.  */
  while (1)
    switch (c)
      {
      case ' ':
      case '\t':
      case '\f':
      case '\r':
      case '\b':
	c = getc (finput);
	break;

      case '\n':
      case '(':
      case '{':
      case '\\':
	c = skip_white_space (c);
      default:
	goto found_nonwhite;
      }
 found_nonwhite:

  token_buffer[0] = c;
  token_buffer[1] = 0;

  switch (c)
    {
    case EOF:
      end_of_file = 1;
      token_buffer[0] = 0;
      value = ENDFILE;
      break;

    case '$':
      if (dollars_in_ident)
	goto letter;
      return '$';

    case 'A':  case 'B':  case 'C':  case 'D':  case 'E':
    case 'F':  case 'G':  case 'H':  case 'I':  case 'J':
    case 'K':  case 'L':  case 'M':  case 'N':  case 'O':
    case 'P':  case 'Q':  case 'R':  case 'S':  case 'T':
    case 'U':  case 'V':  case 'W':  case 'X':  case 'Y':
    case 'Z':
    case 'a':  case 'b':  case 'c':  case 'd':  case 'e':
    case 'f':  case 'g':  case 'h':  case 'i':  case 'j':
    case 'k':  case 'l':  case 'm':  case 'n':  case 'o':
    case 'p':  case 'q':  case 'r':  case 's':  case 't':
    case 'u':  case 'v':  case 'w':  case 'x':  case 'y':
    case 'z':
    case '_':
  letter:
      {
	char prev = '\000';
	p = token_buffer;
	note_underscore = pedantic;
	while (isalnum (c) || c == '_' || c == '$')
	  {
	    if (c == '_')
	      {
		if (note_underscore)
		  {
		    warning
		      ("ISO Pascal does not allow underscores in identifiers");
		    note_underscore = 0;
		  }
		if (! prev)
		  warning ("Extended Pascal does not allow identifiers to start with an underscore");
		if (prev == '_')
		  warning ("Extended Pascal does not allow two adjacent underscores in an identifier");
	      }
	    if (p >= token_buffer + maxtoken)
	      p = extend_token_buffer (p);

	    if (p == token_buffer)
	      {
		if (islower (c))
		  c = toupper (c); /* all words start with a CAPITAL letter */
	      }
	    else if (isupper (c))
	      c = tolower(c);	   /* Leave only the first letter upper case */

	    *p++ = prev = c;
	    c = getc (finput);
	  }
      }
      *p = 0;
      nextchar = c;
      
      value = IDENTIFIER;
      yylval.itype = 0;
      
      /* Recognize reserved words */
      
      {
	register struct resword *ptr;
	
	if (ptr = is_reserved_word (token_buffer, p - token_buffer))
	  {
	    if (ptr->iclass >= flag_what_pascal)
	      value = (int) ptr->token;
	  }
      }
      
      /* If we did not find a keyword, look for an identifier declaration */
      
      if (value == IDENTIFIER)
	{
          yylval.ttype = get_identifier (token_buffer);
	  lastiddecl = lookup_name (yylval.ttype);
	  
	  /* if the declaration found has its PASCAL_REDEFINABLE_DECL attribute
	     bit set, it has the "real" type of the node in it's TREE_TYPE field
	     (if it is a predefined type name, it has TYPE_DECL node there;
	     others have integer_type_node (to prevent crashes in gpc-decl.c)
	     and they have the actual value to return in
	     DECL_INITIAL (decl) */

	  if (check_known_ids)
	    {
	      if (lastiddecl != 0 && PASCAL_REDEFINABLE_DECL (lastiddecl))
		if (TREE_TYPE (lastiddecl) != NULL_TREE && /* type specified &*/
		    DECL_INITIAL (lastiddecl) == NULL_TREE)   /* not a constant */
		  lastiddecl = TREE_TYPE (lastiddecl);
		else
		  value = TREE_INT_CST_LOW (DECL_INITIAL (lastiddecl));
	    }
#ifdef NEW_IDENTIFIER
	  /* jtv@hut.fi: this was a test to do things otherwise */
	  else
	    value = NEW_IDENTIFIER;
#endif
	}
      break;

    case '(':
      c = getc (finput);
      if (c == '.') {
	  c = getc (finput);
	  ungetc (c, finput);
	  if (c == '.') {	/* Allow all params to be unknown (...) */
	      have_DOTS = 1;
	      value = '(';
	  } else {
	      token_buffer[1] = '.';
	      token_buffer[2] = 0;
	      value = LBRACKET;
	  }
      } else {
	  ungetc (c, finput);
	  value = '(';
      }
      break;

    case '.':
      if (have_DOTS == 0)
	  c = getc (finput);
      else
	  have_DOTS = 0;
      switch (c) {
      case '.':
	  token_buffer[1] = '.';
	  c = getc (finput);
	  if (c == '.') {
	      token_buffer[2] = '.';
	      token_buffer[3] = 0;
	      value = ELLIPSIS;
	  } else {
	      ungetc (c, finput);
	      token_buffer[2] = 0;
	      value = TWODOTS;
	  }
	  break;
      case ')':
	  value = RBRACKET;
	  token_buffer[1] = c;
	  token_buffer[2] = 0;
	  break;
      default:
	  ungetc (c, finput);
	  value = '.';
	  break;
      }
      break;

    case '0':  case '1':  case '2':  case '3':  case '4':
    case '5':  case '6':  case '7':  case '8':  case '9':
      {
	int baseflag = 0;
	int base = 10;
	int count = 0;
	int largest_digit = 0;
	int numdigits = 0;
	/* for multi-precision arithmetic,
	   we store only 8 live bits in each short,
	   giving us 64 bits of reliable precision */
	short shorts[8];
	int floatflag = 0;  /* Set 1 if we learn this is a floating constant */

	for (count = 0; count < 8; count++)
	  shorts[count] = 0;

	p = token_buffer;
	*p++ = c;

	if (c == '0' && flag_c_numbers) /* allow c style 0octal, 0xhex etc */
	  {
	    *p++ = (c = getc (finput));
	    if ((c == 'x') || (c == 'X'))
	      {
		base = 16;
		*p++ = (c = getc (finput));
	      }
	    else
	      {
		base = 8;
		numdigits++;
	      }
	  }

	/* Read all the digits.  */

	while (isalnum (c) || (c == '.' && floatflag == 0) ||
	       (c == '#' && baseflag++ == 0)) {
	    if (isdigit (c)) {
		c = c - '0';
	    } else if (c == '.') {
		c = getc (finput);
		if (c == '.') {
		    have_DOTS = 1;
		    ungetc (c, finput); /* only peek at character after . */
		    break;	/* INTEGER.. is not floating point */
		} else {
		    if (base != 10)
			warning("floating constants must be in radix 10");
		    base = 10;
		    if (! isdigit(c)) {
		      if (c == ')')		/* '.)' == ']' */
			{
			  nextchar = '.'; /* we just missed one */
			  break;	  /* Do ungetc() later  */
			}
		      ungetc (c, finput);
		      error("ISO Pascal requires a digit after decimal point");
		      c = '0'; /* assume zero was present */
		    }
		    *p++ = c;
		    c -= '0';
		    floatflag = 1;
		}
	    } else if (c == '#') {
		if (pedantic)
		    warning ("ISO Pascal does not define `radix#value'");
		/* Using Extended Pascals way to input values in different
		 * bases:
		 *     base#value
		 * Base may be in range 2..36
		 */
		base = (shorts[3]<<24) + (shorts[2]<<16) + (shorts[1]<<8) + shorts[0];
		if ((base < 2 || base > 36) ||
		    shorts[7] | shorts[6] | shorts[5] | shorts[4]) {
		    warning ("base value out of range; assuming base 10");
		    base = 10;
		}
	        for (count = 0; count < 8; count++)
		    shorts[count] = 0;
		largest_digit = 0;
		numdigits = 0;
		*p++ = (c = getc (finput));
		continue;
	    } else if (base <= 10) {
		if (base == 10 && (c&~040) == 'E') {
		    floatflag = 1;
		    break;   /* start of exponent */
		}
		error ("nondigits in number whose radix <= 10");
		c = 0;
	    } else if (c >= 'a') {
		c = c - 'a' + 10;
	    } else {
		c = c - 'A' + 10;
	    }
	    if (c >= largest_digit)
		largest_digit = c;
	    numdigits++;
	    
	    if (floatflag == 0)
		for (count = 0; count < 8; count++) {
		    (shorts[count] *= base);
		    if (count) {
			shorts[count] += (shorts[count-1] >> 8);
			shorts[count-1] &= (1<<8)-1;
		    } else shorts[0] += c;
		}
    
	    if (p >= token_buffer + maxtoken - 3)
		p = extend_token_buffer (p);
	    *p++ = (c = getc (finput));
	}

	if (c == '.' && have_DOTS == 0)
	    error("Only one decimal point in floating point number allowed");

	if (numdigits == 0)
	  error ("numeric constant with no digits");

	if (largest_digit >= base)
	  error ("numeric constant contains digits beyond the radix");

	/* Remove terminating char from the token buffer 
	   and delimit the string */
	*--p = 0;

	if (floatflag)
	  {
	    tree type = double_type_node;
	    char f_seen = 0;
	    char l_seen = 0;
	    int esign   = 1;
	    int expon   = 0;
	    char *temp  = p;

	    /* Compress out the leading zeros by adjusting the exponent */
	    int adjust_exp = compress_float (token_buffer, &temp);

	    p = temp;

	    /* Read explicit exponent if any, and put it in tokenbuf.  */

	    if ((c == 'e') || (c == 'E'))
	      {
		if (p >= token_buffer + maxtoken - 3)
		  p = extend_token_buffer (p);
		*p++ = c;
		c = getc (finput);
		if ((c == '+') || (c == '-'))
		  {
		    *p++ = c;
		    if (c == '-')
		      esign = -1;
		    c = getc (finput);
		  }
		if (! isdigit (c))
		  error ("floating constant exponent has no digits");
	        while (isdigit (c))
		  {
		    expon = 10*expon+c-'0';
		    c = getc (finput);
		  }
		expon = esign*expon+adjust_exp;
		expon *= esign;
		exp_store_loc = p;
		store_exp (expon);
		p = exp_store_loc;
	      }
	    else if (adjust_exp)
	      {
		*p++ = 'E';
		if (adjust_exp < 0)
		  {
		    *p++ = '-';
		    adjust_exp = -adjust_exp;
		  }
		exp_store_loc = p;
		store_exp (adjust_exp);
		p = exp_store_loc;
	      }

	    *p = 0;
	    yylval.ttype = build_real (type, atof (token_buffer));

	    if (isalnum (c)) {
		error ("garbage at end of number");
		while (isalnum (c)) {
		    if (p >= token_buffer + maxtoken - 3)
			p = extend_token_buffer (p);
		    *p++ = c;
		    c = getc (finput);
		}
	    }

	    ungetc (c, finput);
	    *p = 0;

	    TREE_TYPE (yylval.ttype) = type;
	    value = UNSIGNED_REAL;
	  }
	else
	  {
	    tree type;

	    if (have_DOTS == 0) { /* already did ungetc for have_DOTS */
		if (isalnum (c)) {
		    error ("garbage at end of number");
		    while (isalnum (c)) {
			if (p >= token_buffer + maxtoken - 3)
			    p = extend_token_buffer (p);
			*p++ = c;
			c = getc (finput);
		    }	
		}
		ungetc (c, finput);
	    }
	    *p = 0;

	    if (shorts[7] | shorts[6] | shorts[5] | shorts[4])
	      warning ("integer constant out of range");

	    /* This is simplified by the fact that our constant
	       is always positive.  */
	    yylval.ttype
	      = build_int_2 ((shorts[3]<<24) + (shorts[2]<<16) + (shorts[1]<<8) + shorts[0],
			     0);
    
	    if (int_fits_type_p (yylval.ttype, integer_type_node))
	      type = integer_type_node;
	    else {
	      error ("integer constant does not fit in integer type");
	      type = integer_type_node;
	    }
	    TREE_TYPE (yylval.ttype) = type;
	    value = UNSIGNED_INTEGER;
	  }
	break;
      }

    case '"' :
	if (pedantic)
	   warning ("ISO Pascal does not allow double quoted strings");
	/* FALLTHROUGH */
    case '\'':
      {
	char quote_char = c;
	c = getc (finput);    /* first char */
	p = token_buffer + 1;
	if (quote_char == '"')
	    value = STRING_LITERAL;
	else
	    value = CHAR_LITERAL; /* Let's all WIRTH together 3 times! */

	while (1) {
	    int num;
	    while(c != quote_char) {
		if (flag_c_escapes && c == '\\') {
		    c = readescape ();
		    if (c < 0)
			goto skipnewline;
		} else if (c == '\n') {
		    if (pedantic)
			if (value == CHAR_LITERAL) /* not accurate :-) */
			    warning ("ISO Pascal forbids newline in char constant");
		        else
			    warning ("string not closed before end of line");
		    lineno++;
		}
		if (p == token_buffer + maxtoken)
		    p = extend_token_buffer (p);
		*p++ = c;
	    skipnewline:
		c = getc (finput);
	    }
	    /* Do we have a closing quote? */
	    num = 0;
	    do {
		if (p == token_buffer + maxtoken)
		    p = extend_token_buffer (p);
		if (num & 1)
		    *p++ = c;
		num++;
		c = getc (finput);
	    } while (c == quote_char);
	    *p = 0;

	    if (p - token_buffer > 2)
		value = STRING_LITERAL;

	    if (num & 1) { /* string/char constant has terminated */
		ungetc (c, finput);
		break; /* while (1) */
	    }
	    /* String did not terminate, continue reading it */
	}

	if (p - token_buffer == 1) {
	    if (value == CHAR_LITERAL)
		error ("Empty character literal -- null assumed");
	    else
		if (pedantic)
		    warning ("Empty string literal");
	}

	if (value == CHAR_LITERAL) {
	    char code = token_buffer[1];
	    /* If char type is signed, sign-extend the constant.  */
	    if (TREE_UNSIGNED (char_type_node)
		|| ((code >> (BITS_PER_UNIT - 1)) & 1) == 0)
	       yylval.ttype = build_int_2(code & ((1 << BITS_PER_UNIT) - 1), 0);
	    else
	       yylval.ttype = build_int_2(code | ((-1) << BITS_PER_UNIT), -1);

	    TREE_TYPE (yylval.ttype) = char_type_node;
        } else { /* string constant */
	    yylval.ttype = build_string (p - token_buffer, token_buffer + 1);
	    TREE_TYPE (yylval.ttype) = char_array_type_node;
	}
	break;
      }      
      
    case '&':
      {
	char c1; 
	token_buffer[ 1 ] = c1 = getc (finput);
        if (c1 == '&')
	  {
	    value = ANDAND; 		/* To take address of a label */
	    token_buffer[ 2 ] = '\000';
	  }
	else
	  {
	    value = '&';
	    ungetc (c1, finput);
	    token_buffer[ 1 ] = '\000';
	  }
      }

    case '+':
    case '-':
    case '<':
    case '>':
    case '*':
    case '/':
    case '=':
    case ':':
      {
	register int c1;

	token_buffer[1] = c1 = getc (finput);
	token_buffer[2] = 0;

	if (c1 == '=') {
	    switch (c) {
	    case ':': value = ASSIGN; goto done;
	    case '<': value = LTE; yylval.code = LE_EXPR; goto done;
	    case '>': value = GTE; yylval.code = GE_EXPR; goto done;
	    }
	} else if (c == '<' && c1 == '>') {
	    value = NEQ; yylval.code = NE_EXPR; goto done;
	} else if (c == '>' && c1 == '<') {
	    value = SYMMETRIC_DIFF; goto done;
	} else if (c == '=' && c1 == '>') {
	    value = RENAME; goto done;
	} else if (c == '*' && c1 == '*') {
	    value = EXPON; goto done;
	}
	ungetc (c1, finput);
	token_buffer[1] = 0;

	switch (c) {
	  case '=':
	    yylval.code = EQ_EXPR; break;
	  case '+':
	    yylval.code = PLUS_EXPR; break;
	  case '-':
	    yylval.code = MINUS_EXPR; break;
	  case '*':
	    yylval.code = MULT_EXPR; break;
	  case '/':
	    yylval.code = RDIV_EXPR; break;
	  case '<':
	    yylval.code = LT_EXPR; break;
	  case '>':
	    yylval.code = GT_EXPR; break;
	}	
	value = c;
	goto done;
      }

    default:
      value = c;
    }

done:
  return value;
}
