/* A Bison parser for the programming language Pascal.
   Copyright (C) 1989, 1990, 1991 Free Software Foundation, Inc.

This file is part of GNU GCC.

GNU GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/*
 * Author: Jukka Virtanen <jtv@hut.fi>
 *         Helsinki University of Technology
 *         Computing Centre
 *         Finland
 *
 * PRE-alpha-version. (Actually it now works also on Alpha :-)
 *
 *
 * Bison parser for ISO 7185 Pascal language originally written
 * 3 Feb 1987 by Jukka Virtanen <jtv@hut.fi>
 *
 * Edit history:
 * jtv: begin modification to gpc at Arpril 1988 Sometime during
 *      June 1988 the gpc-1.21 was already generating some code, a
 *      C-style Pascal language :-} After this I had no time and
 *      lost interest to do this for a long time.
 *
 * A problem: Word INLINE can not be redefined if the syntax is
 * inline procedure ...
 * (The problem with doing it with directives is that
 *  the actual procedure declaration has to be done
 *  without parameter lists and return types. I think this
 *  is silly; but maybe I will add another non-standard way to
 *  go around this...)
 */

/* expect no shift/reduce conflicts */
/* dangling else solved by precedence rules */
%expect 0

%{
#include <stdio.h>
#include <ctype.h>
#include <errno.h>

#include "config.h"
#include "tree.h"
#include "input.h"
#include "c-tree.h"
#include "flags.h"
#include "gpc-defs.h"

#ifndef errno
extern int errno;
#endif

/* Cause the `yydebug' variable to be defined.  */
#define YYDEBUG 1

extern char *token_buffer;	/* Pointer to token buffer.
				   Actual allocated length is maxtoken + 2.  */

extern void warning ();

void yyerror();

/* Like YYERROR but do call yyerror */
#define YYERROR1 { yyerror ("syntax error"); YYERROR; }

void position_after_white_space ();

%}

%start program_component

%union {long itype; tree ttype; enum tree_code code;
        char * filename; int lineno; }

%token IDENTIFIER

/* reserved words */
/* FILE_ has the underscore appended to avoid problems */
%token AND ARRAY BEGIN CASE CONST DIV DO DOWNTO ELSE END
%token FILE_ FOR FUNCTION GOTO IF IN LABEL MOD NIL NOT OF OR
%token PACKED PROCEDURE PROGRAM RECORD REPEAT SET THEN TO TYPE
%token UNTIL VAR WHILE WITH 

/* additional reserved words in Extended Pascal */
%token AND_THEN BINDABLE EXPORT IMPORT MODULE ONLY OR_ELSE
%token OTHERWISE POW PRIVATE PROTECTED QUALIFIED VALUE

%token RENAME 	      /* Extended Pascal token '=>' for export renaming */
%token SYMMETRIC_DIFF /* Extended Pascal: Set symmetric difference (XOR) */

/* The following `r_*' are not tokens, however they are used
 * in the same context as some tokens, so assign unique numbers
 * to them.
 */
%token r_WRITE r_READ r_INITFDR
%token r_SET_IN r_SET_COPY r_SET_RANGE r_SET_SINGLE r_SET_UNION
%token r_SET_DIFF
%token r_SET_INTERSECTION  r_SET_EQUAL r_SET_LE r_SET_SYMDIFF
%token r_SET_CARD

/* redefinable standard identifiers */
%token p_INPUT p_OUTPUT p_REWRITE p_RESET p_PUT p_GET p_WRITE p_READ
%token p_WRITELN p_READLN p_PAGE p_NEW p_DISPOSE 
%token p_ABS p_SQR p_SIN p_COS p_EXP p_LN p_SQRT p_ARCTAN 
%token p_TRUNC p_ROUND p_PACK p_UNPACK p_ORD p_CHR p_SUCC p_PRED
%token p_ODD p_EOF p_EOLN
%token p_MAXINT p_TRUE p_FALSE

/* Additional redefinable identifiers for Extended Pascal */
%token p_EXTEND p_UPDATE p_SEEKWRITE p_SEEKREAD p_SEEKUPDATE
%token p_READSTR p_WRITESTR p_BIND p_UNBIND p_HALT p_GETTIMESTAMP
%token p_ARG p_RE p_IM p_CARD p_CMPLX p_POLAR
%token p_EMPTY p_POSITION p_LASTPOSITION
%token p_LENGTH p_INDEX p_SUBSTR p_TRIM
%token p_EQ p_LT p_GT p_NE p_LE p_GE
%token p_BINDING
%token p_DATE p_TIME 
%token p_MAXCHAR p_MAXREAL p_MINREAL p_EPSREAL

/* redefinable GPC keyword extensions */
%token ASM BREAK CONTINUE p_MARK p_RELEASE RETURN
%token DEFAULT OTHERS INLINE p_CLOSE
%token SIZEOF ALIGNOF ANDAND CONJUGATE
%token p_DEFINESIZE

/* GPC internal tokens */
%token UNSIGNED_INTEGER STRING_LITERAL CHAR_LITERAL UNSIGNED_REAL
%token GTE LTE NEQ EXPON
%token ASSIGN LBRACKET RBRACKET

/* ELLIPSIS, "...", used for functions with variable arglists.  */
%token TWODOTS ELLIPSIS

/* Precedence rules */
/* IF and ELSE are just to take care of dangling else */
%nonassoc IF
%nonassoc ELSE

%right  <code> ASSIGN
%binary	<code> '<'	'='	'>'	IN	NEQ	GTE	LTE
%left	<code> '-'	'+'     OR
%left	<code> '/'	'*'	DIV	MOD	AND
%left	<code> POW	EXPON
%left	<code> NOT

/* types for the nonterminals */
%type <code> for_direction
%type <code> relational_operator multiplying_operator adding_operator
%type <code> sign assign_operator

%type <ttype> IDENTIFIER UNSIGNED_INTEGER UNSIGNED_REAL 
%type <ttype> STRING_LITERAL CHAR_LITERAL NIL

%type <ttype> p_MAXINT p_TRUE p_FALSE
%type <ttype> p_INPUT p_OUTPUT p_REWRITE p_RESET p_PUT p_GET p_WRITE p_READ
%type <ttype> p_WRITELN p_READLN p_PAGE p_NEW p_DISPOSE 
%type <ttype> p_ABS p_SQR p_SIN p_COS p_EXP p_LN p_SQRT p_ARCTAN 
%type <ttype> p_TRUNC p_ROUND p_PACK p_UNPACK p_ORD p_CHR p_SUCC p_PRED p_ODD
%type <ttype> p_EOF p_EOLN

/* Extended Pascal */
%type <ttype> p_EXTEND p_UPDATE p_SEEKWRITE p_SEEKREAD p_SEEKUPDATE
%type <ttype> p_READSTR p_WRITESTR p_BIND p_UNBIND p_HALT p_GETTIMESTAMP
%type <ttype> p_ARG p_RE p_IM p_CARD p_CMPLX p_POLAR
%type <ttype> p_EMPTY p_POSITION p_LASTPOSITION
%type <ttype> p_LENGTH p_INDEX p_SUBSTR p_TRIM
%type <ttype> p_EQ p_LT p_GT p_NE p_LE p_GE
%type <ttype> p_BINDING
%type <ttype> p_DATE p_TIME 
%type <ttype> p_MAXCHAR p_MAXREAL p_MINREAL p_EPSREAL

%type <ttype> OTHERWISE
 
/* redefinable keyword extensions */
%type <ttype> ASM BREAK CONTINUE p_MARK p_RELEASE RETURN
%type <ttype> DEFAULT OTHERS INLINE p_CLOSE
%type <ttype> SIZEOF ALIGNOF CONJUGATE
%type <ttype> p_DEFINESIZE

%type <ttype> optional_par_id_list par_id_list id_list

%type <ttype> program_heading label
%type <ttype> identifier new_identifier case_default
%type <ttype> constant constant_literal predefined_literal 
%type <ttype> number unsigned_number

%type <ttype> typename type_denoter string
%type <ttype> new_ordinal_type new_pointer_type new_structured_type
%type <ttype> pointer_domain_type enumerated_type enumlist enumerator
%type <ttype> subrange_type unpacked_structured_type
%type <ttype> array_type array_index_list 
%type <ttype> file_type set_type ordinal_index_type
%type <ttype> record_type
%type <ttype> record_field_list fixed_part variant_part record_section
%type <ttype> variant_selector variant_list variant
%type <ttype> case_constant_list one_case_constant
%type <ttype> function_heading optional_inline directive
%type <ttype> optional_par_formal_parameter_list parmlist1 
%type <ttype> optional_retval_def functiontype
%type <ttype> conformant_array_schema
%type <ttype> packed_conformant_array_schema
%type <ttype> unpacked_conformant_array_schema
%type <ttype> index_type_specification
%type <ttype> index_type_specification_list
%type <ttype> structured_variable_list
%type <ttype> type_inquiry

%type <ttype> asm_operands asm_operand asm_clobbers nonnull_asm_operands

%type <ttype> variable_access variable_access_no_id identifier_or_typename
%type <ttype> variable_access_or_typename structured_variable
%type <ttype> optional_par_actual_parameter_list
%type <ttype> actual_parameter_list actual_parameter
%type <ttype> index_expression_list optional_filename 

%type <ttype> optional_par_actual_parameter
%type <ttype> optional_par_write_parameter_list
%type <ttype> write_actual_parameter_list write_actual_parameter

%type <ttype> static_expression expression expr_1 boolean_expression 
%type <ttype> simple_expression any_term term primary factor
%type <ttype> standard_functions
%type <ttype> set_constructor_element_list set_constructor member_designator

%type <itype> setspecs
%type <itype> rts_file_open
%type <itype> rts_proc_onepar rts_proc_parlist 
%type <itype> rts_fun_onepar rts_fun_optpar rts_fun_twopar rts_fun_parlist
%type <itype> or_else and_then OR_ELSE AND_THEN

%type <lineno>   save_lineno
%type <filename> save_filename

%{
/* the declaration found for the last IDENTIFIER token read in. */
extern tree lastiddecl;

extern void reinit_parse_for_function ();

/* List of types and structure classes of the current declaration */
extern tree current_declspecs;

/* Stack of saved values of current_declspecs.  */
extern tree declspec_stack;

extern char *input_filename;		/* file being read */
extern char *main_input_filename;	/* top-level source file */

/* A minor kludge for assigning set constructors */
static tree set_assign_destination = NULL_TREE;

static char *concat ();

extern int yylex ();

/* Number of statements (loosely speaking) seen so far.  */
static int stmt_count;

static char *if_stmt_file;
static int if_stmt_line;

/* To allow build_enumerator() to set up the correct type for
 * enumerated type elements (In C they are all ints).
 * Think about [ enumtype ] set constructor.
 */
extern tree current_enum_type;

/*
 * Nonzero if the LABEL declaration part is for the main program
 */
int top_level_labels = 0;
tree main_program_labels;

tree forward_pointer_stack = NULL_TREE;

tree forward_pointer_list  = NULL_TREE;

tree lookup_forward_pointer();
%}

%%

/*
  Normal pascal parser starts at clause main_program_declaration
*/

program_component :
	  main_program_declaration dot_or_error
	| module_declaration dot_or_error
	;

dot_or_error :
	  '.'
	| error
  		{ warning ("Missing `.' at the end of program"); }
	;

/* ISO Pascal parser starts here */

/* @@@@@ NOTE: Non-local gotos to main program DO NOT WORK
   in this version.
 */

main_program_declaration :
	  program_heading semi
  		{ 
		  /* This code tries to declare the function_decl node
		     of the main program BEFORE the main program is actually
		     declared. This is required by label declarations with
		     non-local gotos. */

		  main_program_context =
		    declare_main_program_context
		      (build_tree_list (NULL_TREE, /* declspecs */
					void_type_node),
		       build_nt (CALL_EXPR,	    /* declarator */
				 get_identifier(GPC_MAIN),
				 /* next is a void terminated paramlist */
				 tree_cons (NULL_TREE,
					    NULL_TREE,
					    tree_cons (NULL_TREE,
						       void_type_node,
						       NULL_TREE)),
		                 NULL_TREE));

		  top_level_labels = 1;
		}
	  simple_declaration_part
		{
		  associate_external_files ();
		  $<ttype>$ = getdecls ();
		  top_level_labels = 0;
#ifdef NOTYET
		  push_c_function_context ();
#else
		  main_program_labels = declared_labels;
#endif
		  current_function_decl = NULL_TREE;
	        }
	  function_declaration_part
		{ 
		  tree link;

#ifdef NOTYET
		  pop_c_function_context ();
#else
		  /* Restore the labels for main program */
		  declared_labels = main_program_labels;
#endif
		  current_function_decl = NULL_TREE; /* avoid infinite loops */
		  if (! start_function (NULL_TREE, /* Special case for GPC */
					main_program_context, /* handled in start_function() */
					0))
		      YYERROR1;
		  
		  reinit_parse_for_function ();
		  store_parm_decls ();
		  un_initialize_block ($<ttype>5, 0);
		  init_std_files ();
		}
  	  statement_part 
		{ un_initialize_block ($<ttype>5, 1);
		  finish_function (0);
		}
	;

program_heading :
	  PROGRAM new_identifier optional_par_id_list
		{ external_file_name_list = $3;
		  $$ = $2;
		}
	| PROGRAM error
		{ $$ = error_mark_node; }
	;

optional_par_id_list :
	  /* empty */
		{ $$ = NULL_TREE; }
	| par_id_list
	;

par_id_list :
	  '(' id_list ')'
		{ $$ = $2; }
	| '(' error ')'
		{ $$ = build_tree_list (NULL_TREE, NULL_TREE); }
	;

id_list :
	  new_identifier 
		{ $$ = build_tree_list (NULL_TREE, $1); }
	| id_list ',' new_identifier 
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $3)); }
	;

typename :
	  IDENTIFIER
		{ if (lastiddecl != 0 && TREE_CODE (lastiddecl) == TYPE_DECL)
		    {
		      $$ = TREE_TYPE (lastiddecl);
		      if (TREE_CODE ($$) == POINTER_TYPE)
			{
			  /* If this type name refers to a pointer
			   * that was forward defined return the type
			   * it points to instead of the name
			   *
			   * This solves the following problem:
			   *
			   * foo = ^bar;
			   * bar = record
			   *          z: foo;
			   *       end;
			   */

			  tree fptr = lookup_forward_pointer
			    			(DECL_NAME(lastiddecl), 0);

			  if (fptr)
			    $$ = fptr;
			}
		    }
		  else {
		    error ("type name expected -- identifier `%s' given", 
			   IDENTIFIER_POINTER ($1));
		    $$ = error_mark_node;
		  }
		}
	;

identifier :
	  IDENTIFIER
		{
		    $$ = check_identifier (lastiddecl, $1);
		}
	;

new_identifier :
	  IDENTIFIER
/* maxint    */
	| p_MAXINT
/* boolean values */
	| p_TRUE
	| p_FALSE
/* maxchar */
	| p_MAXCHAR
/* real values */
	| p_MAXREAL
	| p_MINREAL
	| p_EPSREAL
/* standard files */
	| p_INPUT
	| p_OUTPUT
/* Pascal I/O */
	| p_REWRITE
	| p_RESET
	| p_PUT
	| p_GET
	| p_WRITE
	| p_READ
	| p_WRITELN
	| p_READLN
	| p_PAGE
/* Extended Pascal File I/O */
	| p_EXTEND
	| p_UPDATE
	| p_SEEKWRITE
	| p_SEEKREAD
	| p_SEEKUPDATE
	| p_EMPTY
	| p_POSITION
	| p_LASTPOSITION
/* heap handling */
	| p_NEW
	| p_DISPOSE
/* arithmetic functions */	
	| p_ABS
	| p_SQR
	| p_SIN
	| p_COS
	| p_EXP
	| p_LN
	| p_SQRT
	| p_ARCTAN
/* Extended Pascal arithmetic */
	| p_ARG
	| p_RE
	| p_IM
/* transfer functions */
	| p_TRUNC
	| p_ROUND
	| p_PACK
	| p_UNPACK
/* Extended Pascal transfer funs */
	| p_CARD
	| p_CMPLX
	| p_POLAR
/* ordinal functions */
	| p_ORD
	| p_CHR
	| p_SUCC
	| p_PRED
/* boolean functions */
	| p_ODD
	| p_EOF
	| p_EOLN
/* String routines */
	| p_READSTR
	| p_WRITESTR
	| p_LENGTH
	| p_INDEX
	| p_SUBSTR
	| p_TRIM
	| p_EQ
	| p_LT
	| p_GT
	| p_NE
	| p_LE
	| p_GE
/* binding */
	| p_BIND
	| p_UNBIND
	| p_BINDING
/* procedure halt */
	| p_HALT
/* Time routines */
	| p_GETTIMESTAMP
	| p_DATE
	| p_TIME 
/* This is a reserved word in extended Pascal */
	| OTHERWISE
/* extensions to Extended Pascal: */
	| ASM
	| ALIGNOF
	| BREAK
	| CONTINUE
	| p_CLOSE
	| DEFAULT
	| p_MARK
	| OTHERS
	| p_RELEASE
	| RETURN
	| SIZEOF
  	| CONJUGATE
	| p_DEFINESIZE
	;


any_declaration_part :
	  /* empty */
	| any_decl_list
	;

any_decl_list :
	  any_decl
	| any_decl_list any_decl
	;

any_decl :
	  simple_decl
	| function_declaration
	;

simple_declaration_part :
	  /* empty */
	| setspecs
	  simple_decl_list
		{ resume_momentary ($1); }
	;
	      
simple_decl_list :
	  simple_decl
	| simple_decl_list simple_decl
	;

simple_decl :
	  label_declaration_part
	| constant_definition_part
	| type_definition_part
	| variable_declaration_part
	;


/* Label declaration part */

label_declaration_part : 
	  LABEL label_list semi
  	| LABEL error semi
	;

label_list :
	  label
		{ declare_label ($1, top_level_labels); }
	| label_list ',' label
		{ declare_label ($3, top_level_labels); }
	;

/*
 * Labels are returned as identifier nodes for compatibility with gcc
 */
label :
	  UNSIGNED_INTEGER
	      { char *ptr;
		if (pedantic &&
		    ((TREE_INT_CST_HIGH ($1) != 0) ||
		     (TREE_INT_CST_LOW  ($1) > 9999)))
		  warning("ISO Pascal does not allow label values greater than 9999");
		for (ptr = token_buffer; *ptr == '0'; ptr++);
		if (*ptr == '\0') --ptr;
		$$ = get_identifier (ptr); }
	| new_identifier
	      { if (pedantic)
		  warning("ISO Pascal does not allow non-numeric labels");
		$$ = $1;
	      }
	;


/* constant definition part */

constant_definition_part : 
	  CONST constant_definition_list semi
	| CONST error semi
	;

constant_definition_list : 
	  constant_definition
	| constant_definition_list semi constant_definition 
	;

constant_definition : 
	  new_identifier '=' static_expression
		{ tree d = start_decl ($1, tree_cons (NULL_TREE, TREE_TYPE ($3),
				             build_tree_list (NULL_TREE,
							      const_id)),
				       1);
		  TREE_CONSTANT (d) = TREE_CONSTANT ($3);
		  finish_decl (d, $3, NULL_TREE);
		  if (pedantic && TREE_CONSTANT (d) == 0)
		     warning ("constant declared as a read-only variable");
		}
	;

constant :
	  identifier
	| sign identifier
		{ $$ = build_unary_op ($1, $2, 0); }
	| number
	| constant_literal
	;		      

number :
	  sign unsigned_number
		{ $$ = build_unary_op ($1, $2, 0); }
    	| unsigned_number
	;

unsigned_number :
	  UNSIGNED_INTEGER
	| UNSIGNED_REAL
	| p_MAXINT
		{ $$ = copy_node (integer_maxint_node); }
	| p_MAXREAL
		{ if (pedantic)
		    warning ("ISO Pascal does not define MAXREAL");
		  $$ = copy_node (real_max_node);
		}
	| p_MINREAL
		{ if (pedantic)
		    warning ("ISO Pascal does not define MINREAL");
		  $$ = copy_node (real_min_node);
		}
	| p_EPSREAL
		{ if (pedantic)
		    warning ("ISO Pascal does not define EPSREAL");
		  $$ = copy_node (real_eps_node);
		}
	| p_MAXCHAR
		{ if (pedantic)
		    warning ("ISO Pascal does not define MAXCHAR");
		  $$ = copy_node (char_max_node);
		}
	;

sign :
	  '+' 
		{ $$ = CONVERT_EXPR; }
	| '-' 
		{ $$ = NEGATE_EXPR; }
	;

constant_literal :
	  string
		{ $$ = combine_strings ($1); }
	| CHAR_LITERAL
	| predefined_literal
	;

predefined_literal :
	  NIL
		{ $$ = copy_node (null_pointer_node); }
	| p_FALSE
		{ $$ = copy_node (boolean_false_node); }
	| p_TRUE
		{ $$ = copy_node (boolean_true_node); }
	;

string :
	  STRING_LITERAL
	| string STRING_LITERAL
		{ $$ = chainon ($1, $2); }
	;


/* type definition part */

type_definition_part : 
	  TYPE type_definition_list semi
	| TYPE error semi
	;

type_definition_list :
	  type_definition
	| type_definition_list semi type_definition
	;

/* $1 = IDENTIFIER_NODE
   $3 = ..._TYPE node
 */
type_definition :
	  new_identifier '=' type_denoter
		{ tree d = start_decl ($1, tree_cons (NULL_TREE, $3,
						      build_tree_list(NULL_TREE,
								      type_id)),
				       0);

		  if (TREE_CODE ($3) == POINTER_TYPE &&
		      TREE_CODE (TREE_TYPE ($3)) == IDENTIFIER_NODE)
# ifdef OLD_fwd
		    forward_pointer_list =
		      tree_cons (TREE_TYPE ($3),
				 TREE_TYPE (d),
				 forward_pointer_list);
# else
		    forward_pointer_list =
		      tree_cons ($1, $3, forward_pointer_list);
# endif

		  resolve_forward_pointer ($1, $3);

		  finish_decl (d, NULL_TREE, NULL_TREE);
		}
	;

/* this rule must return ..._TYPE nodes */
type_denoter : 
	  typename 
	| new_ordinal_type
	| new_pointer_type
	| new_structured_type
	| type_inquiry
	;
	
new_ordinal_type : 
	  enumerated_type 
	| subrange_type
	;

/* ENUMERATED */

enumerated_type : 
	  '(' setspecs
		{ if (current_enum_type) /* @@@ does it need a stack? */
		    abort ();
		  current_enum_type = $$ = start_enum (NULL_TREE);
		}
	  enumlist ')'
		{ $$ = finish_enum ($<ttype>3, nreverse ($4));
		  current_enum_type = NULL_TREE;
		  resume_momentary ($2); }
	| '(' error ')'
	        { $$ = error_mark_node; } /* @@@@ ??? */
	;

/* We chain the enumerators in reverse order.
   they are later put in forward order */

enumlist:
	  enumerator
	| enumlist ',' enumerator
		{ $$ = chainon ($3, $1); }
	;

/* ord(first_enumerator) == 0 */
/* second arg to build enumerator is the ord number exp of the enumerator */
enumerator:
	  new_identifier
		{ $$ = build_enumerator ($1, NULL_TREE); }
	;

/* SUBRANGE */
/* Here is a problem: In extended Pascal subrange bounds are allowed
   to be arbitrary expressions. If just done the obvious way I get
   49 shift/reduce conflicts. (one for each new_identifier)

   So I currently allow expressions only as the upper bound
   of the range, the lower bound must be a constant until
   I figure out a way to handle this. "(a) .. j;" can be mixed up
   with enumerated_type definition.
 */
subrange_type : 
	  constant TWODOTS expression
		{ if (pedantic && ! TREE_CONSTANT ($3))
		      warning ("ISO Pascal does not allow non-constant range bounds\n");
		  $$ = build_range_type ($1, $3); }
	;

/* POINTER */

new_pointer_type : 
	  pointer_char pointer_domain_type
		{ $$ = build_pointer_type ($2); }
	;

pointer_char :
	  '^'
    	| '@'
	;

/* new_identifier allows forward referencing. */
/* can not use the following as it causes conflicts:
pointer_domain_type :
	  IDENTIFIER
	| typename
	;
*/

/* this is either IDENTIFIER or predefined word */
pointer_domain_type :
	  new_identifier
		{ if (lastiddecl != 0 && TREE_CODE (lastiddecl) == TYPE_DECL)
		    $$ = TREE_TYPE (lastiddecl);
		  else /* a forward declared pointer to an unknown type */
		    $$ = $1;
		}
	;

/* STRUCTURED TYPES */

new_structured_type :
	  PACKED unpacked_structured_type
		{ $$ = grok_packed ($2); }
	| unpacked_structured_type
	;

unpacked_structured_type : 
	  array_type
	| file_type
	| set_type
	| record_type
	;

/* ARRAY */

array_type : ARRAY lbracket array_index_list rbracket OF type_denoter
		{ $$ = build_pascal_array_type ($6, nreverse ($3)); }
	;

array_index_list : 
	  ordinal_index_type
		{ $$ = $1; }
	| array_index_list ',' ordinal_index_type
		{ $$ = chainon ($1, $3); }
	;


/*
 * TREE_LIST whose TREE_PURPOSE is the type and TREE_VALUE
 * is a integer range
 *
 * only for arrays and sets.
 */
ordinal_index_type : 
	  new_ordinal_type
		{ $$ = build_tree_list ($1, convert_type_to_range ($1)); }
	| typename
		{ $$ = build_tree_list ($1, convert_type_to_range ($1)); }
	;

/* FILE */

file_type : 
	  FILE_ OF type_denoter
		{ $$ = build_file_type ($3); }
	;

/* SET */

set_type :
	  SET OF ordinal_index_type
		{ $$ = build_set_type ($3); }
	| SET OF error
		{  /* @@@@ ???? */
		  $$ = build_set_type
		        (build_tree_list
			    (void_type_node,
			     build_index_2_type (size_zero_node,
						 build_int_2 (-1, 0))));
		}
	;

/* RECORD */

record_type :
	  RECORD record_field_list END
		{ $$ = $2; }
	| RECORD error END	/* Return empty record */
		{ $$ = finish_struct (start_struct (RECORD_TYPE, NULL_TREE),
				      NULL_TREE);
		  TYPE_TAG_FIELD ($$) = NULL_TREE;
		}
	;

/* returns RECORD_TYPE */
record_field_list : 
	  /* empty */
		{ $$ = finish_struct (start_struct (RECORD_TYPE, NULL_TREE),
				      NULL_TREE);
		  TYPE_TAG_FIELD ($$) = NULL_TREE;
		}
	| fixed_part optional_semicolon
		{ $$ = finish_struct (start_struct (RECORD_TYPE, NULL_TREE),
				      $1); 
		  TYPE_TAG_FIELD ($$) = NULL_TREE;
		}
	| fixed_part semi variant_part
		{ 
		  /* Chain the tag field or NULL_TREE */
		  tree fields = chainon ($1, TREE_VALUE (TREE_PURPOSE ($3)));
		  /* Chain the variant part */
		  fields = chainon ($1, TREE_VALUE ($3));
		  $$ = finish_struct (start_struct (RECORD_TYPE, NULL_TREE),
				      fields);
		  TYPE_TAG_FIELD ($$) = TREE_PURPOSE ($3);
		}
	| variant_part
		{ 
		  /* Store the tag field or NULL_TREE */
		  tree fields = TREE_VALUE (TREE_PURPOSE ($1));
		  /* Chain the variant part */
		  fields = chainon (fields, TREE_VALUE ($1));
		  $$ = finish_struct (start_struct (RECORD_TYPE, NULL_TREE),
				      fields);
		  TYPE_TAG_FIELD ($$) = TREE_PURPOSE ($1);
		}
	;

/* RECORD FIXED PART */
/* chain on FIELD_DECL */
fixed_part :
	  record_section
	| fixed_part semi record_section
		{ $$ = chainon ($1, $3); }
	;

/* chain of FIELD_DECL */
/* @@@ check *_momentary use here? */
record_section :
	  setspecs id_list ':' type_denoter
		{ $$ = p_grokfields ($2, $4);
		  resume_momentary ($1); }
	;

/* RECORD VARIANTS */

/* returns a TREE_LIST node whose
 * PURPOSE : TREE_LIST of variant_selector
 * VALUE   : unnamed FIELD_DECL node of type UNION_TYPE
 */
variant_part :
	  CASE variant_selector OF variant_list optional_semicolon
		{ $$ = build_tree_list ($2,
			  build_record_variant_part (input_filename,
						     lineno,
						     $2, $4)); }
	;

/* returns a TREE_LIST node whose
 * PURPOSE is the type of the selector
 * and
 * VALUE is a FIELD_DECL node if the variant selector actually exists,
 *       NULL_TREE otherwise.
 */
variant_selector : 
	  new_identifier ':' typename
		{ $$ = build_tree_list ($3,
			  grokfield (input_filename,
				     lineno,
				     $1,
				     build_tree_list (NULL_TREE, $3),
				     NULL_TREE));
		}
	| typename
		{ $$ = build_tree_list ($1, NULL_TREE); }
	;

/* returns a chain of TREE_LIST nodes whose
 * PURPOSE field is the case list
 * and
 * VALUE is the FIELD_DECL node of RECORD_TYPE
 */
variant_list :
	  variant
	| variant_list semi variant
		{ $$ = chainon ($1, $3); }
	;

variant :
	  case_constant_list ':' '(' record_field_list ')'
		{ $$ = build_tree_list ($1,
			 grokfield (input_filename, lineno,
				    NULL_TREE,   /* field name */
				    build_tree_list (NULL_TREE, $4),
				    NULL_TREE)); /* field width */
		}
	;

case_constant_list : 
	  one_case_constant
	| case_constant_list ',' one_case_constant
		{ $$ = chainon ($1, $3); }
	;

one_case_constant :
	  static_expression
		{ $$ = build_tree_list (NULL_TREE, $1); }
	| static_expression TWODOTS static_expression
		{ $$ = build_tree_list ($3, $1); }
	;

/* in Extended Pascal */
type_inquiry :
	  TYPE OF expression
		{ $$ = TREE_TYPE ($3);
		  if (pedantic)
		    warning ("ISO Pascal forbids `TYPE OF'") }
	;

/* variable declaration part */

variable_declaration_part :
	  VAR variable_declaration_list semi
	| VAR error semi
	;

variable_declaration_list :
	  variable_declaration
	| variable_declaration_list semi variable_declaration
	;

variable_declaration : 
	  id_list ':' type_denoter
		{ declare_vars ($1, $3, NULL_TREE); }
	;


/* procedure and function declaration part */

function_declaration_part : 
	  /* empty */
	| function_declaration_list
	| error
	;

function_declaration_list :
	  function_declaration
	| function_declaration_list function_declaration
	;

/* PROCEDURE == void_type_node returning function */
/* FUNCTION */

function_declaration : 
	  function_heading semi directive semi
		{
		  grok_directive (TREE_VALUE ($1), TREE_PURPOSE ($1), $3);
	      	}
	| function_heading semi
		{
		  /* set flag if we are defining a nested function */
		  $<itype>2 = ! global_bindings_p ();

		  if ($<itype>2)
		    push_c_function_context ();
		  if (! ($<ttype>$ =  start_pascal_function (TREE_VALUE($1),
							     TREE_PURPOSE($1),
							     $<itype>2)))
		    {
		      if ($<itype>2)
			pop_c_function_context ();
		      YYERROR1;
		    }
		  reinit_parse_for_function ();
		  store_parm_decls ();
		}
	  pushlevel1
  		{
		  /* Declare the function RESULT_DECL */
		  tree rtype = $<ttype>3;
		  if (rtype != void_type_node)
		    finish_decl (
		       start_decl
			    (TREE_CHAIN ($1),
#if 1
			     /* -O optimized result assignments away,
				so now it has to be volatile */
			     chainon (build_tree_list (NULL_TREE, volatile_id),
				      build_tree_list (NULL_TREE, rtype)),
#else
				      build_tree_list (NULL_TREE, rtype),
#endif
			     0),
		       NULL_TREE,
		       NULL_TREE);
		  push_label_level ();
		}
	  any_declaration_part
		{
	          un_initialize_block (getdecls (), 0);
		}
	  statement_part semi
		{ tree result = DECL_RESULT (current_function_decl);
		  un_initialize_block (getdecls (), 1);

		  pop_label_level ();
		  if (TREE_TYPE (result) != void_type_node)
		    {
		      tree d;
		      /* I do it like this, since if the code
			 directly assigns to DECL_RESULT it will lose
			 on recursive functions.
			     
			 TREE_CHAIN($1) contains the unique name for
			 the return value parameter of this function.
		       */
		      if (! TREE_CHAIN ($1) ||
			  !(d = lookup_name (TREE_CHAIN ($1))))
			abort ();

		      if (   ! PASCAL_VALUE_ASSIGNED (result)
			  && ! PASCAL_VALUE_ASSIGNED (d))
			warning ("return value of function not assigned");

		      expand_expr_stmt
			(build_modify_expr (result, NOP_EXPR, d));
		    }
	        }
	  poplevel1
		{
		  finish_function ($<itype>2);
		  if ($<itype>2)
		    pop_c_function_context ();
		}
	;

/* dummy `no_setspecs' is missing in this version */
function_heading : 
	  optional_inline
	  PROCEDURE
	  new_identifier optional_par_formal_parameter_list
		{ $$ = tree_cons (build_nt (CALL_EXPR, $3, $4, NULL_TREE),
				  tree_cons (NULL_TREE, void_type_node, $1),
				  NULL_TREE);
		}
	| optional_inline
	  FUNCTION 
	  new_identifier optional_par_formal_parameter_list 
	  optional_retval_def
	  functiontype
		{ tree retval = $5;
		  /* If user did not specify a name, we create one. */
		  if (retval == NULL_TREE)
		    retval =
		      get_identifier
			(concat ("retval_", IDENTIFIER_POINTER ($3), ""));

		  $$ = tree_cons (build_nt (CALL_EXPR, $3, $4, NULL_TREE),
				  tree_cons (NULL_TREE, $6, $1),
				  retval);
		}
	;

optional_inline :
	  /* empty */
		{ $$ = NULL_TREE; }
	| INLINE
		{ $$ = build_tree_list (NULL_TREE, inline_id); }
	;

directive :
	  new_identifier
	;

optional_retval_def :
	  /* empty */
		{ $$ = NULL_TREE }
	| '=' new_identifier
		{ if (pedantic)
		    warning ("ISO Pascal does not allow function result variable specification");
		  $$ = $2; }
	;

functiontype :
	  /* Only when function has already been declared with a directive */
          /* this will be altered to the real return type                  */
		{ $$ = void_type_node; }
	| ':' typename
		{ $$ = $2; }
	| error
		{ $$ = void_type_node; }
	;


/* parameter specification section */

optional_par_formal_parameter_list :
		{ pushlevel (0);
	          declare_parm_level (0); }	/* @@@@ New parameter: 0 = declaration */
	  parmlist1
		{ $$ = $2;
		  parmlist_tags_warning ();
		  poplevel (0, 0, 0); }

parmlist1 :
	  /* empty */
		{ $$ = get_parm_info (1); }
	| '(' formal_parameter_list ')'
		{ $$ = get_parm_info (1); }
	| '(' formal_parameter_list optional_semicolon ELLIPSIS ')'
		{ if (pedantic)
		    warning ("ISO Pascal denies variable number of arguments");
		  $$ = get_parm_info (0); }
	| '(' ELLIPSIS ')'
		{ if (pedantic)
		    warning ("ISO Pascal denies variable number of arguments");
		  $$ = get_parm_info (0); }
	| '(' error ')'
		{ $$ = tree_cons (NULL_TREE, NULL_TREE, NULL_TREE); }
	;

formal_parameter_list :
	  formal_parameter
	| formal_parameter_list semi formal_parameter
	;

formal_parameter : 
	  id_list ':' typename
		{ handle_formal_param_list ($1, $3, 0); }
	| VAR id_list ':' typename
		{ handle_formal_param_list ($2, $4, 1); }
	| function_heading
		{ handle_formal_param_list (TREE_PURPOSE ($1), TREE_VALUE ($1), 0); }
	| conformant_array_parameter_specification
	;

conformant_array_parameter_specification :
	  value_conformant_array_specification
	| variable_conformant_array_specification
	;

value_conformant_array_specification : 
	id_list ':' conformant_array_schema
		{ handle_formal_conf_array_param_list ($1, $3, 0); }
	;

variable_conformant_array_specification :
	VAR id_list ':' conformant_array_schema
		{ handle_formal_conf_array_param_list ($2, $4, 1); }
	;

conformant_array_schema : 
	  packed_conformant_array_schema
	| unpacked_conformant_array_schema
	;

/* Packed conformant array formal parameter */
packed_conformant_array_schema :
	  PACKED ARRAY lbracket index_type_specification rbracket OF typename
		{ $$ = build_tree_list (error_mark_node,
					build_tree_list ($4, $7));
		}
	;

/* Unpacked conformant array formal parameter */
unpacked_conformant_array_schema :
	  ARRAY lbracket index_type_specification_list rbracket OF typename
		{ $$ = build_tree_list (NULL_TREE,
					build_tree_list ($3, $6));
		}
	;

index_type_specification :
	  new_identifier TWODOTS new_identifier ':' typename
  		{ $$ = build_tree_list (build_tree_list ($1, $3), $5); }
	;

index_type_specification_list :
	  index_type_specification
	| index_type_specification_list semi index_type_specification
		{ $$ = chainon ($1, $3); } 
	;


/* statement part */

statement_part :
	  compound_statement
  		{ stmt_count++; }
	;

compound_statement :
	  BEGIN pushlevel statement_sequence poplevel END
	;

/* @@@@ Do we need to clear_momentary here??? */
statement_sequence : 
	  lineno_statement
	| statement_sequence semi lineno_statement
	;

save_filename :
		{ $$ = input_filename; }
	;

save_lineno :
		{ $$ = lineno; }
	;

lineno_statement :
  	  save_filename save_lineno statement
  		{ }
	;

statement : 
	  label ':'
		{ tree label =
		   define_label (input_filename, lineno, $1);
		  stmt_count++;
		  emit_nop ();
		  if (label)
		    expand_label (label);
		  position_after_white_space (); }
	  save_filename save_lineno unlabelled_statement
	| unlabelled_statement
	| declaring_statement
	| error
	;

/* GPC extension */
declaring_statement :
	  VAR id_list ':' type_denoter
		{ if (pedantic)
		    warning ("ISO Pascal does not allow variable declarations in statement part\n");
		  declare_vars ($2, $4, NULL_TREE); }
	;

unlabelled_statement :
	  structured_statement
	| simple_statement
	;

/* STRUCTURED STATEMENT */
structured_statement :
	  compound_statement
	| with_statement
	| conditional_statement
	| repetitive_statement
	;

/* WITH STATEMENT */
with_statement :
	  WITH structured_variable_list DO
		{ $<itype>$ = list_length ($2); }
	  lineno_statement
		{ int i;
		  for (i = $<itype>4; i ;i--)
		     poplevel (0, 0, 0);
		}
	;

/* Actually only the length of the list matters */
structured_variable_list : 
	  structured_variable
		{ $$ = build_tree_list (NULL_TREE, $1); }
	| structured_variable_list ',' structured_variable
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $3)); }
	;

structured_variable : 
	  variable_access
		{ if (TREE_CODE (TREE_TYPE ($1)) != RECORD_TYPE)
		      error ("WITH element must be of record type");
		  else
	    	      shadow_record_fields ($1);
		  $$ = $1;
		}
	;

/* CONDITIONAL STATEMENT */
conditional_statement :
	  if_statement 
	| case_statement
	;

/* IF THEN ELSE (shift/reduce conflict avoided with %nonassoc rules) */

simple_if:
	  IF boolean_expression THEN
		{ emit_line_note ($<filename>-1, $<lineno>0);
		  expand_start_cond ($2, 0);
		  $<itype>1 = stmt_count;
		  if_stmt_file = $<filename>-1;
		  if_stmt_line = $<lineno>0;
		  position_after_white_space ();
		}
	  lineno_statement
	;

if_statement :
	  simple_if ELSE
  		{ expand_start_else ();
		  $<itype>1 = stmt_count;
		  position_after_white_space (); }
	  lineno_statement
		{ expand_end_cond ();
		  if (extra_warnings && stmt_count == $<itype>1)
		    warning("empty body in an else statement");
		}
	| simple_if %prec IF
		{ expand_end_cond ();
		  if (extra_warnings && stmt_count == $<itype>1)
		    warning_with_file_and_line (if_stmt_file, if_stmt_line,
						"empty body in an if statement");
		}
	;

/* CASE (see also VARIANT RECORDS) */
case_statement :
	  CASE expression OF
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  c_expand_start_case ($2);
		  /* Don't let the tree nodes for $2 be discarded by
		     clear_momentary during the parsing of the next stmt.  */
		position_after_white_space (); }
	  case_element_list optional_semicolon END
		{ expand_end_case ($2); }
	;

case_element_list : 
	  case_element
	| case_element_list semi case_element
	;

/* now this looks more like a C switch statement */
case_element :
	  case_constant_list ':' 
		{ register tree link;
		  register tree value1, value2;
		  register tree label;
		  tree duplicate;

		  for (link = $1; link; link = TREE_CHAIN (link)) {
		    value1 = TREE_VALUE (link);
		    value2 = TREE_PURPOSE (link);
		    label = build_decl (LABEL_DECL, NULL_TREE, NULL_TREE);

		    /* @@@ use check_case_value ()??? */
		    if (TREE_CODE (value1) != INTEGER_CST
			&& value1 != error_mark_node) {
		      error ("case label does not reduce to an integer constant");
		      continue;
		    }
		    if (value2
			&& TREE_CODE (value2) != INTEGER_CST
			&& value2 != error_mark_node) {
		      error ("upper value of case range does not reduce to an integer constant");
		      continue;
		    }
		    if (value1 != error_mark_node && value2 != error_mark_node) {
		      int success;
		      if (value2) {
			  if (pedantic)
			      warning ("ISO Pascal does not allow case ranges");
			  success = pushcase_range (value1, value2, label, &duplicate);
		      } else
			  success = pushcase (value1, label, &duplicate);
		      if (success == 1)
			error ("case label not within case statement");
		      else if (success == 2)
			{
			  error ("duplicate case label value in case statement");
			  error_with_decl (duplicate, "this is the first entry for that value");
			}
		      else if (success == 3)
			warning ("case label out of range");
		      else if (success == 4)
			warning ("empty case label range");
		    }
		  }
		  position_after_white_space ();
		}
	  lineno_statement
		{ expand_exit_something (); }
	| case_default optional_colon
		{
		  tree duplicate;
		  register tree label
		    = build_decl (LABEL_DECL, NULL_TREE, NULL_TREE);
		  int success = pushcase (NULL_TREE, label, &duplicate);
		  stmt_count++;
		  if (pedantic)
		    warning ("ISO Pascal does not allow default label `%s' in CASE statement",
			     IDENTIFIER_POINTER ($1));
		  if (success == 1)
		    error ("default label not within a case statement");
		  else if (success == 2)
		    {
		      error ("multiple default labels in one case statement");
		      error_with_decl (duplicate, "this is the first entry for that value");
		    }
		  position_after_white_space ();
		}
	  lineno_statement
		{ expand_exit_something (); }
	;

/* all of these are non-standard */
case_default :
	  DEFAULT
	| OTHERS
	| OTHERWISE
	;

/* REPETITIVE STATEMENT */

repetitive_statement : 
	  repeat_statement
	| while_statement
	| for_statement 
	;

/* REPEAT */
repeat_statement :
	  REPEAT
		{ emit_nop ();
		  stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  expand_start_loop_continue_elsewhere (1); }
	  statement_sequence UNTIL
		{ expand_loop_continue_here (); }
	  boolean_expression
		{ emit_line_note (input_filename, lineno);
		  expand_exit_loop_if_false (0,
			build_unary_op (TRUTH_NOT_EXPR, $6, 0));
		  expand_end_loop ();
		  /* clear_momentary (); */ }
	;


/* WHILE */
while_statement :
	  WHILE
		{ emit_nop ();
		  stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  expand_start_loop (1); }
	  boolean_expression 
		{ emit_line_note (input_filename, lineno);
		  expand_exit_loop_if_false (0, $3); }
	  DO lineno_statement
		{ expand_end_loop (); }
	;

/* FOR */
/* I did not manage to do this without a temporary variable.
 *
 *
 *
 *
 * Currently this is done so that we allocate a temporary VAR_DECL
 * for the loop upper bound; and store the expression there.
 *
 * ARGV!!! Both the lower and upper bounds must be evaluated
 * before the for loop control variable is assigned. 
 * if not, the following does not work:
 *    for i:= (i+1) to (i+10) do
 *
 * So allocate another temporary variable for the lower bound...
 */
for_statement :
	  FOR identifier ASSIGN expression for_direction expression
		{ tree for_l_bound, for_u_bound;
		  emit_nop ();
		  stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (TREE_CONSTANT ($4))
		      for_l_bound = $4;
		  else {
		      for_l_bound = start_decl
			  (get_unique_identifier ("for_lower"),
			   build_tree_list(NULL_TREE,
					   TREE_TYPE ($2)),
			   0);
		      finish_decl (for_l_bound, NULL_TREE, NULL_TREE);

		      /* Assign the lower bound to temp */
		      /* @@@ c_expand_expr_stmt() ??? */
		      expand_expr_stmt (build_modify_expr (for_l_bound,
							   NOP_EXPR, $4));
		  }

		  if (TREE_CONSTANT ($6))
		      for_u_bound = $6;
		  else {
		      for_u_bound = start_decl
			  (get_unique_identifier ("for_upper"),
			   build_tree_list(NULL_TREE,
					   TREE_TYPE ($2)),
			   0);
		      finish_decl (for_u_bound, NULL_TREE, NULL_TREE);

		      emit_line_note ($<filename>-1, $<lineno>0);
		      /* and the upper one */
		      expand_expr_stmt (build_modify_expr (for_u_bound,
							   NOP_EXPR, $6));
		  }
		  $<ttype>$ = for_u_bound;

		  /* necessary to prevent infinite loops when
 		   * incrementing/decrementing would cause
		   * wrap-around at maxint/-maxint
		   */
		  expand_start_cond (
		    fold (build_binary_op ($5, for_l_bound,
					   for_u_bound, 0)), 0);
		  emit_line_note ($<filename>-1, $<lineno>0);
		  expand_expr_stmt (build_modify_expr ($2, NOP_EXPR, for_l_bound));
		  expand_start_loop_continue_elsewhere (1);
		}
	  DO
		{ position_after_white_space ();
		}
	  lineno_statement
		{ emit_line_note ($<filename>-1, $<lineno>0);
		  expand_loop_continue_here ();
		  expand_exit_loop_if_false (0,
		     fold (build_binary_op
			   ($5 == LE_EXPR ? LT_EXPR : GT_EXPR,
			    $2, $<ttype>7,0)));
		  expand_expr_stmt (build_unary_op (
		     ($5 == GE_EXPR ? POSTDECREMENT_EXPR : POSTINCREMENT_EXPR),
		     $2, 0));
		  expand_end_loop ();
		  expand_end_cond ();
		}
	;

for_direction :
	  TO     
		{ $$ = LE_EXPR; }
	| DOWNTO 
		{ $$ = GE_EXPR; }
	;

/* SIMPLE STATEMENT */
simple_statement : 
	  empty_statement
	| goto_statement
	| procedure_statement
	| assignment_statement
	| standard_function_statement
	| statement_extensions
	;

empty_statement :
	  /* empty */
	;

goto_statement : 
	  GOTO label
	      	{ tree decl;
		  stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  decl = lookup_label ($2);
		  if (decl != 0) {
		      TREE_USED (decl) = 1;
		      expand_goto (decl); }
	        }
	| GOTO '*' expression
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (pedantic)
		    warning ("ISO Pascal does not allow indirect jumps");
		  expand_computed_goto (convert (ptr_type_node, $3)); }

	;

/* PROCEDURE CALL */

procedure_statement : 
	  identifier optional_par_actual_parameter_list
		{ expand_expr_stmt (build_function_call ($1, $2)); }
	;

optional_par_actual_parameter_list :
	  /* empty */
		{ $$ = NULL_TREE; }
	| '('
		{ $<itype>1 = suspend_function_calls (); }
          actual_parameter_list ')'
		{
		  $$ = $3;
		  resume_function_calls ($<itype>1);
	        }
	;

actual_parameter_list :
	  actual_parameter
		{ $$ = build_tree_list (NULL_TREE, $1); }
	| actual_parameter_list ',' actual_parameter
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $3)); }
	;

actual_parameter : 
	  expression
	;

/* ASSIGNMENT */
/* Should also implement threatening of optional_retval_def */
assignment_statement :
	  variable_access assign_operator expression
		{ tree target = $1;
		  tree source = $3;
		  if (TREE_CODE ($1) == FUNCTION_DECL)
		    {
		      if (TREE_TYPE (TREE_TYPE ($1)) == void_type_node)
			{
			  error ("You can't assign to a procedure name");
			  target = error_mark_node;
			}
		      else
			{
			  if ($1 == current_function_decl)
			    current_function_returns_value = 1;
			  else if (! containing_function_p ($1))
			    {
			      error ("You can't assign to a function not containing this one");
			      target = error_mark_node;
			    }
			  if (target != error_mark_node)
			    {
			      target =
				lookup_name
				  (get_identifier
				   (concat
				    ("retval_",
				     IDENTIFIER_POINTER (DECL_NAME (target)),
				     "")));
			      if (! target)
				abort ();
			    }
			}
		    } 
		  
		  /* To mark that we have assigned to this variable.
		     Currently only used to flag function value assignements,
		     but it cannot be inside the FUNCTION_DECL conditional
		     above, since GPC allows you to define a name for the
		     return value of the function, and that name is not
		     (at least currently) recorded in the function
		     declaration. GPC assigns a VAR_DECL node for the name. */
		  if (TREE_CODE (target) == VAR_DECL)
		    PASCAL_VALUE_ASSIGNED (target) = 1;
		    
		  emit_line_note (input_filename, lineno);

		  /* Construct a set directly to the set variable */
		  if (TREE_CODE (TREE_TYPE (target)) == SET_TYPE &&
		      TREE_CODE (source) == CONSTRUCTOR)
		    source = construct_set (source, target, 0);
		    
		  if (source)
		    expand_expr_stmt
		      (build_modify_expr (target, NOP_EXPR, source));
		}
	;

assign_operator :
  	  ASSIGN
	| '='
		{ warning ("Using '=' instead of ':=' in assignment");
		  $<code>$ = $1;
		}
  	;

standard_function_statement :
	  p_HALT optional_par_actual_parameter
		{ if (pedantic && $2 != NULL_TREE)
		      warning ("Extended Pascal does not allow parameters to HALT");
		  build_rts_call (p_HALT, $2); }
	| rts_file_open '(' actual_parameter optional_filename ')'
		{ if (pedantic && $4 != NULL_TREE)
		    warning ("ISO Pascal does not allow file name parameter to reset/rewrite");
		  build_rts_call ($1, 
				 tree_cons (NULL_TREE, $3,
				            build_tree_list (NULL_TREE, $4)));
		}
	| rts_proc_onepar '(' actual_parameter ')'
		{ build_rts_call ($1, build_tree_list(NULL_TREE, $3)); }
	| rts_proc_parlist '(' actual_parameter_list ')'
		{ build_rts_call ($1, $3); }
	| p_WRITE '(' write_actual_parameter_list ')'
		{ build_rts_call (p_WRITE, $3); }
	| p_WRITELN optional_par_write_parameter_list
		{ build_rts_call (p_WRITELN, $2); }
	| p_READLN optional_par_actual_parameter_list
		{ build_rts_call (p_READLN, $2); }
	;

rts_file_open :
	  p_REWRITE
		{ $$ = p_REWRITE; }
	| p_RESET
		{ $$ = p_RESET; }
	| p_EXTEND
		{ $$ = p_EXTEND; }
	;

optional_filename :
	  /* empty */
		{ $$ = null_pointer_node; }
	| ',' actual_parameter
		{ $$ = $2; }
	;

optional_par_write_parameter_list :
	  /* empty */
		{ $$ = NULL_TREE; }
	| '(' write_actual_parameter_list ')'
		{ $$ = $2; }
	;

write_actual_parameter_list :
	  write_actual_parameter
	| write_actual_parameter_list ',' write_actual_parameter
		{ $$ = chainon ($1, $3); }
	;


/* how to represent the : expressions? */
/* yes, as a list in TREE_PURPOSE of each actual parameter */
write_actual_parameter :
	  actual_parameter
		{ $$ = build_tree_list (NULL_TREE, $1); }
 	| actual_parameter ':' expression
		{ $$ = build_tree_list (build_tree_list (NULL_TREE, $3), $1); }
	| actual_parameter ':' expression ':' expression
		{ $$ = build_tree_list (build_tree_list ($5, $3), $1); }
	;

/* run time system calls with one parameter */
rts_proc_onepar :
	  p_PUT
		{ $$ = p_PUT; }
	| p_GET
		{ $$ = p_GET; }
	| p_PAGE
		{ $$ = p_PAGE; }
	| p_MARK
		{ $$ = p_MARK; }
	| p_RELEASE
		{ $$ = p_RELEASE; }
	| p_CLOSE
		{ $$ = p_CLOSE; }
	| p_UPDATE
		{ $$ = p_UPDATE; }
	| p_GETTIMESTAMP
		{ $$ = p_GETTIMESTAMP; }
	;

rts_proc_parlist :
	  p_READ	/* varargs */
		{ $$ = p_READ; }
	| p_NEW		/* varargs */
		{ $$ = p_NEW; }
	| p_DISPOSE	/* varargs */
		{ $$ = p_DISPOSE; }
	| p_PACK	/* Three args */
		{ $$ = p_PACK; }
	| p_UNPACK	/* Three args */
		{ $$ = p_UNPACK; }
	| p_BIND	/* Two args */
		{ $$ = p_BIND; }
	| p_SEEKREAD
		{ $$ = p_SEEKREAD; }
	| p_SEEKWRITE
		{ $$ = p_SEEKWRITE; }
	| p_SEEKUPDATE
		{ $$ = p_SEEKUPDATE; }
/* GPC extension: Define random access file size */
	| p_DEFINESIZE  /* Two args */
		{ $$ = p_DEFINESIZE; }
	;

/* statement extensions to ISO Pascal */
statement_extensions :
	  return_statement
	| continue_statement
	| break_statement
	| asm_statement
	;

return_statement :
	  RETURN
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (pedantic)
		    warning ("ISO Pascal does not allow return statement");
		  c_expand_return (NULL_TREE); }
	| RETURN expression
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (pedantic)
		    warning ("ISO Pascal does not allow return statement");
		  current_function_returns_value = 1;
		  c_expand_return ($2); }
	;

break_statement :
	  BREAK
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (pedantic)
		    warning ("ISO Pascal does not allow break statement");
		  if ( ! expand_exit_something ())
		    error ("break statement not within loop or case"); }
	;

continue_statement :
	  CONTINUE
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (pedantic)
		    warning ("ISO Pascal does not allow continue statement");
		  if (! expand_continue_loop (0))	/* @@@ NP */
		    error ("continue statement not within a loop"); }
	;

asm_statement :
	  ASM asm_qualifier '(' string ')'
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (TREE_CHAIN ($4)) $4 = combine_strings ($4);
		  expand_asm ($4); }
	/* This is the case with just output operands.  */
	| ASM asm_qualifier '(' string ':' asm_operands ')'
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (TREE_CHAIN ($4)) $4 = combine_strings ($4);
		  c_expand_asm_operands ($4, $6, NULL_TREE, NULL_TREE,
					 0, input_filename, lineno); }
	/* This is the case with input operands as well.  */
	| ASM asm_qualifier '(' string ':' asm_operands ':' asm_operands ')'
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (TREE_CHAIN ($4)) $4 = combine_strings ($4);
		  c_expand_asm_operands ($4, $6, $8, NULL_TREE,
					 0, input_filename, lineno); }
	/* This is the case with clobbered registers as well.  */
	| ASM asm_qualifier '(' string ':' asm_operands ':' asm_operands ':' asm_clobbers ')'
		{ stmt_count++;
		  emit_line_note ($<filename>-1, $<lineno>0);
		  if (TREE_CHAIN ($4)) $4 = combine_strings ($4);
 		  c_expand_asm_operands ($4, $6, $8, $10,
					 0, input_filename, lineno); }
	;

/* Now yet decided what to do */
asm_qualifier:
  	  /* empty */
	        { if (pedantic)
		    warning ("ISO Pascal forbids use of `asm' keyword"); } 
	;

/* These are the operands other than the first string and colon
   in  asm ("addextend %2,%1": "=dm" (x): "0" (y), "g" (*x))  */
asm_operands: /* empty */
		{ $$ = NULL_TREE; }
	| nonnull_asm_operands
	;

nonnull_asm_operands:
	  asm_operand
	| nonnull_asm_operands ',' asm_operand
		{ $$ = chainon ($1, $3); }
	;

asm_operand:
	  STRING_LITERAL '(' expression ')'
		{ $$ = build_tree_list ($1, $3); }
	;

asm_clobbers:
	  string
		{ $$ = tree_cons (NULL_TREE, $1, NULL_TREE); }
	| asm_clobbers ',' string
		{ $$ = tree_cons (NULL_TREE, $3, $1); }
	;


/* variable access */
/* also matches function identifier to avoid conflicts */
variable_access : 
	  identifier
		{ 
		  if (TREE_CODE (TREE_TYPE ($1)) == REFERENCE_TYPE)
		      $$ = build_indirect_ref ($1, "VAR parameter ref"); 
		  else
		      $$ = $1;
		}
	| variable_access_no_id
  	;

/* Separated for sizeof & alignof to avoid s/r conflicts */
variable_access_no_id :
	  p_OUTPUT
		{ $$ = output_file_node; }
	| p_INPUT
		{ $$ = input_file_node; }
	| variable_access '.' new_identifier
		{ $$ = build_component_ref ($1, $3); }
	| variable_access pointer_char
		{ if (TREE_CODE (TREE_TYPE ($$)) == FILE_TYPE)
		    $$ = build_buffer_ref ($1);
		  else
		    $$ = build_indirect_ref ($1, "^");
		}
	| variable_access lbracket index_expression_list rbracket
		{ $$ = build_pascal_array_ref ($1, $3); }
	;

/* For sizeof & alignof */
variable_access_or_typename :
  	  variable_access_no_id
	| identifier_or_typename
	;

identifier_or_typename :
	  IDENTIFIER
		{ if (lastiddecl != 0 && TREE_CODE (lastiddecl) == TYPE_DECL)
		    $$ = lastiddecl;
		  else {
		    tree id = check_identifier (lastiddecl, $1);
		    if (TREE_CODE (TREE_TYPE (id)) == REFERENCE_TYPE)
		      $$ = build_indirect_ref (id, "VAR parameter ref"); 
		    else
		      $$ = id;
		  }
		}
	;

index_expression_list :
	  expression
		{ $$ = build_tree_list (NULL_TREE, $1); }
	| index_expression_list ',' expression
		{ $$ = chainon ($1, build_tree_list (NULL_TREE, $3)); }
	;


/* expressions */

/* We should check that this really is compile time constant */
static_expression :
	  expression
	;

/* the expression result must be type boolean */
/* This should be checked */
boolean_expression :
	  expr_1 
		{ $$ = fold ($1); }
	;

expression :
	  expr_1
		{ $$ = fold ($1); }

expr_1 :
	  simple_expression relational_operator simple_expression
		{ $$ = build_pascal_op ($2, $1, $3); }
	| simple_expression IN simple_expression
		{ $$ = build_pascal_op (IN_EXPR, $1, $3); }
	| simple_expression
	;

simple_expression :
	  any_term
	| simple_expression adding_operator any_term
		{ $$ = build_pascal_op ($2, $1, $3); }
	| simple_expression OR any_term
		{ enum tree_code code =
			flag_short_circuit ? TRUTH_ORIF_EXPR : TRUTH_OR_EXPR;
		  $$ = build_binary_op (code, $1, $3, 0); }
	| simple_expression or_else any_term
		{ if (pedantic)
		      warning ("ISO Pascal does not allow 'OR_ELSE'");
		  $$ = build_binary_op (TRUTH_ORIF_EXPR, $1, $3, 0); }
	;

or_else :
	  OR ELSE
		{ $$ = OR_ELSE; }
	| OR_ELSE
	;

and_then :
	  AND THEN
		{ $$ = AND_THEN; }
	| AND_THEN
	;


any_term :
	  sign term
		{ $$ = build_unary_op ($1, $2, 0); }
	| term
	;


term :
	  primary
	| term multiplying_operator primary
		{ $$ = build_pascal_op ($2, $1, $3); }
	| term SYMMETRIC_DIFF primary
		{ $$ = build_pascal_op (BIT_XOR_EXPR, $1, $3); }
	| term AND primary
		{ enum tree_code code =
			flag_short_circuit ? TRUTH_ANDIF_EXPR : TRUTH_AND_EXPR;
		  $$ = build_binary_op (code, $1, $3, 0); }
	| term and_then primary
		{ if (pedantic)
		    warning ("ISO Pascal does not allow 'AND_THEN'");
		  $$ = build_binary_op (TRUTH_ANDIF_EXPR, $1, $3, 0); }
	;


primary :
  	  factor
	| primary POW factor
		{ if (TREE_CODE (TREE_TYPE ($3)) != INTEGER_TYPE)
		    {
		      error ("`pow' exponent is not of integer type");
		      $$ = error_mark_node;
		    }
		  else
		    $$ = build_pascal_op (EXPON_EXPR, $1, $3);
		}
	| primary EXPON factor
		{ if (TREE_CODE (TREE_TYPE ($3)) != REAL_TYPE)
		    {
		      error ("`**' exponent is not of real type");
		      $$ = error_mark_node;
		    }
		  else
		    $$ = build_pascal_op (EXPON_EXPR, $1, $3);
		}
	;


factor :
	  variable_access	 
		{ tree temp = $1;
		  if (TREE_CODE ($1) == FUNCTION_DECL
		      || (TREE_CODE ($1) == PARM_DECL
			  && TREE_CODE (TREE_TYPE ($1)) == POINTER_TYPE
			  && TREE_CODE (TREE_TYPE (TREE_TYPE ($1))) ==
			     FUNCTION_TYPE))
		     $$ = maybe_call_function ($1);
		  else
		      $$ = $1;
		}
	| constant_literal
	| unsigned_number
	| set_constructor
	| identifier '('
		{ $<itype>2 = suspend_function_calls ();
		  $$ = lastiddecl;
		}
	  actual_parameter_list ')'
		{ resume_function_calls ($<itype>2);
		  if ($<ttype>3 && TREE_CODE ($<ttype>3) == TYPE_DECL) {
		      if (pedantic)
			 warning ("ISO Pascal does not allow type casts");
		      if (list_length ($4) != 1) {
			 error ("type cast expects one expression argument");
			 $$ = error_mark_node;
		      } else {
			 tree type = 
			   groktypename (
				build_tree_list (build_tree_list (NULL_TREE,
						    TREE_TYPE ($<ttype>3)),
						 DECL_NAME ($<ttype>3)));
		         $$ = build_c_cast (type, TREE_VALUE ($4));
		      }
		  } else
		      $$ = build_function_call ($1, $4);
		}
	| '(' expression ')'
		{ $$ = $2; }
	/* Check later that NOT is taken from BOOLEAN_TYPE */
	| NOT factor
		{ $$ = build_unary_op (TRUTH_NOT_EXPR, $2, 0); }
	| standard_functions
	| '&' identifier
		{ if (pedantic)
		    warning ("ISO Pascal does not allow to take addresses of objects");
		  $$ = build_unary_op (ADDR_EXPR, $2, 0); }

/* More GNU Pascal extensions */
	/* Refer to the address of a label as a pointer.  */
        /* For computed GOTO * statement */
	| ANDAND label
		{ tree label = lookup_label ($2);
		  TREE_USED (label) = 1;
		  $$ = build1 (ADDR_EXPR, ptr_type_node, label);
		  TREE_CONSTANT ($$) = 1; }
	| CONJUGATE '(' expression ')'
		{ $$ = build_unary_op (CONJ_EXPR, $3, 1); }
	| SIZEOF '(' variable_access_or_typename ')'
		{ if (TREE_CODE ($3) == COMPONENT_REF
		      && DECL_BIT_FIELD (TREE_OPERAND ($3, 1)))
		    error ("`sizeof' applied to a bit-field");
		  $$ = c_sizeof (TREE_TYPE ($3)); }
	| ALIGNOF '(' variable_access_or_typename ')'
		{ if (TREE_CODE ($3) == COMPONENT_REF
		      && DECL_BIT_FIELD (TREE_OPERAND ($3, 1)))
		    error ("`alignof' applied to a bit-field");
		  if (TREE_CODE ($3) == INDIRECT_REF)
		    {
		      tree t = TREE_OPERAND ($3, 0);
		      tree best = t;
		      int bestalign = TYPE_ALIGN (TREE_TYPE (TREE_TYPE (t)));
		      while (TREE_CODE (t) == NOP_EXPR
			     && TREE_CODE (TREE_TYPE (TREE_OPERAND (t, 0))) == POINTER_TYPE)
			{
			  int thisalign;
			  t = TREE_OPERAND (t, 0);
			  thisalign = TYPE_ALIGN (TREE_TYPE (TREE_TYPE (t)));
			  if (thisalign > bestalign)
			    best = t, bestalign = thisalign;
			}
		      $$ = c_alignof (TREE_TYPE (TREE_TYPE (best)));
		    }
		  else
		    $$ = c_alignof (TREE_TYPE ($3)); }
	;


/* SET CONSTRUCTOR,
 * save some work with a small kludge.
 */
set_constructor : lbracket set_constructor_element_list rbracket
		{ $$ = build_set_constructor ($2); }
	;

set_constructor_element_list :
	  /* empty */
		{ $$ = NULL_TREE; }
	| member_designator
	| set_constructor_element_list ',' member_designator
		{ if ($3 != NULL_TREE) /* Don't chain empty ranges */
		    $$ = chainon ($1, $3);
		}
	;

/* Returns a TREE_LIST node with TREE_PURPOSE as the range low bound,
   and TREE_VALUE as the range upper bound */
member_designator : 
	  expression
		{ $$ = construct_set_member ($1, NULL_TREE); }
	| expression TWODOTS expression
		{ $$ = construct_set_member ($1, $3);}
	;

standard_functions :
	  rts_fun_onepar '(' actual_parameter ')'
		{ $$ = build_rts_call ($1, build_tree_list (NULL_TREE, $3)); }
	| rts_fun_twopar '(' actual_parameter ',' actual_parameter ')'
		{ $$ = build_rts_call ($1,
			      chainon (build_tree_list (NULL_TREE, $3),
				       build_tree_list (NULL_TREE, $5)));
		}
	| rts_fun_optpar optional_par_actual_parameter
		{ $$ = build_rts_call ($1, $2); }
	| rts_fun_parlist '(' actual_parameter_list ')'
		{ $$ = build_rts_call ($1, $3); }
	;


optional_par_actual_parameter :
	  /* empty */
		{ $$ = NULL_TREE; }
	|  '(' actual_parameter ')'
		{ $$ = build_tree_list (NULL_TREE, $2); }
	;

rts_fun_optpar :
	  p_EOF 
		{ $$ = p_EOF; }
	| p_EOLN
		{ $$ = p_EOLN; }
	;

rts_fun_onepar :
/* arithmetic functions */	
	  p_ABS
		{ $$ = p_ABS; }
	| p_SQR
		{ $$ = p_SQR; }
	| p_SIN
		{ $$ = p_SIN; }
	| p_COS
		{ $$ = p_COS; }
	| p_EXP
		{ $$ = p_EXP; }
	| p_LN
		{ $$ = p_LN; }
	| p_SQRT
		{ $$ = p_SQRT; }
	| p_ARCTAN
		{ $$ = p_ARCTAN; }
	| p_ARG
		{ $$ = p_ARG; }
	| p_RE
		{ $$ = p_RE; }
	| p_IM
		{ $$ = p_IM; }
/* transfer functions */
	| p_TRUNC
		{ $$ = p_TRUNC; }
	| p_ROUND
		{ $$ = p_ROUND; }
	| p_CARD
		{ $$ = p_CARD; }
/* ordinal functions */
	| p_ORD
		{ $$ = p_ORD; }
	| p_CHR
		{ $$ = p_CHR; }
/* boolean functions */
	| p_ODD
		{ $$ = p_ODD; }
	| p_EMPTY
		{ $$ = p_EMPTY; }
/* Direct access position functions */
	| p_POSITION
		{ $$ = p_POSITION; }
	| p_LASTPOSITION
		{ $$ = p_LASTPOSITION; }
/* String functions */
	| p_LENGTH
		{ $$ = p_LENGTH; }
	| p_TRIM
		{ $$ = p_TRIM; }
/* Binding functions */
	| p_UNBIND
		{ $$ = p_UNBIND; }
/* time functions */
	| p_DATE
		{ $$ = p_DATE; }
	| p_TIME
		{ $$ = p_TIME; }
	;

rts_fun_twopar :
/* Complex functions */
	  p_CMPLX
		{ $$ = p_CMPLX; }
	| p_POLAR
		{ $$ = p_POLAR; }
/* String functions */
	| p_INDEX
		{ $$ = p_INDEX; }
	| p_EQ
		{ $$ = p_EQ; }
	| p_LT
		{ $$ = p_LT; }
	| p_GT
		{ $$ = p_GT; }
	| p_NE
		{ $$ = p_NE; }
	| p_LE
		{ $$ = p_LE; }
	| p_GE
		{ $$ = p_GE; }
	;

rts_fun_parlist :
	  p_SUCC	/* One or two args */
		{ $$ = p_SUCC; }
	| p_PRED	/* One or two args */
		{ $$ = p_PRED; }
        | p_SUBSTR	/* Two or three args */
		{ $$ = p_SUBSTR; }
	;

/* operator definitions */

relational_operator :
	  NEQ
		{ $$ = NE_EXPR; }
	| LTE
		{ $$ = LE_EXPR; }
	| GTE
		{ $$ = GE_EXPR; }
	| '=' 
	| '<'
	| '>'
	;

multiplying_operator :
	  DIV
		{ $$ = TRUNC_DIV_EXPR; }
	| MOD
		{ $$ = FLOOR_MOD_EXPR; }
	| '/'
	| '*'
	;

adding_operator :
	  '-'
	| '+'
	;

semi :
	  ';'
		{ }
/* ARGH. This causes s/r conflicts. Check them!
	| error
		{ warning ("missing ';'\n"); }
 */
	;

optional_semicolon :
	  /* empty */
	| ';'
	;

optional_colon :
	  /* empty */
	| ':'
	;

lbracket :
	  '['
	| LBRACKET
	;

rbracket :
	  ']'
	| RBRACKET
	;

/* These are support states, not pascal syntax */

pushlevel :
	  /* empty */
		{   emit_line_note (input_filename, lineno);
		    pushlevel (0);
		    clear_last_expr ();
		    expand_start_bindings (0); 
		}
	;

/* Do not push_momentary() */
pushlevel1 :
	  /* empty */
		{   emit_line_note (input_filename, lineno);
		    pushlevel (0);
		    clear_last_expr ();
		    expand_start_bindings (0); 
		}
	;

/* @@@ Note that this does not return the value of poplevel().
   GCC uses that value to handle braced-group expressions, which
   are not implemented in GPC (yet :-)) */
poplevel :
	poplevel1
		{ 
		}
	;

poplevel1 :
	  /* empty */
		{ tree decls = getdecls ();
		  emit_line_note (input_filename, lineno);
		  if (decls != NULL_TREE) {
		    expand_end_bindings (decls, 1, 0);
		    poplevel (1, 1, 0);
		  } else {
		    expand_end_bindings (decls, kept_level_p (), 0);
		    poplevel (kept_level_p (), 0, 0);
		  }
		}
	;

setspecs :
	  /* empty */
		{ $$ = suspend_momentary (); }
	;

/*
  The EXTENDED PASCAL support below currently only parses things, so don't
  try to use e.g. modules for your programs. They DO NOT generate any
  code. The module support is according to an unapproved
  draft (Jan 29 1989) of Programming Language Extended Pascal

  You will propably crash the compiler if trying to use these rules.

			Mon Jul 31 15:21:27 1989

			jtv@hut.fi
 */

module_declaration :
	  module_name rest_of_module
	;

/* optional_module_type_directive may be one of
   INTERFACE
   IMPLEMENTATION

   If neither is present, default to INTERFACE
 */

module_name :
	  MODULE new_identifier optional_module_type_directive
	;

rest_of_module :
	  par_id_list semi module_interface optional_module_block  /* Interface */
		{ }
	| semi module_interface optional_module_block		   /* Interface */
	| semi module_block					   /* Implementation */
	;

module_interface :
	  interface_specification_part
	  optional_import_part
	  simple_declaration_part
	  END
	;

optional_module_block :
	  /* empty */
	| semi module_block
	;

module_block :
	  import_part
	  simple_declaration_part
	  optional_init_and_final_part
	  END
	;

optional_init_and_final_part :
	  TO BEGIN DO statement semi optional_finalization
	| optional_finalization
	;

/* Handle here also the case where both init & final parts are missing */
optional_finalization :
	  /* empty */
	| TO END DO statement semi
	;

/* Interface or implementation, if missing, default to interface */
optional_module_type_directive :
	  /* Empty */
		{ }
	| directive
		{ }
	;

interface_specification_part :
	  EXPORT export_part_list semi
	;

export_part_list :
	  export_part
	| export_part_list semi export_part
	;

export_part :
	  new_identifier '(' export_list ')'
		{ }
	;

export_list :
	  export_list_item
	| export_list ',' export_list_item
	;

export_list_item :
	  export_clause_or_range
	;

/* This is not quite correct, but done like this to avoid conflicts */
/* optional_protected should be allowed only before a variable_name */
export_clause_or_range :
	  optional_protected identifier rest_of_clause_or_range
		{ }
	;

/* The TWODOTS part should not allow RESTRICTED in front... */
rest_of_clause_or_range :
	  TWODOTS identifier
	| optional_renaming_part
	;

/* Here are the exact rules from export_list_item down:
   (The export_clause_or_range should check inside the
    action clause that it's not too general)

export_list_item :
	  export_clause
	| export_range
	;


export_clause :
	  exportable_name
	| export_renaming_clause
	;

exportable_name :
	  constant_name
	| type_name
	| schema_name
	| optional_protected variable_name
	| procedure_name
	| function_name
	;

export_renaming_clause :
	  exportable_name RENAME new_identifier
	;

export_range :
	  least TWODOTS greatest
	;

least :
	  constant_name
	;

greatest :
	  constant_name
	;

*/

optional_protected :
	  /* Empty */
	| PROTECTED
	;

optional_renaming_part :
	  /* Empty */
	| RENAME new_identifier
	;

optional_import_part :
	  /* Empty */
	| import_part
	;

import_part :
	  IMPORT import_specification_list semi
	;

import_specification_list :
	  import_specification
	| import_specification_list semi import_specification
	;	

import_specification :
	  identifier optional_access_qualifier optional_import_qualifier
		{ }
	;

optional_access_qualifier :
	  /* Empty */
	| QUALIFIED
	;

optional_import_qualifier :
	  /* Empty */
	| ONLY '(' import_list ')'
	| '(' import_list ')'
	;
	  
import_list :
	  import_clause
		{ }
	| import_list ',' import_clause
		{ }
	;

import_clause :
	  identifier optional_renaming_part
		{ }
	;

%%
/* Lexical analyzer moved to gpc-lex.c */


/* Return a newly-allocated string whose contents concatenate
   those of s1, s2, s3.  */

static char *
concat (s1, s2, s3)
     char *s1, *s2, *s3;
{
  int len1 = strlen (s1), len2 = strlen (s2), len3 = strlen (s3);
  char *result = (char *) xmalloc (len1 + len2 + len3 + 1);

  strcpy (result, s1);
  strcpy (result + len1, s2);
  strcpy (result + len1 + len2, s3);
  *(result + len1 + len2 + len3) = 0;

  return result;
}


/* Sets the value of the 'yydebug' varable to VALUE.
   This is a function so we don't have to have YYDEBUG defined
   in order to build the compiler.  */
void
set_yydebug (value)
     int value;
{
#if YYDEBUG != 0
  yydebug = value;
#else
  warning ("YYDEBUG not defined.");
#endif
}

/*
Local variables:
mode:c
tab-width: 8
version-control: t
End:
*/
