/* Misc routines & definitions for the GPC 
   Copyright (C) 1987, 1988, 1989 Free Software Foundation, Inc.

This file is part of GNU GCC.

GNU GCC is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 1, or (at your option)
any later version.

GNU GCC is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with GNU GCC; see the file COPYING.  If not, write to
the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.  */

/* TODO:
 *  implement all extended pascal libcalls
 *
 *
 */

/*
 * Author: Jukka Virtanen <jtv@hut.fi>
 */
#include "config.h"
#include <stdio.h>

#include "gvarargs.h"

#include "tree.h"
#include "c-tree.h"
#include "input.h"
#include "rtl.h"
#include "obstack.h"
#include "flags.h"

#include "gpc-parse.h"
#include "gpc-defs.h"
#include "rts/rts_types.h"

/* Testing emit_library_call */
#define EMIT_LIBCALL

extern char *language_string;

/* This is not used. It's required to compile gpc-decl.c */
tree ridpointers[1];

static void init_any ();

extern tree initialize_rts_set_parameter (); /* Forward declaration */

/* Option -fsetlimit:NNNNN changes the upper bound of "unbound"
 * integer type sets (default upper bound = DEFAULT_SET_SIZE)
 */

/* When reading options we can't make tree nodes yet! */
int  requested_set_size = -1;
tree integer_set_size   = NULL_TREE;

/* Default set type for unbounded integer sets, size is
 * integer_set_size
 */
tree integer_set_type_node;

/* Size of the file control block in bits */
/* This is referenced by stor-layout.c */
tree size_of_file_type;

/* BOOLEAN_TYPE node */
tree boolean_type_node;

/* TEXT and FILE type nodes */
/* TEXT is a predefined file type in ISO Pascal */
tree text_type_node;

/* COMPLEX type node */
tree complex_type_node;

/* pascal nodes */
tree boolean_false_node;
tree boolean_true_node;
tree integer_maxint_node;
tree input_file_node;
tree output_file_node;
tree external_file_name_list;

tree real_zero_node;

/* index type for character arrays (strings) */
tree char_array_index_type_node;

/* Extended Pascal nodes */
tree complex_zero_node;
tree real_max_node;
tree real_min_node;
tree real_eps_node;
tree char_max_node;

/*
 * these *_id's are only used to tell
 * grokdeclarator what we are currently defining.
 */
tree const_id;
tree type_id;
tree inline_id;
tree varparm_id;
tree volatile_id;

/*
 * Generate (temporary) unique identifier_nodes of the form "<%s_%d>"
 */
tree
get_unique_identifier (lead)
char *lead;
{
  char idbuf [ 256 ];
  static int idcount = 0;
  
  sprintf (idbuf, "<%s_%d>", lead, idcount++);
  return get_identifier (idbuf);
}

tree
error_level (str)
char *str;
{
    error ("Level 0 ISO Pascal does not implement `%s'", str);
    return error_mark_node;
}

static tree
get_empty_set_node ()
{
  
  /* This node represents the empty set; the decl is created
   * when required and referenced only in this file.
   * The range of the domain is special.
   */
  static tree empty_set_node = NULL_TREE;

  /* This is the node representing the empty set */
  /* @@@ This is stupid. Get rid of the variable, declare only
     a type and use that. When so, be sure to change calls to set-library.
   */
  if (! empty_set_node)
    {
      empty_set_node =
	start_decl
	  (get_identifier ("empty_set"),
	   build_tree_list
	   (NULL_TREE, 
	    build_set_type
	    (build_tree_list (void_type_node,
			      build_index_2_type (size_zero_node,
						  build_int_2 (-1, 0))))),
	   0);

      finish_decl (empty_set_node, NULL_TREE, NULL_TREE);
    }
  return empty_set_node;
}

/* Clear the target set */
clear_set (the_set)
     tree the_set;
{
  tree type  = TREE_TYPE (the_set);
  rtx target = expand_expr (the_set,
			    NULL_TREE,
			    TYPE_MODE (type),
			    0);
  clear_storage (target, int_size_in_bytes (type));
}

/*
  Construct one set member from the LOWER and UPPER as range bounds.
  If UPPER is NULL_TREE, then the member is not a closed interval,
  but rather a single member.

  Returns a TREE_LIST node with TREE_PURPOSE as the lower bound and
  TREE_VALUE as the upper bound of each set member.

 */

tree
construct_set_member (lower, upper)
tree lower, upper;
{
  tree t1;

  if (lower == error_mark_node || upper == error_mark_node)
    return build_tree_list (NULL_TREE, NULL_TREE);

  t1 = TREE_TYPE (lower);

  if (! ORDINAL_TYPE (TREE_CODE (t1)))
    {
      error ("Set constructor elements must be of ordinal type");
      return build_tree_list (NULL_TREE, NULL_TREE);
    }
    
  if (upper == NULL_TREE)
    return build_tree_list (lower, upper);
  else
    {
      if (TREE_CODE (t1) != TREE_CODE (TREE_TYPE (upper)))
	{
	  error ("set range upper and lower bound are of incompatible type");
	  return build_tree_list (NULL_TREE, NULL_TREE);
	}
      if (TREE_CODE (lower) == INTEGER_CST &&
	  TREE_CODE (upper) == INTEGER_CST &&
	  tree_int_cst_lt (upper, lower))
	{
	  warning ("specified set member range is empty");
	  return NULL_TREE;
	}
    }
  return build_tree_list (lower, upper);
}

/*
 * Build a constructor node for set elements.
 * This is later converted to a set with construct_set ()
 *
 */
tree
build_set_constructor (members)
tree members;
{
  tree node;
  tree m;
  int konstant = 1;

  if (members == error_mark_node)
    return error_mark_node;

  for (m = members; m ; m = TREE_CHAIN(m))
    {
      if (!TREE_PURPOSE (m) && !TREE_VALUE (m))
	return error_mark_node;			/* Error in elements */
      else if (! TREE_CONSTANT (TREE_PURPOSE (m)) ||
	       (TREE_VALUE (m) && ! TREE_CONSTANT (TREE_VALUE (m))))
	konstant = 0;
    }
  
  node = make_node (CONSTRUCTOR);

  /* Mark it as a SET_TYPE constructor */
  /* The correct type is not known yet, so this is as good as any other */
  TREE_TYPE (node) = build_set_type
    			(build_tree_list
			 	(void_type_node,
				 build_index_2_type (size_zero_node,
						     build_int_2 (-1, 0))));
  CONSTRUCTOR_ELTS (node) = members;
  TREE_CONSTANT (node)    = konstant;
  return node;
}

/* Build a SET_TYPE node from a set constructor.

   CONSTRUCTOR is a CONSTRUCTOR type node whose CONSTRUCTOR_ELTS
   contains a TREE_LIST of the elements(/pairs) of the set.

   If ARG_TYPE == 0, TARGET_OR_TYPE is a VAR_DECL node where we should
   	construct our set.
   If ARG_TYPE == 1, TARGET_OR_TYPE is a SET_TYPE node which we should
   	model our new set after.
   If ARG_TYPE == 2, TARGET_OR_TYPE is a SET_TYPE node passed as a
        parameter to a function. (Special case for empty set constructor
	passing)

   TARGET_OR_TYPE is NULL_TREE if we don't know the destination
   where we should put the constructed set, nor the type we should
   be cloning to our constructed set.

   Return NULL_TREE if we assigned the set to the existing
   TARGET_SET, else return the node that contains the constructed
   set.

   @@@ Rewrite (something like this):
       For constant set members, build the set as a bit vector
       in here and store it to the set as a constant. After that,
       call r_SET_RANGE only on the variable parts. This way we
       can mark the some sets constant and optimize.
 */

tree
construct_set (constructor, target_or_type, arg_type)
tree constructor;
tree target_or_type;
int  arg_type;
{
  tree elements;
  tree result = NULL_TREE;
  tree set_low, set_high;
  tree type;
  rtx target;

  if (constructor == error_mark_node)
    return error_mark_node;

  elements = CONSTRUCTOR_ELTS (constructor);
  
  if (! elements && arg_type != 2)
    if (arg_type == 0 && target_or_type) /* Clear the storage in the target */
      {
	clear_set (target_or_type);
      }
    else
      result = get_empty_set_node ();
  else
    {
      tree this_set_type;
      tree cset;	/* target set built of the constructor elements */

      if (arg_type == 0 && target_or_type)
	{
	  cset = target_or_type;
	  this_set_type = TREE_TYPE (target_or_type);
	  set_low  = TYPE_MIN_VALUE (this_set_type);
	  set_high = TYPE_MAX_VALUE (this_set_type);
	}
      else
	{
	  int  set_is_ready = 0;
	  tree type;
	  enum tree_code tcode;

	  if (elements)
	    type = TREE_TYPE (TREE_PURPOSE (elements));
	  else if (! target_or_type)
	    /* Empty set constructor to unknown set? */
	    abort ();
	  else
	    type = TREE_TYPE (target_or_type);

	  tcode = TREE_CODE (type);

	  if (target_or_type)
	    {
	      this_set_type = target_or_type;
	      set_low  = TYPE_MIN_VALUE (this_set_type);
	      set_high = TYPE_MAX_VALUE (this_set_type);
	      set_is_ready = 1;
	    }
	  else if (tcode == INTEGER_TYPE && ! TREE_TYPE (type))
	    {
	      /* avoids [ -maxint .. maxint ] storage request :-) */
	      tree min = NULL_TREE;
	      tree max = NULL_TREE;
	      tree elem;
	      int varies = 0;

	      /* Scan for the min and max values */
	      for (elem = elements; elem; elem = TREE_CHAIN (elem))
		{
		  tree min_c = TREE_PURPOSE (elem);
		  tree max_c = TREE_VALUE (elem);

		  /* If this is a single element, not a range */
		  if (! max_c)
		    max_c = min_c;

		  if (TREE_CODE (min_c) != INTEGER_CST)
		    varies++;
		  else
		    if (!min || tree_int_cst_lt (min_c, min))
		      min = min_c;

		  if (TREE_CODE (max_c) != INTEGER_CST)
		    varies++;
		  else
		    if (!max || tree_int_cst_lt (max, max_c))
		      max = max_c;
		}
	      
	      /* @@@ I do not yet check if the type of the variable part
		 is a subrange whose allowed ranges would
		 change the calculated set bounds, or prevent
		 the warnings since it could not be outside
		 bounds.
	       */
	      
	      if (!varies)  /* Constant low bound */
		{
		  set_low  = min;
		  set_high = max;
		  
		  /* If constant bounds, but too big */
		  if (max && tree_int_cst_lt
		      	       (integer_set_size,
		       		fold
				  (build
				     (PLUS_EXPR, integer_type_node,
				      integer_one_node,
				      fold
				        (build (MINUS_EXPR,
						integer_type_node,
						max, min))))))
		    max = NULL_TREE;

		  if (!max)
		    {
		      set_high = fold (build (PLUS_EXPR, integer_type_node,
					      min, integer_set_size));
		      warning ("constructing limited integer set `%d..%d'",
			       TREE_INT_CST_LOW (set_low),
			       TREE_INT_CST_LOW (set_high));
		    }
		}
	      else
		{
		  this_set_type = integer_set_type_node;
		  
		  set_low  = TYPE_MIN_VALUE (TYPE_DOMAIN (this_set_type));
		  set_high = TYPE_MAX_VALUE (TYPE_DOMAIN (this_set_type));
		  set_is_ready = 1;
		  warning ("constructing limited integer set `%d..%d'",
			   TREE_INT_CST_LOW (set_low),
			   TREE_INT_CST_LOW (set_high));
		}
	      
	      if (varies)
		warning ("The limited integer set contains variable elements");
	    }
	  else
	    {
	      if (tcode == INTEGER_TYPE) /* subrange type */
	        {
		  set_low  = TYPE_MIN_VALUE (TREE_TYPE (type));
		  set_high = TYPE_MAX_VALUE (TREE_TYPE (type));
		}
	      else
		{
		  set_low  = TYPE_MIN_VALUE (type);
		  set_high = TYPE_MAX_VALUE (type);
		}
	    }

	  if (! set_is_ready)
	    this_set_type =
	      build_set_type
		(build_tree_list
		 (type, build_index_2_type (set_low, set_high)));
	  
	  /* Declare a temporary variable large enough to hold the set */
	  cset = start_decl (get_unique_identifier ("set_constructor"),
			     build_tree_list(NULL_TREE, this_set_type),
			     0);
	  finish_decl (cset, NULL_TREE, NULL_TREE);

	  if (int_size_in_bytes (type) == -1)
	    abort ();   /* non-constant bounds not yet supported (Why???) */
	  else
	    init_any (cset, 0);

	  /* We constructed a temporary set to hold the constructor elems. */
	  result = cset;
	}

      /* Clear the storage in the cset */
      clear_set (cset);

      {
	/* Add the set members to the set; checking that each member
	 * is of the same base type
	 */

	int testit = 1;
	tree m;
	tree basetype = set_base_type (this_set_type);
	enum tree_code set_type_code = TREE_CODE (basetype);

	/* Do no conversion before pushing args.
	 *  -> must inform the run time system how long
	 *     the set elements are
	 *     (all of them are same size than the set base type)
	 */
	tree element_size = size_in_bytes (basetype);
	
	/* Scan for single elements in constructor */
	tree single_list = NULL_TREE;
	tree range_list  = NULL_TREE;

	for (m = elements; m ; m = TREE_CHAIN(m))
	  {
	    tree p = TREE_PURPOSE (m);
	    tree v = TREE_VALUE (m);
	    enum tree_code pcode = TREE_CODE (TREE_TYPE (p));
	    enum tree_code vcode;
	    /* @@@ Do we need the pcode, vcode to int conversion ??? */

	    /* @@@@ Generate checks to see if nodes are in set */
	    
	    /* Add the element to the set. */
	    if (v == NULL_TREE) {
	      /* @@@ Allows too much! */
	      if (pcode != set_type_code &&
		  (pcode != INTEGER_TYPE ||		/* Subrange */
		   TREE_CODE (TREE_TYPE (TREE_TYPE (p))) != set_type_code))
		error
		  ("Type of set constructor element must be the set basetype");

/* @@@@@@@@@@@@@@@@ Hmm, test without conversion */
if (testit)
	      single_list =
		chainon
		  (single_list,
		   build_tree_list (NULL_TREE,
				    /* Converting crashes emit_move_insn */
				    p));
else
	      single_list =
		chainon
		  (single_list,
		   build_tree_list (NULL_TREE,
				    pcode == INTEGER_TYPE ?
				    p : convert (integer_type_node, p)));
	    } else {
	      vcode = TREE_CODE (TREE_TYPE (v));
	      
	      if (pcode != set_type_code || vcode != set_type_code)
		error
		  ("Type of set constructor range bounds must be the set basetype");
	      
/* @@@@@@@@@@@@@@@@ Hmm, test without conversion */
if (testit)
	      range_list =
		chainon
		  (chainon (range_list,
			    build_tree_list (NULL_TREE,
					 /* Converting crashes emit_move_insn */
					     p)),
		   build_tree_list (NULL_TREE, v));
else
	      range_list =
		chainon
		  (chainon (range_list,
			    build_tree_list (NULL_TREE,
					     pcode == INTEGER_TYPE ?
					     p : convert(integer_type_node, p))),
		   build_tree_list (NULL_TREE,
				    vcode == INTEGER_TYPE ?
				    v : convert (integer_type_node, v)));
	    }
	  }
	
	if (single_list)
	  {
	    /* Prepend element size and number of singles to the front of
	     * the list
	     */
	    single_list =
	      tree_cons (NULL_TREE,
			 element_size,
			 tree_cons (NULL_TREE,
				    build_int_2 (list_length (single_list), 0),
				    single_list));


	    build_rts_call (r_SET_SINGLE,		/* Add singles */
			    tree_cons (NULL_TREE,
				       initialize_rts_set_parameter (cset, 1),
				       single_list));
	  }

	if (range_list)
	  {
	    int range_length = list_length (range_list);

	    if (range_length & 1)
	      abort ();

	    /* Prepend element size and  number of ranges to the front of
	     * the list
	     */
	    range_list =
	      tree_cons (NULL_TREE,
			 element_size,
			 tree_cons (NULL_TREE,
				    build_int_2 (range_length/2, 0),
				    range_list));

	    build_rts_call (r_SET_RANGE,		/* ranges */
			    tree_cons (NULL_TREE,
				       initialize_rts_set_parameter (cset, 1),
				       range_list));
	  }
      }
    }

  /* Return NULL_TREE if we assigned the set to target_set.
     Otherwise return the set type variable the constructor
     is now stored in or empty_set_node if that was specified.
   */

  return result;
}

/* Construct, lay out and return the set of type with the number of
   elements specified by the range of values of MEMBERS.

   MEMBERS is a TREE_LIST whose TREE_PURPOSE is the type of the
   set and TREE_VALUE is the range of values is has.

   The SET_TYPE has the same fields as the ARRAY_TYPE.

   The TREE_TYPE of the SET_TYPE is the type of the set members.
   (If that is subrange, then TREE_TYPE(TREE_TYPE(type)) is the
    base type of the set)
 */

tree
build_set_type (members)
tree members;
{
  tree t = make_node (SET_TYPE);

  tree d     = TREE_VALUE (members);
  tree dlow  = TYPE_MIN_VALUE (d);
  tree dhigh = TYPE_MAX_VALUE (d); 

  TREE_TYPE (t)   = TREE_PURPOSE (members);
  TYPE_DOMAIN (t) = d;

  /* @@ hash set types... */

  if (TYPE_SIZE (t) == 0)
    {
      /* Check if it is a nonlimited integer type, like "set of integer" */
      if (TREE_TYPE (t) == integer_type_node &&
	  ! TREE_TYPE (TREE_TYPE (t))        && /* Allows huge subrange sets */
	  TREE_CODE (dlow)  == INTEGER_CST   &&
	  TREE_CODE (dhigh) == INTEGER_CST   &&
	  TREE_INT_CST_LOW (integer_set_size)   <
	    TREE_INT_CST_LOW (dhigh) - TREE_INT_CST_LOW (dlow) + 1)
	{
	  warning ("Integer set size limited to `%d' elements from low bound",
		   TREE_INT_CST_LOW (integer_set_size));
	  warning ("Use -fsetlimit:NUMBER to change it at compile time");
	  TYPE_DOMAIN (t) = 
	    build_index_2_type
	      (dlow,
	       build_int_2 (TREE_INT_CST_LOW (dlow) +
			    TREE_INT_CST_LOW (integer_set_size) - 1, 0));
	}
      layout_type (t);
    }
  return t;
}

/*
 * Returns the base type of the SET_TYPE TYPE
 */
tree
set_base_type (type)
tree type;
{
  tree basetype;

  if (TREE_CODE (type) != SET_TYPE)
    abort ();

  basetype = TREE_TYPE (type);

  /* Check for ordinal subranges */
  if (TREE_CODE (basetype) == INTEGER_TYPE
      && TREE_TYPE (basetype)
         /* If it is an integer subrange return the range */
      && TREE_CODE (TREE_TYPE (basetype)) != INTEGER_TYPE)
    basetype = TREE_TYPE (basetype);

  return basetype;
}

tree
build_file_type (filetype)
tree filetype;
{
    tree File = make_node (FILE_TYPE);

    /* This file is not a TEXT type file */
    TYPE_FILE_TEXT (File) = integer_zero_node;

    TREE_TYPE (File) = filetype;

    layout_type (File);
    if (!TYPE_SIZE (File))
      TYPE_SIZE (File) = size_of_file_type;

    return File;
}

/* information in variant field into tree list node:
 *
 * VARSEL:
 *  TREE_LIST(PURPOSE (PURPOSE(selector_id), VALUE(selector_type)),
 *	      VALUE   (PURPOSE(file),        VALUE(line)))
 *            
 *
 * FIELDLIST:
 *  TREE_LIST(PURPOSE   (case_list),
 *	      VALUE	(field_decl_node))
 */
/* @@@@ Extended Pascal: Implement ranges as case selectors */
tree
build_record_variant_part (file, line, varsel, fieldlist)
char *file;
int line;
tree varsel, fieldlist;
{
    tree ret, type;
    tree link, fields = NULL_TREE;

    /* @@@@ check here that the case labels are of type
       TREE_PURPOSE (varsel) */

    for (link = fieldlist; link; link = TREE_CHAIN (link))
	fields = chainon (fields, TREE_VALUE (link));

    type = finish_struct (start_struct (UNION_TYPE, NULL_TREE), fields);

    ret = grokfield (file, line,
		     NULL_TREE, /* field name */
		     build_tree_list (NULL_TREE, type),
		     NULL_TREE); /* field width */
    return ret;
}

/*
 * Set the packed flag of the type.
 * @@@@ This flag should also be checked...
 */
tree
grok_packed (type)
tree type;
{
  PASCAL_TYPE_PACKED (type) = 1;
  return type;
}

/* Declare variables.
 *
 * @@ In Extended Pascal the initial value is associated with the TYPE like:
 * var i,j,k,l : integer value 0;
 * (not implemented yet.)
 */
void
declare_vars (var_list, type, initial_value)
tree  var_list, type, initial_value;
{
    tree v;
    tree d;
    tree tt = build_tree_list (NULL_TREE, type);
    
    for(v = var_list; v; v = TREE_CHAIN(v)) {
	d = start_decl (TREE_VALUE (v), tt, initial_value != NULL_TREE);
	finish_decl (d, initial_value, NULL_TREE);
    }
}

static tree
pascal_conformant_type (type)
     tree type;
{
  extern int flag_gen_aux_info; /* @@@@ KLUDGE ALERT! */
  tree ctype;
  int old_flag = flag_gen_aux_info;

  flag_gen_aux_info = 1;
  
  /* Create a brand new type node we can flag as conformant.
   * The type node is identical in other respects than the old type
   */
  ctype = build_type_variant (type, TYPE_READONLY(type), TYPE_VOLATILE(type));

  PASCAL_TYPE_CONFORMANT_INDEX (ctype) = 1;
  
  flag_gen_aux_info = old_flag;
  return ctype;
}

/*
 * Each conformant array argument is passed as follows:
 *
 * The array and it's conformant bounds are in the following
 * argument order:
 *
 *    1   2      n-1  n  n+1     n+2     n+3
 *   LO1 HI1 ... LOn HIn CARRAY0 CARRAY1 CARRAY2
 *
 * Where the indices always appear paired:
 * LO1 HI1  - Top level indices of CARRAYs
 * ...
 * LOn HIn  - lowest level indices of CARRAYs
 *
 * CARRAYx  - the n/2-dimension arrays.
 *
 * All of the indice's types are flagged with
 * PASCAL_TYPE_CONFORMANT_INDEX. 
 *
 * The first thing after indices is the conformant
 * array type with variable sized bounds.
 *
 * If an ARRAY_TYPE is not preceedeed by a list of indices,
 * its indices are the same as the immediately preceeding
 * conformant array's. (If some of this is not true, abort())
 */
void
handle_formal_conf_array_param_list (idlist, type, varparm)
tree idlist;
tree type;
int varparm;
{
  tree link;
  int  packed     = TREE_PURPOSE (type) != NULL_TREE;
  tree index_list = TREE_PURPOSE (TREE_VALUE (type));
  tree carray     = TREE_VALUE (TREE_VALUE (type));
  tree parm;

  /*
   * Build array with unspecified index. The type is NOT laid
   * out. These are handled specially when actual parameters
   * are passed to this kind of formal in convert_arguments.
   *
   * The TREE_LIST node of the indices part of the conformant
   * array schema is flagged; it tells grokdeclarator to
   * take special care of it.
   */
  for (link = index_list; link; link = TREE_CHAIN (link))
    {
      /* Build a new type-node so that we can safely flag it */
      tree ctype = pascal_conformant_type (TREE_VALUE (link));
      tree lo = TREE_PURPOSE (TREE_PURPOSE (link));
      tree hi = TREE_VALUE   (TREE_PURPOSE (link));
      tree index_type = build_tree_list (NULL_TREE, ctype);
      tree itype;

      /* Push lower bound */
      parm = build_tree_list (index_type, lo);
      push_parm_decl (parm);
      
      /* Push upper bound */
      parm = build_tree_list (index_type, hi);
      push_parm_decl (parm);

      {
	tree lodecl = lookup_name (lo);
	tree hidecl = lookup_name (hi);

	if (! lodecl || ! hidecl ||
	    TREE_CODE (lodecl) != PARM_DECL ||
	    TREE_CODE (hidecl) != PARM_DECL)
	  abort ();

	itype = build_index_2_type (lodecl, hidecl);
      }

      carray = build_array_type (carray, itype);
      C_TYPE_VARIABLE_SIZE (carray) = 1;
    }

  if (packed)
    grok_packed (carray);

  if (varparm)
    {
      carray = build_reference_type (carray);
      C_TYPE_VARIABLE_SIZE (carray) = 1;
    }

  /* add conformant array parameters */
  for (link = idlist; link; link = TREE_CHAIN(link))
    {
      parm = build_tree_list (build_tree_list (NULL_TREE, carray),
			      TREE_VALUE (link));
      push_parm_decl (parm);
    }
}

void
handle_formal_param_list (idlist, type, varparm)
tree idlist;
tree type;
int varparm;
{
    register tree link;

    if (TREE_CODE (idlist) == CALL_EXPR) {
	tree parm;
	/* idlist == CALL_EXPR
	 * type   == TREE_LIST (NULL_TREE, type_node)
	 */
	parm  = build_tree_list (type, idlist);

	push_parm_decl (parm);
    } else {
	if (varparm)
	    type = build_reference_type (type);

	type = build_tree_list (NULL_TREE, type);

	for (link = idlist; link; link = TREE_CHAIN(link)) {
	    tree parm = build_tree_list (type, TREE_VALUE (link));
	    push_parm_decl (parm);
	}
    }
}

tree
p_grokfields (list, type)
tree list, type;
{
    tree link, ret = NULL_TREE;

    type = build_tree_list (NULL_TREE, type);

    for (link = list; link; link = TREE_CHAIN (link))
	ret = chainon(ret, grokfield ("", 0, 
				      TREE_VALUE(link), type, NULL_TREE));

    return ret;
}

/* no params to function */
tree ptype_void;

/* single param rts calls */
tree ptype_int;
tree ptype_char;
tree ptype_bool;
tree ptype_double;

/* any number of any parms */
tree ptype_vararg;

/* Some rts function formal parameters are constructed
   by this routine. */
tree
do_ptype(type, var, last)
tree type;
int  var;	/* 0 : use type; 1 : build_reference; 2 : build_pointer */
int  last;
{
    tree temp;

    if (var == 1)
	type = build_reference_type (type);
    else if (var == 2)
	type = build_pointer_type (type);

    temp = build_tree_list (NULL_TREE, type);

    /* True if this is the last parameter to the function */
    if (last)
	temp = chainon (temp, ptype_void);

    return temp;
}

/* Declares some useful rts call types */
void
declare_rts_types ()
{
    ptype_void   = build_tree_list (NULL_TREE, void_type_node);
    ptype_vararg = NULL_TREE;

    ptype_int    = do_ptype (integer_type_node, 0, 1);
    ptype_char   = do_ptype (char_type_node, 0, 1);

    ptype_bool   = do_ptype (boolean_type_node, 0, 1);
    ptype_double = do_ptype (double_type_node, 0, 1);
}

/* inline functions are not here */
static struct
  {
      int  val;
      char *name;
  } rts[] = {
    { r_WRITE,	"_p_write"	},
    { r_READ,	"_p_read"	},
    { r_INITFDR,"_p_initfdr"	},

    /* These are NOT used if inline set functions are compiled */
						/* returns */
    { r_SET_IN,		"_p_set_in"	},	/* boolean */
    { r_SET_COPY,	"_p_set_copy"	},	/* *set */
    { r_SET_RANGE,	"_p_set_range"	},	/* void */
    { r_SET_SINGLE,	"_p_set_single"	},	/* void */
    { r_SET_UNION,	"_p_set_union"	},	/* *set */
    { r_SET_DIFF,	"_p_set_diff"	},	/* *set */
    { r_SET_INTERSECTION,"_p_set_intersection"},/* *set */
    { r_SET_EQUAL,	"_p_set_equal"	},	/* boolean */	
    { r_SET_LE,		"_p_set_le"	},	/* boolean */
    { r_SET_SYMDIFF,	"_p_set_symdiff"},	/* *set (Extended Pascal) */
    { r_SET_CARD,	"_p_set_card"	},	/* int  (Extended Pascal) */

    { p_ARCTAN, "_p_arctan"	},
    { p_COS,	"_p_cos" 	},
    { p_DISPOSE,"_p_dispose" 	},
    { p_EXP,	"_p_exp" 	},
    { p_EOF,	"_p_eof" 	},
    { p_EOLN,	"_p_eoln" 	},
    { p_GET,	"_p_get" 	},
    { p_LN,	"_p_ln" 	},
    { p_NEW,	"_p_new" 	},
    { p_PAGE,	"_p_page" 	},
    { p_PUT,	"_p_put" 	},
    { p_RESET,	"_p_reset" 	},
    { p_REWRITE,"_p_rewrite" 	},
    { p_SIN,	"_p_sin" 	},
    { p_SQRT,	"_p_sqrt" 	},
    { p_MARK,	"_p_mark" 	},
    { p_RELEASE,"_p_release" 	},
    { p_CLOSE,	"_p_close"	},

    /* Extended Pascal libcalls */

    { p_GETTIMESTAMP, "_p_gettimestamp", },
    { p_DATE,   "_p_date"	},
    { p_TIME,   "_p_time"	},
    { p_HALT,   "_p_halt"	},
    { p_EMPTY,  "_p_empty"	},
    { p_EXTEND, "_p_extend"	},
    { p_UPDATE, "_p_update"     },
    { p_POSITION,  "_p_position"},
    { p_LASTPOSITION,"_p_lastposition" },
    { p_SEEKWRITE, "_p_seekwrite" },
    { p_SEEKREAD,  "_p_seekread"  },
    { p_SEEKUPDATE,"_p_seekupdate"},
    /* GPC extensions */
    { p_DEFINESIZE,"_p_definesize" },
    { 0,        NULL		}
};

/*
   RTS_ID : What routine we are calling
   VALUE_TYPE : return value from calling NAME
   FORMAL_PARAMS : List of formal parameters; NULL_TREE means varargs.
   PARAMS is a list--a chain of TREE_LIST nodes--in which the
          TREE_VALUE of each node is a parameter-expression.
   SET_REF is nonzero if we should do a build_indirect_ref
   	   to the value before returning it
*/

static tree
rts_call (rts_id, value_type, formal_params, params, set_ref)
int rts_id;
tree value_type;
tree formal_params;
tree params;
int  set_ref;
{
    register tree fun;
    register int i;
    register int value;
    tree retval;
    char *name;
    
    for (i = 0; (value = rts[i].val) != rts_id; i++)
      ASSERT (value, "no such run time system routine");
    name = rts[ i ].name;

    fun = build_decl (FUNCTION_DECL,
		      get_identifier(name),
		      build_function_type (value_type, formal_params));

    DECL_EXTERNAL (fun) = 1;
    TREE_PUBLIC (fun) = 1;

    DECL_SOURCE_FILE (fun) = "Pascal run time library";
    DECL_SOURCE_LINE (fun) = 1;

    rest_of_decl_compilation (fun, 0, 0, 0);

    retval = build_function_call (fun, params);

    if (set_ref)
      retval = build_indirect_ref (retval, "rts_call deref");

    return retval;
}

tree rts_set_type = NULL_TREE;

/* I am tired to hack this for time being, sorry. */
tree horrible_set_kludge_type = NULL_TREE;

void
create_rts_set_type ()
{
  tree fields;
  tree set_type;
  fields = chainon
    	     (p_grokfields (build_tree_list (NULL_TREE,
					     get_identifier ("Set")),
			    ptr_type_node),
	      p_grokfields
	        (chainon
		   (build_tree_list (NULL_TREE,
				     get_identifier ("lo")),
		    build_tree_list (NULL_TREE,
				     get_identifier ("hi"))),
		 integer_type_node));
  
  set_type = finish_struct (start_struct (RECORD_TYPE, NULL_TREE),
			    fields); 
  
  TYPE_TAG_FIELD (set_type) = NULL_TREE;
  
  rts_set_type = set_type;
}

/*
 * This routine creates and initializes structures for passing
 * set type nodes to the runtime system.
 *
 * If we are on global_binding_level, we re-use the global storage,
 * as the structures need to exist only to the next rts_call.
 *
 * Since there are max 3 set operands to the rts set routines,
 * we need 3 variables on the global_binding_level
 *
 */

#define RTS_SET_PARMS 3

tree
initialize_rts_set_parameter (op, which)
tree op;
int which;
{
  /* Generate code to fill the runtime set parameter
     struct {
     int *Set;
     int lo;
     int hi;
     };
   */
  
  int global = global_bindings_p ();
  tree the_set = NULL_TREE;
  static tree global_set[ RTS_SET_PARMS ] =
    { NULL_TREE, NULL_TREE, NULL_TREE };	/* well, yes */
  
  if (TREE_CODE (TREE_TYPE (op)) != SET_TYPE)
    abort ();

  if (global)
    {
      if (which < 0 || which > (RTS_SET_PARMS - 1))
	abort ();
      
      the_set = global_set[ which ];
    }
  
  if (! the_set)
    {
      the_set = start_decl (get_unique_identifier ("_rts_set_parameter"),
			    build_tree_list (NULL_TREE, rts_set_type),
			    0);
      finish_decl (the_set, NULL_TREE, NULL_TREE);
      
      if (global)
	global_set[ which ] = the_set;
    }
  
  {
    /*
     * Initialize the fields at runtime
     */
    tree target;
    tree source;
    tree type = TREE_TYPE (op);

    target = build_component_ref (the_set, get_identifier ("Set"));

    if (set_empty_p (type))
      emit_move_insn (expand_expr (target, NULL_RTX, VOIDmode, 0),
		      const0_rtx);
    else
      {
	source = build_unary_op (ADDR_EXPR, op, 1);
	expand_expr_stmt
	  (build_modify_expr (target, NOP_EXPR, source));
      }

    target = build_component_ref (the_set, get_identifier ("lo"));
    source = TYPE_MIN_VALUE (TYPE_DOMAIN (type));
    expand_expr_stmt
      (build_modify_expr (target, NOP_EXPR, source));

    target = build_component_ref (the_set, get_identifier ("hi"));
    source = TYPE_MAX_VALUE (TYPE_DOMAIN (type));
    expand_expr_stmt
      (build_modify_expr (target, NOP_EXPR, source));
  }
  return build_unary_op (ADDR_EXPR, the_set, 0);
}

/*
 * This routine is called from expr.c to generate library calls to
 * handle all set operations. All of these should really be inlined.
 * CODE is the operation code, OP1 and OP2 are arguments, RESULT
 * is an out parameter that contains the result of the operation.
 *
 * Return codes:
 *  -1		failure
 *   0 		success
 *   1		no-operation (result does not need to be moved to destination)
 */
int
rts_set (code, op1, op2, result)
enum tree_code code;
tree op1, op2;
tree *result;
{
  int retval = 0;
  tree store;

  int  match1 = 1;	/* If returning unmodified nodes passed to us */
  int  match2 = 1;	/*  as parameters, we use these */

  tree type1;
  enum tree_code opcode1  = TREE_CODE (op1);

  tree type2;
  enum tree_code opcode2  = VOID_TYPE;

  tree result_type = NULL_TREE;

  tree setA, setB, setR = NULL_TREE;

  if (TREE_CODE (TREE_TYPE (op1)) != SET_TYPE)
    abort();

  /* We need a temporary variable for the runtime VAR parameter (sick) */
  if (code != IN_EXPR &&
      opcode1 != VAR_DECL && opcode1 != RESULT_DECL && opcode1 != PARM_DECL)
    {
      tree par1 = start_decl (get_unique_identifier ("_set_parameter"),
			      build_tree_list (NULL_TREE, TREE_TYPE (op1)),
			 0);
      finish_decl (par1, NULL_TREE, NULL_TREE);

      /* assign the set to the temporary */
      expand_assignment (par1, op1, 0, 0);

      op1 = par1;
      match1 = 0;
    }

  type1    = TREE_TYPE (op1);

  if (op2)
    {
      opcode2 = TREE_CODE (op2);    

      /* We need a temporary variable for the runtime VAR parameter (sick) */
      if (code != MODIFY_EXPR && code != CARD_EXPR &&
	  opcode2 != VAR_DECL && opcode2 != RESULT_DECL &&
	  opcode2 != PARM_DECL)
	{
	  tree par2 = start_decl (get_unique_identifier ("_set_parameter"),
				  build_tree_list (NULL_TREE, TREE_TYPE (op2)),
				  0);
	  finish_decl (par2, NULL_TREE, NULL_TREE);
      
	  /* assign the set to the temporary */
	  expand_assignment (par2, op2, 0, 0);

	  op2 = par2;
	  match2 = 0;
	}

      type2 = TREE_TYPE (op2);
    }

  if (code == BIT_AND_EXPR || code == BIT_ANDTC_EXPR ||
      code == BIT_IOR_EXPR || code == BIT_XOR_EXPR)
    {
      tree dom1 = TYPE_DOMAIN (type1),
      	   dom2 = TYPE_DOMAIN (type2);

      /* Both operands must be sets.
       * Elements of the sets must be of same type.
       * @@@@ Verify that the code checks this someday!
       */
      if (TREE_CODE (type2) != SET_TYPE)
	abort();

      /* @@@ Variable sized sets??? Should be able to support */
      if (TREE_CODE (TYPE_SIZE (type1)) != INTEGER_CST ||
	  TREE_CODE (TYPE_SIZE (type2)) != INTEGER_CST)
	abort ();

      if (dom1 == dom2)
	result_type = type1;
      else
	{
	  tree result_lo, result_hi;

	  switch (code) {
	  case BIT_XOR_EXPR: 
	  case BIT_IOR_EXPR:
	    /* Make the set large enough to hold both */
	    if (tree_int_cst_lt (TYPE_MIN_VALUE (dom1), TYPE_MIN_VALUE (dom2)))
	      result_lo = TYPE_MIN_VALUE (dom1);
	    else 
	      result_lo = TYPE_MIN_VALUE (dom2);
	    
	    if (tree_int_cst_lt (TYPE_MAX_VALUE (dom1), TYPE_MAX_VALUE (dom2)))
	      result_hi = TYPE_MAX_VALUE (dom2);
	    else 
	      result_hi = TYPE_MAX_VALUE (dom1);
	    
	    result_type = TREE_TYPE (type1); /* type of set */

	    if (TREE_TYPE (result_type))     /* Oops, an integer subrange */
	      result_type = TREE_TYPE (result_type); /* Get containing type */
	    
	    result_type = 
	      build_set_type
		(build_tree_list (result_type,
				  build_index_2_type (result_lo,
						      result_hi)));
	    break;
	    
	  case BIT_AND_EXPR:
	    /* Here we MAY make the set smaller, choose either one,
	       or take the larger (It does not matter)
	       (Well, check that the library works, though)
	       
	       Now we choose the first one.
	     */
	    result_type = type1;
	    break;

	  case BIT_ANDTC_EXPR:
	    result_type = type1;
	    break;
	  }
	}

      /* Declare a temporary variable for the resulting set */
      store = start_decl (get_unique_identifier ("set_result"),
			  build_tree_list (NULL_TREE, result_type),
			  0);
      finish_decl (store, NULL_TREE, NULL_TREE);
      setR = error_mark_node;
    }

  /* First, try to optimize trivialities before we
   * stack parameters to the runtime system
   */
  switch (code) {
  /* SET IN */
  case IN_EXPR:		/* op2 IN op1 ; ordinal IN set */
    /* IN_EXPR is INLINED; this should never happen */
    if (set_empty_p (TREE_TYPE (op1)))
      *result = boolean_false_node;
    break;
  /* SET CARD */
  case CARD_EXPR:	/* CARD(op1) */
    if (set_empty_p (TREE_TYPE (op1)))
      *result = integer_zero_node;
    break;

  /* SET BOOLEAN */
  case NE_EXPR:		/* op1 <> op2 ; !(op1 == op2) */
  case EQ_EXPR:		/* op1 == op2 */
    break;
  case LE_EXPR:		/* op1 <= op2 ; op1 is included in op2 */
    if (set_empty_p (TREE_TYPE (op1)))
      *result = boolean_true_node;
    break;
  case GE_EXPR:		/* op1 => op2 ; swap operands */
    if (set_empty_p (TREE_TYPE (op2)))
      *result = boolean_true_node;
    break;

  /* SET RETURNING OPERATIONS */
  case BIT_AND_EXPR:	/* op1 * op2 */
    if (set_empty_p (TREE_TYPE (op1)) || set_empty_p (TREE_TYPE (op2)))
      *result = get_empty_set_node (), retval = 1;
    break;

  case BIT_ANDTC_EXPR:	/* op1 - op2  == op1 AND (NOT op2) */
    if (set_empty_p (TREE_TYPE (op2)))
      *result = op1, retval = match1;
    else if (set_empty_p (TREE_TYPE (op1)))
      *result = get_empty_set_node (), retval = 1;
    break;

  case BIT_IOR_EXPR:	/* op1 + op2 */
    if (set_empty_p (TREE_TYPE (op1)))
      *result = op2, retval = match2;
    else if (set_empty_p (TREE_TYPE (op2)))
      *result = op1, retval = match1;
    break;

  case BIT_XOR_EXPR:	/* op1 >< op2 */
    if (set_empty_p (TREE_TYPE (op1)))
      *result = op2, retval = match2;
    else if (set_empty_p (TREE_TYPE (op2)))
      *result = op1, retval = match1;
    break;

  /* SET ASSIGNMENT */
  case MODIFY_EXPR:	/* op2 := op1 */
    break;

  default:
    abort();
  }

  if (retval)
    return retval;  /* If we optimized the operation away, return */

  /* Form run time system set parameters */
  setA = initialize_rts_set_parameter (op1, 1);

  if (TREE_CODE (type2) == SET_TYPE)
    setB = initialize_rts_set_parameter (op2, 2);

  if (setR)
    setR = initialize_rts_set_parameter (store, 3);

  /* Now do the real thing */
  switch (code) {
  /* SET IN */
  case IN_EXPR:		/* op2 IN op1 ; ordinal IN set */
    /* IN_EXPR is INLINED; this should never happen */
    *result = build_rts_call
      		 (r_SET_IN,
		  tree_cons (NULL_TREE, setA,
			     build_tree_list (NULL_TREE, op2)));
    break;
  /* SET CARD */
  case CARD_EXPR:	/* CARD(op1) */
    *result = build_rts_call
      		 (r_SET_CARD,
		  build_tree_list (NULL_TREE, setA));
    break;

  /* SET BOOLEAN */
  case NE_EXPR:		/* op1 <> op2 ; !(op1 == op2) */
    store = build_rts_call
      	       (r_SET_EQUAL,
		tree_cons (NULL_TREE, setA,
			   build_tree_list (NULL_TREE, setB)));
    *result = build_unary_op (TRUTH_NOT_EXPR, store, 0);
    break;
  case EQ_EXPR:		/* op1 == op2 */
    *result = build_rts_call
      		 (r_SET_EQUAL,
		  tree_cons (NULL_TREE, setA,
			     build_tree_list (NULL_TREE, setB)));
    break;
  case LE_EXPR:		/* op1 <= op2 ; op1 is included in op2 */
    *result = build_rts_call
                 (r_SET_LE,
		    tree_cons (NULL_TREE, setA,
			       build_tree_list (NULL_TREE, setB)));
    break;
  case GE_EXPR:		/* op1 => op2 ; swap operands */
    *result = build_rts_call
      		 (r_SET_LE,
		  tree_cons (NULL_TREE, setB,
			     build_tree_list (NULL_TREE, setA)));
    break;

  /* SET RETURNING OPERATIONS */
  case BIT_AND_EXPR:	/* op1 * op2 */
    horrible_set_kludge_type = result_type;
    *result = build_rts_call
		 (r_SET_INTERSECTION,
		  tree_cons (NULL_TREE, setA,
			     tree_cons (NULL_TREE, setB,
					build_tree_list (NULL_TREE, setR))));
    break;

  case BIT_ANDTC_EXPR:	/* op1 - op2  == op1 AND (NOT op2) */
    horrible_set_kludge_type = result_type;
    *result = build_rts_call
      		 (r_SET_DIFF,
		  tree_cons (NULL_TREE, setA,
			     tree_cons (NULL_TREE, setB,
					build_tree_list (NULL_TREE, setR))));
    break;

  case BIT_IOR_EXPR:	/* op1 + op2 */
    horrible_set_kludge_type = result_type;
    *result = build_rts_call
      		 (r_SET_UNION,
		  tree_cons (NULL_TREE, setA,
			     tree_cons (NULL_TREE, setB,
					build_tree_list (NULL_TREE, setR))));
    break;

  case BIT_XOR_EXPR:	/* op1 >< op2 */
    horrible_set_kludge_type = result_type;
    *result = build_rts_call
      		 (r_SET_SYMDIFF,
		  tree_cons (NULL_TREE, setA,
			     tree_cons (NULL_TREE, setB,
					build_tree_list (NULL_TREE, setR))));
    break;

  /* SET ASSIGNMENT */
  case MODIFY_EXPR:	/* op2 := op1 */
    *result = build_rts_call
      		(r_SET_COPY,
		 tree_cons (NULL_TREE, setB,
			    build_tree_list (NULL_TREE, setA)));
    break;

  default:
    abort();
  }

  return retval;
}

void
rts_read(rts_id, params)
int rts_id;
tree params;
{
    tree parm;
    tree file;
    int length;
    int is_text;
    tree arglist = NULL_TREE;

    if (params &&
	TREE_CODE (TREE_TYPE (TREE_VALUE (params))) == FILE_TYPE) {
	file = TREE_VALUE (params);
	params = TREE_CHAIN (params);
    } else
	file = input_file_node;

    if (file == NULL_TREE) {
	error ("File `input' was not mentioned in program header");
	input_file_node = error_mark_node;
	return;
    }

    is_text = TYPE_FILE_TEXT (TREE_TYPE (file)) == integer_one_node;

    if (rts_id == p_READLN)
	if (! is_text) {
	    error ("`Readln' is allowed only when reading from files of type `Text'");
	    return;
	}
    
    if (rts_id == p_READ) {
	if (params == NULL_TREE) {
	    error ("Too few parameters to predefined procedure `Read'");
	    return;
	}

	/* Non TEXT file reads */
	if (! is_text) {
	    tree the_file = build_tree_list (NULL_TREE, file);
	    tree buffer = build_buffer_ref (file);
	    for (parm = params; parm; parm = TREE_CHAIN (parm)) {

		/* @@@@ Should check that the types match! */

		/* Move the contents of the file buffer to the read param */
		expand_assignment (TREE_VALUE (parm), buffer, 0, 0);
		
		/* Do a get from the file */
		build_rts_call (p_GET, the_file);
	    }
	    return;
	}
    }

    /* find out the number of args we are reading */
    length = list_length (params);
    length += rts_id == p_READLN;	/* add P_LINE */
	
    arglist = tree_cons (NULL_TREE, build_unary_op (ADDR_EXPR, file, 0),
			 build_tree_list (NULL_TREE, size_int (length)));

    /* TEXT file reads */
    for (parm = params; parm; parm = TREE_CHAIN (parm)) {
	tree p = TREE_VALUE (parm);
	enum tree_code code = TREE_CODE (TREE_TYPE (p));
	int what;

	length--;
	switch (code) {
	case INTEGER_TYPE :
	    what = P_INT;
	    break;
	case CHAR_TYPE :
	    what = P_CHAR;
	    break;
	case REAL_TYPE :
	    what = P_REAL;
	    break;
#if 0
/* some day allow reading of strings... */
        case STRING_TYPE:
	case ARRAY_TYPE:
	    what = P_STRING;
	    push the address & max length
	    continue;
#endif
	default :
	    error ("parameter of Read/Readln from TEXT file is of wrong type");
	    return;
	}

	/* Inform the run time system of the next arg type */
	arglist = chainon (arglist,
			   build_tree_list (NULL_TREE, size_int (what)));

	/* Pass the address of the variable we want to read */
	arglist = chainon (arglist, 
		     build_tree_list (NULL_TREE,
				      build_unary_op (ADDR_EXPR, p, 0)));
    }

    if (rts_id == p_READLN) {
	/* Inform the run time system that we should do a readln */
	arglist = chainon (arglist, build_tree_list (NULL_TREE,
						     size_int (P_LINE)));
	length--;
    }

    if (length)
	abort ();

    expand_expr_stmt (rts_call (r_READ, void_type_node,
				NULL_TREE, arglist, 0));
}

void
rts_write(rts_id, params)
int rts_id;
tree params;
{
    tree parm;
    tree file;
    tree arglist;
    int length;
    int is_text;
    
    if (params &&
	TREE_CODE (TREE_TYPE (TREE_VALUE (params))) == FILE_TYPE) {
	file = TREE_VALUE (params);
	params = TREE_CHAIN (params);
    } else
	file = output_file_node;
    
    if (file == NULL_TREE) {
	error ("File `output' was not mentioned in program header");
	output_file_node = error_mark_node;
	return;
    }

    is_text  = TYPE_FILE_TEXT (TREE_TYPE (file)) == integer_one_node;

    if (rts_id == p_WRITELN)
	if (! is_text) {
	    error ("WRITELN is allowed only when writing to files of type TEXT");
	    return;
	}
    
    if (rts_id == p_WRITE) {
	if (params == NULL_TREE) {
	    error ("Too few parameters to predefined procedure WRITE");
	    return;
	}

	/* Non TEXT file writes */
	if (! is_text) {
	    tree the_file = build_tree_list (NULL_TREE, file);
	    tree buffer = build_buffer_ref (file);
	    for (parm = params; parm; parm = TREE_CHAIN (parm)) {

		/* @@@@ Should check that the types match! */

		/* Move the contents of the write parm to the file buffer */
		expand_assignment (buffer, TREE_VALUE (parm), 0, 0);
		
		/* Do a put to the file */
		build_rts_call (p_PUT, the_file);
	    }
	    return;
	}
    }

    /* find out the number of args we are writing */
    length = list_length (params);
    length += rts_id == p_WRITELN;	/* add P_LINE */
	
    arglist = tree_cons (NULL_TREE, build_unary_op (ADDR_EXPR, file, 0),
			 build_tree_list (NULL_TREE, size_int (length)));

    /* TEXT file writes */
    for (parm = params; parm; parm = TREE_CHAIN (parm)) {
	tree field1;
	tree field2;
	int what;
	tree p = TREE_VALUE (parm);
	enum tree_code code = TREE_CODE (TREE_TYPE (p));

	length--;

	if (TREE_PURPOSE (parm)) {
	    field1 = TREE_VALUE (TREE_PURPOSE (parm));
	    field2 = TREE_PURPOSE (TREE_PURPOSE (parm));
	} else {
	    field1 = NULL_TREE;
	    field2 = NULL_TREE;
        }

	switch (code) {
	case INTEGER_TYPE :
	    if (field1)
		what = P_FIX_INT;
	    else
		what = P_INT;
	    break;
	case BOOLEAN_TYPE :
	    if (field1)
		what = P_FIX_BOOL;
	    else
		what = P_BOOL;
	    break;
	case CHAR_TYPE :
	    if (field1)
		what = P_FIX_CHAR;
	    else
		what = P_CHAR;
	    break;
	case REAL_TYPE :
	    if (field1)
		if (field2)
		    what = P_FIX2_REAL;
		else
		    what = P_FIX1_REAL;
	    else
		what = P_REAL;
	    break;
	case ARRAY_TYPE:
	    if (! is_string_type (p)) {
		error ("Only Pascal strings may be written to TEXT files");
		continue;
	    }
	    
	    if (field1) {
		what = P_FIX_STRING;
		if (field2)
		    warning ("second field width allowed only when writing REAL type");
		field2 = field1;
	    } else
		what = P_STRING;

	    if (TREE_CODE (p) == STRING_CST)
		field1 = size_int (TREE_STRING_LENGTH (p) - 1);
	    else
		field1 = TYPE_MAX_VALUE (TYPE_DOMAIN (TREE_TYPE (p)));
	    /* pass the address of the string */
	    p = build_unary_op (ADDR_EXPR, p, 0);
	    break;
	    ;

	default :
	    error ("parameter to Write/Writeln to Textfile is of wrong type");
	    return;
	}
	if (field2 && what != P_FIX2_REAL && what != P_FIX_STRING)
	    warning ("second field width allowed only when writing REAL type");
	
	/* Inform the run time system of the next arg type */
	arglist = chainon (arglist,
			   build_tree_list (NULL_TREE, size_int (what)));
	
	/* Pass the variable we want to write */
	arglist = chainon (arglist, build_tree_list (NULL_TREE, p));

	if (field1)
	    arglist = chainon (arglist, build_tree_list (NULL_TREE, field1));

	if (field2)
	    arglist = chainon (arglist, build_tree_list (NULL_TREE, field2));
    }

    if (rts_id == p_WRITELN) {
	/* Inform the run time system that we should do a writeln */
	arglist = chainon (arglist, build_tree_list (NULL_TREE,
						     size_int (P_LINE)));
	length--;
    }

    if (length)
	abort ();

    expand_expr_stmt (rts_call (r_WRITE, void_type_node,
				NULL_TREE, arglist, 0));
}

/*
 * @@@. Sigh. I just noticed that array_type_nelts thinks that
 * the array type is max-min long. However, I assume that arrays are
 * max-min+1 long, so I had to make yet another routine...
 */

/* Return, as an INTEGER_CST node, the number of elements for
   TYPE (which is an ARRAY_TYPE).  */
tree
pascal_array_type_nelts (type)
     tree type;
{
  tree index_type = TYPE_DOMAIN (type);
  tree len =
    tree_int_cst_equal (TYPE_MIN_VALUE (index_type),
			integer_zero_node)
      ? TYPE_MAX_VALUE (index_type)
      : fold (build (MINUS_EXPR, integer_type_node,
		     TYPE_MAX_VALUE (index_type),
		     TYPE_MIN_VALUE (index_type)));

  return fold (build (PLUS_EXPR, integer_type_node,
		      len, integer_one_node));
}

tree
object_size (type)
tree type;
{
    if (TREE_CODE (type) != ARRAY_TYPE)
	return size_in_bytes (type);
    else
	return fold (build_binary_op (MULT_EXPR,
				      pascal_array_type_nelts (type),
				      object_size (TREE_TYPE (type)),
				      1));
}

/*
 * Implement pascal pack and unpack transfer procedures.
 */
tree
pascal_unpack_and_pack (unpack, unpacked, packed, ustart)
int  unpack;			/* Nonzero if UNPACK */
tree unpacked;
tree packed;
tree ustart;
{
  tree utype;
  tree ptype;
  tree elem_size;
  tree len;
  tree pmin, pmax, umin, umax;
  tree check;			/* Check of arguments */

  rtx x,y;
  rtx target;
  rtx source;
  rtx length;
  
  if (unpacked == error_mark_node || packed == error_mark_node)
    return error_mark_node;

  utype = TREE_TYPE (unpacked);
  ptype = TREE_TYPE (packed);
  pmin  = TYPE_MIN_VALUE (TYPE_DOMAIN (ptype));
  pmax  = TYPE_MAX_VALUE (TYPE_DOMAIN (ptype));
  umin  = TYPE_MIN_VALUE (TYPE_DOMAIN (utype));
  umax  = TYPE_MAX_VALUE (TYPE_DOMAIN (utype));

  /* Elements must be same size, get size in bytes */
  elem_size = size_in_bytes (TREE_TYPE (utype));
  
  x = expand_expr
    (build_indirect_ref
     (build_unary_op (ADDR_EXPR, build_array_ref (unpacked, ustart), 0),
      "unpacked"),
     0, BLKmode, 0);
  
  y = expand_expr
    (build_indirect_ref
     (build_unary_op (ADDR_EXPR, packed, 0),
      "packed"),
     0, BLKmode, 0);

  if (unpack)
    {
      target = x;
      source = y;
    }
  else
    {
      target = y;
      source = x;
    }

  /* Length we copy is the length of the packed array */
  len = fold (build (PLUS_EXPR, integer_type_node,
		     fold (build (MINUS_EXPR, integer_type_node, pmax, pmin)),
		     integer_one_node));
      
  /* Sanity check */
  /* Check that (ustart >= umin) and (ustart..umax >= pmin..pmax) */
  check =
    fold (build (TRUTH_AND_EXPR, boolean_type_node,
		 fold (build (GE_EXPR, boolean_type_node,
			      ustart, umin)),
		 fold (build (GE_EXPR, boolean_type_node,
			      fold (build (PLUS_EXPR, integer_type_node,
					   fold (build (MINUS_EXPR,
							integer_type_node,
							umax, ustart)),
					   integer_one_node)),
			      len))));

  if (! TREE_CONSTANT (check))
    /* @@@ Runtime check here variable size array */;
  else if (integer_zerop (check))
    {
      error ("Illegal arguments to %s", unpack ? "Unpack" : "Pack");
      return error_mark_node;
    }

  length = expand_expr
    	       (fold (build (MULT_EXPR, integer_type_node, elem_size, len)),
		0, SImode, 0);
  
  if (GET_CODE (source) == MEM && GET_CODE (target) == MEM)
    {
      extern rtx change_address ();

      if (GET_MODE (target) != BLKmode)
	target = change_address (target, BLKmode, 0);

      if (GET_MODE (source) != BLKmode)
	source = change_address (source, BLKmode, 0);

      emit_block_move
	(target, source, length, TYPE_ALIGN (ptype) / BITS_PER_UNIT);
    }
  else
    abort ();

  return NULL_TREE;
}

/*
 * routine constructs Pascal run time system calls with correct arguments.
 *
 * RTS_ID is the %token number of the run time system routine to call.
 * APAR is a TREE_LIST chain of arguments; args are in the TREE_VALUE field.
 * if there is something in the TREE_PURPOSE field, it is a TREE_LIST
 * node of write | writeln output field length expressions, the first 
 * expression is in TREE_VALUE and the second one is in TREE_PURPOSE
 * i.e. actual_parameter : expression : expression
 *
 */

tree
build_rts_call (rts_id, apar)
int rts_id;
register tree apar;   /* actual parameters of the routine */
{
  tree fpar = NULL_TREE;       /* formal parameters, default to varargs */
  tree rval = void_type_node;  /* return value; void_type_node if !function */

  tree un_init_this = NULL_TREE; /* If nonzero, try to un-initalize the beast */

  tree val = NULL_TREE, val2 = NULL_TREE;
  enum tree_code code = MINUS_EXPR; /* Something that is impossible */
  tree type, type2;
  int set_ref = 0;	       /* 1 only for SET returning functions */
  int length;		       /* length of the actual parameter list */
  int rts_inline   = 0;	       /* set to 1 if this is compiled inline */
  int wins         = 1;	       /* Will be zeroed if formal params are not
				  constants */
  tree temp;

  char *errstr = (char *)NULL;
  tree retval   = NULL_TREE;

  length = list_length (apar);

#define INLINE_RTS_LENGTH(nam, len) \
 { if (length != len) {						\
    error ("incorrect number of parameters to `%s'", nam); 	\
    return error_mark_node; } 					\
   else rts_inline = 1; }

  if (apar) {
      val  = TREE_VALUE (apar);
      type = TREE_TYPE (val);
      code = TREE_CODE (type);
      if (TREE_CHAIN (apar))
	{
	  val2  = TREE_VALUE (TREE_CHAIN (apar));
	  type2 = TREE_TYPE (val2);
	}
      /* @@@ wins is not used */
      for (temp = apar; temp && wins; temp = TREE_CHAIN (temp))
	  wins = TREE_CONSTANT (TREE_VALUE (temp));
  }

  switch (rts_id) {

  /* rts_inline functions */
  case p_ABS:
      INLINE_RTS_LENGTH("abs", 1);
      if (INT_REAL(code))
          retval = build_unary_op (ABS_EXPR, val, 0);
      else 
	  errstr = "argument to `abs' must be of integer type or real type";
      break;

  case p_SQR:
      INLINE_RTS_LENGTH("sqr", 1);
      if (INT_REAL(code))
	  retval = build_binary_op(MULT_EXPR, val, val,1);
      else
	  errstr = "argument to `sqr' must be of integer type or real type";
      break;

  case p_TRUNC:	/* check the return value; maybe wrong TRUNC_EXPR used */
      INLINE_RTS_LENGTH("trunc", 1);
      if (code == REAL_TYPE)
	  retval = convert (integer_type_node, val);
      else
	  errstr = "argument to `trunc' must be of real type";
      break;

  case p_ROUND:
      INLINE_RTS_LENGTH("round", 1);
      if (code == REAL_TYPE) {
	  retval = fold (build1 (FIX_ROUND_EXPR, integer_type_node, val));
      } else 
	  errstr = "argument to `round' must be of real type";
      break;

  case p_SUCC:		/* rts_inline */
      INLINE_RTS_LENGTH("succ", 1);
      if (ORDINAL_TYPE(code)) {
	  retval = convert (type,
			    fold (build_binary_op (PLUS_EXPR, val, integer_one_node,0)));
      } else 
	  errstr = "argument to `succ' must be of ordinal type";
      break;

  case p_PRED:		/* rts_inline */
      INLINE_RTS_LENGTH("pred", 1);
      if (ORDINAL_TYPE(code)) {
	  retval = convert (type,
			    fold (build_binary_op (MINUS_EXPR, val, integer_one_node,0)));
      } else 
	  errstr = "argument to `pred' must be of ordinal type";
      break;

  case p_ORD:		/* rts_inline */
      INLINE_RTS_LENGTH("ord", 1);
      if (ORDINAL_TYPE(code))
	  retval = convert (integer_type_node, val);
      else 
	  errstr = "argument to `ord' must be of ordinal type";
      break;

  case p_CHR:		/* rts_inline */
      INLINE_RTS_LENGTH("chr", 1);
      if (code == INTEGER_TYPE)
	  retval = convert (char_type_node, val); 
      else 
	  errstr = "argument to `chr' must be of integer type";
      break;

  case p_ODD:		/* rts_inline */ 
      INLINE_RTS_LENGTH("odd", 1);
      if (code == INTEGER_TYPE)
	  retval = convert (boolean_type_node,
			    fold (build_binary_op (BIT_AND_EXPR, val, integer_one_node,0)));
      else 
	  errstr = "argument to `odd' must be of integer type";
      break;

  case p_CMPLX:		/* rts_inline */
      INLINE_RTS_LENGTH("cmplx", 2);
      if (! INT_REAL (code) || ! INT_REAL (TREE_CODE (TREE_TYPE (val2))))
	  errstr = "Wrong type arguments to `cmplx'";
      else {
	  tree complex = TREE_TYPE (complex_type_node);

	  if (type != complex)
	      val = convert (complex, val);

	  if (TREE_TYPE (val2) != complex)
	      val2 = convert (complex, val2);

	  retval = build (COMPLEX_EXPR, complex_type_node, val, val2);
      }
      break;

  case p_IM:
      INLINE_RTS_LENGTH("im", 1);
      if (code == COMPLEX_TYPE)
	  retval = build_unary_op (IMAGPART_EXPR, val, 1);
      else 
	  errstr = "argument to `im' must be of complex type";
      break;
      
  case p_RE:
      INLINE_RTS_LENGTH("re", 1);
      if (code == COMPLEX_TYPE)
	  retval = build_unary_op (REALPART_EXPR, val, 1);
      else 
	  errstr = "argument to `re' must be of complex type";
      break;

  case p_PACK:
      emit_line_note (input_filename, lineno);
      INLINE_RTS_LENGTH ("pack", 3);
      { tree val3  = TREE_VALUE (TREE_CHAIN (TREE_CHAIN (apar)));
	tree type3 = TREE_TYPE (val3);
	tree unpacked_domain = TYPE_DOMAIN (type);

	if (TREE_CODE (type3) != ARRAY_TYPE
#ifdef CHECK_PACKED
	    || ! PASCAL_TYPE_PACKED (type3)
#endif
	    )
	  error ("Third argument to PACK must be a packed array");
	else if (TREE_CODE (type) != ARRAY_TYPE
#ifdef CHECK_PACKED
		 || PASCAL_TYPE_PACKED (type)
#endif
		 )
	  error ("First argument to PACK must be an unpacked array");
	else if (TREE_CODE (type2) != TREE_CODE (unpacked_domain) &&
		 (TREE_CODE (unpacked_domain) != INTEGER_TYPE ||
		  TREE_CODE (type2) != TREE_CODE(TREE_TYPE (unpacked_domain))))
	  error ("Second argument of PACK must be of unpacked array index type");
	else if (TREE_CODE (TREE_TYPE (type)) != TREE_CODE (TREE_TYPE (type3)))
	  error ("Source and destination arrays in PACK are not of same type");
	else
	  return pascal_unpack_and_pack (0, val, val3, val2);
      }
      return error_mark_node;

  case p_UNPACK:
      emit_line_note (input_filename, lineno);
      INLINE_RTS_LENGTH ("unpack", 3);
      { tree val3  = TREE_VALUE (TREE_CHAIN (TREE_CHAIN (apar)));
	tree type3 = TREE_TYPE (val3);
	tree unpacked_domain = TYPE_DOMAIN (type2);

	/* I just wonder why on venus they had to shuffle these */
	if (TREE_CODE (type2) != ARRAY_TYPE
#ifdef CHECK_PACKED
	    || ! PASCAL_TYPE_PACKED (type2)
#endif
	    )
	  error ("Second argument to UNPACK must be an unpacked array");
	else if (TREE_CODE (type) != ARRAY_TYPE
#ifdef CHECK_PACKED
		 || ! PASCAL_TYPE_PACKED (type)
#endif
		 )
	  error ("First argument to UNPACK must be a packed array");
	else if (TREE_CODE (type3) != TREE_CODE (unpacked_domain) &&
		 (TREE_CODE (unpacked_domain) != INTEGER_TYPE ||
		  TREE_CODE (type3) != TREE_CODE(TREE_TYPE (unpacked_domain))))
	  error ("Third argument of UNPACK must be of unpacked array index type");
	else if (TREE_CODE (TREE_TYPE (type)) != TREE_CODE (TREE_TYPE (type2)))
	  error ("Source and destination arrays in UNPACK are not of same type");
	else
	  return pascal_unpack_and_pack (1, val2, val, val3);
      }
      return error_mark_node;

  /* calls to run time system */

  /* statements */
  case p_NEW:
      if (code != POINTER_TYPE) {
	  errstr = "POINTER type required for NEW";
	  break;
      }
      /* Pass a reference to this pointer */
      fpar = chainon (do_ptype (type, 1, 0),
		      ptype_int);

      if (length != 1) {
	  static int informed = 0;
	  if (! informed)
	    {
	      warning ("tag fields ignored in NEW");
	      informed++;
	    }
	  apar = copy_node (apar);
      }
      un_init_this = build_indirect_ref (TREE_VALUE (apar));
      apar = chainon (apar, build_tree_list (NULL_TREE,
					     object_size (TREE_TYPE (type))));
      break;

  case p_DISPOSE:
      if (code != POINTER_TYPE) {
	  errstr = "POINTER type required for DISPOSE";
	  break;
      }
      fpar = chainon (do_ptype (type, 0, 0),
		      ptype_int);
      if (length != 1) {
	  static int informed = 0;
	  if (! informed)
	    {
	      warning ("tag fields ignored in DISPOSE");
	      informed++;
	    }
	  apar = copy_node (apar);
      }
      apar = chainon (apar, build_tree_list (NULL_TREE,
					     object_size (TREE_TYPE (val))));
      break;

  case p_POSITION:
  case p_LASTPOSITION:
      rval = integer_type_node;
  case p_CLOSE:
  case p_UPDATE:
  case p_PUT:
  case p_GET:
  case p_PAGE:
      if (code != FILE_TYPE)
	  errstr = "FILE type required";
      else
	  fpar = do_ptype (type, 1, 1);
      break;

  case p_HALT:
      if (length == 0)
	  apar = build_tree_list (NULL_TREE, integer_zero_node);
      else if (code != INTEGER_TYPE) {
	  errstr = "HALT requires integer type parameter";
	  break;
      }
      fpar = do_ptype (TREE_TYPE (TREE_VALUE (apar)), 0, 1);
      break;

  case p_DATE:
  case p_TIME:
      /* Should create a standard record TimeStamp
	 @@ This is a temporary test feature
       */
      if (pedantic)
	  warning ("ISO Pascal does not define`Date' or `Time'");
      rval = ptr_type_node; /* This is WRONG! should return string_type */
      if (code != RECORD_TYPE) /* any record :-) */
	  errstr = "Timestamp record type parameter required";
      else
	  fpar = do_ptype (type, 0, 1);
      break;

  case p_GETTIMESTAMP:
      /* Should create a standard record TimeStamp
	 @@ This is a temporary test feature
       */
      if (pedantic)
	  warning ("ISO Pascal does not define`Gettimestamp'");
      if (code != RECORD_TYPE) /* @@@@ accepts any record :-) */
	  errstr = "Timestamp record type parameter required";
      else
	  fpar = do_ptype (type, 1, 1);
      break;

  case p_SEEKUPDATE:
  case p_SEEKREAD:
  case p_SEEKWRITE:
      if (pedantic)
	  warning ("ISO Pascal does not define seek operations");
      if (code != FILE_TYPE)
	  errstr = "FILE type required for seek operations";
      else
	  fpar = chainon (do_ptype (type, 1, 0),
			  do_ptype (integer_type_node, 0, 1));
      break;

  /* functions */
  case p_EOF:
      rval = boolean_type_node;

      if (length == 0)
	  if (input_file_node)
	      apar = build_tree_list (NULL_TREE, input_file_node);
	  else
	      errstr = "File `input' was not mentioned in program heading";
      else if (code != FILE_TYPE) {
	  errstr = "EOF may only be applied to FILE type objects";
	  break;
      }
      if (apar)
	  fpar = do_ptype (TREE_TYPE (TREE_VALUE (apar)), 1, 1);
      break;

  case p_EOLN:
      rval = boolean_type_node;

      if (length == 0)
	  if (input_file_node)
	      apar = build_tree_list (NULL_TREE, input_file_node);
	  else
	      errstr = "File `input' was not mentioned in program heading";
      else if (code != FILE_TYPE || TYPE_FILE_TEXT (TREE_TYPE (val)) != integer_one_node) {
	  errstr = "EOLN may only be applied to TEXT type objects";
	  break;
      }
      if (apar)
	  fpar = do_ptype (TREE_TYPE (TREE_VALUE (apar)), 1, 1);
      break;

  case p_SQRT:
      rval = double_type_node;
      if (code == INTEGER_TYPE || code == REAL_TYPE)
	  fpar = ptype_double;
      else
	  errstr = "argument to `sqrt' must be of integer type or real type";
      break;

  case p_SIN:
  case p_COS:
  case p_EXP:
  case p_LN:
  case p_ARCTAN:
      rval = double_type_node;
      fpar = ptype_double;
      break;

  case p_EXTEND:
      if (pedantic)
	  warning ("ISO Pascal does not define `Extend'");
  case p_RESET:
  case p_REWRITE:
      if (code != FILE_TYPE)
	  errstr = "FILE type required";
      else {
	  tree size = integer_zero_node;
	  if (val2 != null_pointer_node)
	      if (is_string_type (val2)) {
		  fpar = chainon (do_ptype (type, 1, 0),
				  chainon (do_ptype (TREE_TYPE(val2), 1, 0),
					   ptype_int));
		  if (TREE_CODE (val2) == STRING_CST)
		      size = size_int (TREE_STRING_LENGTH (val2) - 1);
		  else
		      size = TYPE_MAX_VALUE (TYPE_DOMAIN (TREE_TYPE (val2)));
	      } else
		  errstr = "Optional file name parameter to REWRITE/RESET/EXTEND must be a string";
	  else
	      fpar = chainon (do_ptype (type, 1, 0),
			      chainon (do_ptype (char_array_type_node, 2, 0),
				       ptype_int));
	  apar = chainon (apar, build_tree_list (NULL_TREE, size));
      }
      break;

  case p_WRITELN:
  case p_WRITE:
      emit_line_note (input_filename, lineno);
      rts_write (rts_id, apar);
      return NULL_TREE;

  case p_READ:
  case p_READLN:
      emit_line_note (input_filename, lineno);
      rts_read (rts_id, apar);
      return NULL_TREE;

  /* extensions */
  case p_MARK:
      if (pedantic)
	  warning ("ISO Pascal does not define `Mark'");
      if (code != POINTER_TYPE)
	  errstr = "Pointer type required for `Mark'";
      else
	  fpar = do_ptype (type, 1, 1);
      break;

  case p_RELEASE:
      if (pedantic)
	  warning ("ISO Pascal does not define `Release'");
      if (code != POINTER_TYPE)
	  errstr = "Pointer type required for `Release'";
      else
	  fpar = do_ptype (type, 0, 1);
      break;

  case p_DEFINESIZE:
      if (pedantic)
	  warning ("ISO Pascal does not define `Definesize'");
      if (code != FILE_TYPE)
	  errstr = "FILE type required for `Definesize'";
      else
	  fpar = chainon (do_ptype (type, 1, 0),
			  do_ptype (integer_type_node, 0, 1));
      break;

  case p_EMPTY:
      if (pedantic)
	  warning ("ISO Pascal does not define `Empty'");
      if (code != FILE_TYPE)
	  errstr = "FILE type required for `Empty'";
      else
	  fpar = do_ptype (type, 1, 1);
      rval = boolean_type_node;
      break;

    /* @@@ Should be more careful with the parameter types.
     * But since compiler generates these calls, and no
     * VAR parameters are generated, don't bother
     *
     * SETs are passed as records to the runtime system
     * struct { int *Set; int lo; int hi; }.
     */
    case r_SET_CARD:
      if (code != POINTER_TYPE)
	errstr = "SET type required";
      rval = integer_type_node;
      break;
    case r_SET_RANGE:
    case r_SET_SINGLE:
      if (code != POINTER_TYPE)
	errstr = "SET type required";
      break;
    case r_SET_IN:
    case r_SET_LE:
    case r_SET_EQUAL:
      if (code != POINTER_TYPE)
	errstr = "SET type required";
      rval = boolean_type_node;
      break;

      /* @@@ Note that type of res set is set according to setA. (bug) */
    case r_SET_COPY:
    case r_SET_UNION:
    case r_SET_DIFF:
    case r_SET_SYMDIFF:
    case r_SET_INTERSECTION:
      if (code != POINTER_TYPE)
	errstr = "SET type required";
      /* AARGH. I apologize for doing this the dirty way. */
      rval = build_pointer_type (horrible_set_kludge_type);
      set_ref = 1;
      break;

    default:
      ASSERT (0, "no such run time system routine");
    }

  horrible_set_kludge_type = NULL_TREE;

  if (errstr) {
      error (errstr);      
      return error_mark_node;
  }

  /* Output a RTL line note */
  emit_line_note (input_filename, lineno);

  /* construct a call to the run time system if not compiled inline */

  if (! rts_inline) {

      retval = rts_call (rts_id, rval, fpar, apar, set_ref);

      /* if this is a statement, built rtl from it */
      /* otherwise let the caller do whatever it likes to do with it */
      if (rval == void_type_node)
	  expand_expr_stmt (retval);

      /* Un-initialize at run-time the node in UN_INIT_THIS */
      if (un_init_this)
	  init_any (un_init_this, 0);
  }      
  return retval;
}

/* the ISO 7185 standard says that the two ways, abbreviated (when
 * giving several index types) and full (when listing each index type
 * separately), of specifying and accessing arrays are equivalent.
 *
 * So we represent arrays internally in their full form.
 */

/* the ELT_TYPE is the user specified element type.
 *
 * the INDEX_LIST is a list of TREE_LIST nodes of index types,
 * in REVERSE ORDER, and thus the ELT_TYPE is the type of the
 * first element in the INDEX_LIST.
 *
 * Returns the type of the complete array specified by the user.
 */
tree
build_pascal_array_type (elt_type, index_list)
tree elt_type;
tree index_list;
{
    tree link;
    for (link = index_list; link; link = TREE_CHAIN (link)) {
	elt_type = build_array_type (elt_type, TREE_VALUE(link));
    }
    return elt_type;
}

/*
 * Constructs a reference to a pascal array. The VAR_ACCESS is
 * a VAR_DECL node and the INDEX_LIST is a TREE_LIST chain of
 * expressions to index the array with. This INDEX_LIST is
 * passed in FORWARD order (not reversed, as in build_pascal_array_type)
 */
tree
build_pascal_array_ref (var_access, index_list)
tree var_access;
tree index_list;
{
    tree link;
    for (link = index_list; link; link = TREE_CHAIN (link)) {
	var_access = build_array_ref (var_access, TREE_VALUE(link));
    }
    return var_access;
}

static int
guess_precision (minval, maxval)
tree minval, maxval;
{
    if (int_fits_type_p (minval, signed_char_type_node) &&
	int_fits_type_p (maxval, signed_char_type_node))
	return TYPE_PRECISION (signed_char_type_node);
    else if (int_fits_type_p (minval, short_integer_type_node) &&
	     int_fits_type_p (maxval, short_integer_type_node))
	return TYPE_PRECISION (short_integer_type_node);
    else if (int_fits_type_p (minval, integer_type_node) &&
	     int_fits_type_p (maxval, integer_type_node))
	return TYPE_PRECISION (integer_type_node);
    else if (int_fits_type_p (minval, long_integer_type_node) &&
	     int_fits_type_p (maxval, long_integer_type_node))
	return TYPE_PRECISION (long_integer_type_node);
    else {
	error ("range will not fit in the largest integer type available");
	return TYPE_PRECISION (long_integer_type_node);
    }
}

/* Create and return a type for signed integers of PRECISION bits, if
   the bounds are constant, otherwise just return a new range type with
   variable bounds. */
tree
make_signed_range (minval, maxval)
tree minval;
tree maxval;
{
  register tree type = make_node (INTEGER_TYPE);

  /* set the upper and lower limits of the range */

  TYPE_MIN_VALUE (type) = minval;
  TYPE_MAX_VALUE (type) = maxval;

  if (TREE_CODE (minval) == INTEGER_CST && TREE_CODE (maxval) == INTEGER_CST) {

      /* Set these things only if the bounds are constant */

      TYPE_PRECISION (type) = guess_precision (minval, maxval);
      
      /* Lay out the type: set its alignment, size, etc.  */
      
      TYPE_SIZE (type) = 0;
      layout_type (type);
  }      
  return type;
}

/* Create a subrange type for Pascal.
   It is (currently always) an integer with TYPE_MIN_VALUE
   and TYPE_MAX_VALUE set to the lower and upper limits
   as specified in the type definition.

   MINVAL and MAXVAL are INTEGER_CST nodes, whose TREE_TYPE
   decides the actual type of the subrange, or they
   are variable bounds in which case we just construct a
   integer_type node to contain the bounds. */

tree
build_range_type (minval, maxval)
tree minval, maxval;
{
    tree itype;
    tree rangetype = TREE_TYPE (minval);

    if (TREE_TYPE (minval) != TREE_TYPE (maxval)) {
	error ("range limits are not of the same type");
	return error_mark_node;
    }

    itype = make_signed_range (minval, maxval);

    /* see comment in tree.def in INTEGER_TYPE */
    TREE_TYPE (itype) = rangetype;
    return itype;
}

/*
 * The First word of the FILE_TYPE BLKmode variable is
 * a pointer to the file buffer we want to access.
 *
 *  So we need to generate:
 *  *(*(&file))
 *
 * If symbol LAZY_IO is defined, each buffer access needs to check
 * that the buffer is valid. If not, we need to do a get before
 * accessing the buffer.
 *
 *  If LAZY_IO is defined, we can select at compile time if we want to use
 * it or not; run time system should support both that and the old method.
 *
 * Lazy I/O helps to overcome the biggest problem of Pascal: It's I/O system.
 *
 * When we do a reset or read something from a file, the old method
 * needs to read new contents to the buffer before the data is
 * actually needeed. This is annoying if you do interactive programs,
 * the output to terminal asking for input comes after you have
 * already given the input to the program, or then you have to code
 * things differently for terminals and files, which is also annoying.
 *
 * The old method GPC uses is the same implementation dependent feature
 * that the Pax compiler uses. The run time system checks each file when
 * it's reset, and if it's a terminal, it sets EOLN implicitely on and places
 * a space in the file buffer. This is valid according to the standard, since
 * it explicitely states that handling of INPUT and OUTPUT is implementation
 * dependent.
 *
 * Lazy I/O means that we must do a PUT as early as we can, and GET as
 * late as we can. The first condition is satisfied either by not
 * buffering output at all, or then flushing output to terminals
 * before each get; the second condition is fulfilled when we check
 * that the buffer is valid each time we generate buffer references.
 */

tree
build_buffer_ref (file)
tree file;
{
  register tree ft = TREE_TYPE (file); 	/* type of variable */
  tree lazy_undefined = size_int (FiUnd);
  
  if (TREE_CODE (ft) == FILE_TYPE) {
      tree t = TREE_TYPE (ft);		/* type of file component */
      register tree ref =
	  build1 (INDIRECT_REF, TYPE_MAIN_VARIANT (t),	       /* type we point to */
		 build1 (INDIRECT_REF, build_pointer_type (t), /* Through this pointer */
			build1 (ADDR_EXPR, ft, file)));	       /* Address of file */

      TREE_READONLY (ref) = TREE_READONLY (t);
      TYPE_VOLATILE (ref) = TYPE_VOLATILE (t) || TYPE_VOLATILE (file);
      TREE_THIS_VOLATILE (ref) = TYPE_VOLATILE (t);
      return ref;
  } else if (file != error_mark_node && ft != error_mark_node)
      error ("file buffer referencing requires a FILE type argument");
  return error_mark_node;
}

void
associate_external_files ()
{
    tree output = get_identifier (OUTPUT_FILE);
    tree input  = get_identifier (INPUT_FILE);
    tree link;
    tree files = NULL_TREE;

    if (pedantic && external_file_name_list == NULL_TREE)
	warning ("No program parameters specified"); /* hehehehe, imitate Pax */

    for (link = external_file_name_list;
	 link;
	 link = TREE_CHAIN (link)) {
	tree id = TREE_VALUE (link);
	tree name = lookup_name (id);
	if (id == output)
	    if (PASCAL_REDEFINABLE_DECL (name)) {
		files = chainon (files, build_tree_list (NULL_TREE, output));
		output_file_node = error_mark_node;
	    } else
		warning ("Identifier `output' in program heading has been redeclared from its standard meaning");
	else if (id == input)
	    if (PASCAL_REDEFINABLE_DECL (name)) {
		files = chainon (files, build_tree_list (NULL_TREE, input));
		input_file_node = error_mark_node;
	    } else
		warning ("Identifier `input' in program heading has been redeclared from its standard meaning");
	else {
	    /* see if we have declared all the names in program param list */
	    if (name == NULL_TREE)
	     warning ("Identifier `%s' in program heading is undefined",
			 IDENTIFIER_POINTER (id));
	    else if (TREE_CODE (TREE_TYPE (name)) != FILE_TYPE)
	     warning ("Identifier `%s' in program heading is not of type FILE",
		      IDENTIFIER_POINTER (id));
	    else
		PASCAL_EXTERNAL_OBJECT (name) = 1;
	}
    }
    if (files) {
	declare_vars (files, text_type_node, NULL_TREE);
	if (input_file_node == error_mark_node) {
	    input_file_node = lookup_name (input);
	    PASCAL_EXTERNAL_OBJECT (input_file_node) = 1;
	}
	if (output_file_node == error_mark_node) {
	    output_file_node = lookup_name (output);
	    PASCAL_EXTERNAL_OBJECT (output_file_node) = 1;
	}
    }
}

/* Internal routine to un_initialize_block.

   This one does the initialization of the object
   (or a component of an object) specified in decl.
   This type is a BASIC type, i.e. not a structured type
   that has been exploded up by init_any ()
 */
static void
init_simple (decl, the_end)
tree decl;
int the_end;
{
    tree type = TREE_TYPE (decl);
    enum tree_code code = TREE_CODE (type);

    /* Clean up only local files */
    if (the_end && code != FILE_TYPE)
	return;	/* Forget them! */

    switch (code) {
    case RECORD_TYPE:
    case ARRAY_TYPE:
	abort ();
    case FILE_TYPE :
	{
	    char *temp;
	    tree temp1 = NULL_TREE;
	    tree file_size;
	    tree file_kind;
	    int lazyfile = 0;
	    
	    if (the_end) {
		/* Close the files on exit */
		build_rts_call (p_CLOSE,
				build_tree_list (NULL_TREE, decl));
		break;
	    }
	    
	    if (PASCAL_EXTERNAL_OBJECT (decl)) {
		temp  = IDENTIFIER_POINTER (DECL_NAME (decl));
		temp1 = build_string (strlen (temp)+1, temp);
		TREE_TYPE (temp1) = char_array_type_node;
		temp1 = build_tree_list (NULL_TREE,
					 build_unary_op (ADDR_EXPR, temp1, 0));
	    } else
		temp1 = build_tree_list (NULL_TREE,
					 integer_zero_node);

	    if (pedantic && PASCAL_TYPE_PACKED (type))
	      warning ("Word `PACKED' has no effect on external file format");
	    
	    /* Make a TREE_LIST node of the size of the file buffer */
	    /* the size is in bits, if the file is packed, else bytes */
	    /* (the run time system does not pack files yet, so it converts */
	    /*  sizes back to bytes) */
	    file_size =
	      build_tree_list
		(NULL_TREE,
		 (! PASCAL_TYPE_PACKED (type) ?
		    size_in_bytes (TREE_TYPE (type)) :
		    fold (build (MULT_EXPR, sizetype,
				 size_in_bytes (TREE_TYPE (type)),
				 size_int (BITS_PER_UNIT)))));

#ifdef LAZY_IO
	    if (flag_lazy_io)
		lazyfile = 1;
#endif
	    
	    file_kind = build_tree_list (NULL_TREE,
			  size_int (TREE_INT_CST_LOW (TYPE_FILE_TEXT (type))
				    | PASCAL_TYPE_PACKED (type)     << 1
				    | PASCAL_EXTERNAL_OBJECT (decl) << 2
				    | lazyfile		            << 3));


	    /* Call and construct parameters to the run time system
	     * routine that initializes a file buffer to a known state
	     */
	    expand_expr_stmt
	      (rts_call (r_INITFDR, void_type_node,
			 do_ptype (type, 1, 0),
			 tree_cons (NULL_TREE, decl,
				    chainon (temp1,
					     chainon (file_size, file_kind))),
			 0));
	    break;
	}
	default :
	    /* nothing yet!!! */ ;		    
    }
}

/* Subroutine of init_any */
static void
init_record (thing, the_end, base)
tree thing;
int the_end;
tree base;
{
    tree field;

    for (field = TYPE_FIELDS (TREE_TYPE (thing));
	 field;
	 field = TREE_CHAIN (field)) {
	if (DECL_NAME (field))
	    init_any (build_component_ref (base, DECL_NAME (field)),
		      the_end);
	else {
	    tree inner_type = TREE_TYPE (field);
	    if (TREE_CODE (inner_type) == RECORD_TYPE || 
		TREE_CODE (inner_type) == UNION_TYPE)
		init_record (field, the_end, base);
	    else
		warning ("strange unnamed field in record");
	}
    }
}

/* Subroutine of un_initialize_block
 *
 * Recursively propagate through a structured type
 * un_initializing all simple types with init_simple ()
 */
static void
init_any (thing, the_end)
tree thing;
int the_end;
{
    tree temp1;
    int i, maxint = TREE_INT_CST_LOW (integer_maxint_node);
    tree type = TREE_TYPE (thing);
    enum tree_code code = TREE_CODE (type);
    switch (code) {
    case SET_TYPE :
        if (! the_end)
	  {
	    /* nothing currently */
	  }
        break;
    case FILE_TYPE :
	init_simple (thing, the_end);
	break;
    case ARRAY_TYPE : {
	temp1 = TYPE_DOMAIN (type);

	/* Avoid wraparound if i is maxint */
	/* @@@Lets hope min_value is less than max_value :-) */
	if (TREE_CODE (TYPE_MIN_VALUE (temp1)) == INTEGER_CST &&
	    TREE_CODE (TYPE_MAX_VALUE (temp1)) == INTEGER_CST)
	    for (i = TREE_INT_CST_LOW (TYPE_MIN_VALUE (temp1));
		 i <= TREE_INT_CST_LOW (TYPE_MAX_VALUE (temp1)) &&
		 i <  maxint; i++)
		init_any (build_array_ref (thing, size_int (i)),
			  the_end);
	else if (pedantic)
	    warning ("Dynamic arrays are not initialized properly");
	break;
    }
    case UNION_TYPE :	/* In Pascal only in variant records */
    case RECORD_TYPE :
	init_record (thing, the_end, thing);
	break;
    }
}

/*
 * This code will (some day) un-initialize all variables
 * declared in the current block level.
 *
 * It does not do it yet, however, except for FILE_TYPE
 * nodes, which won't work if not un-initialized.
 *
 * THE_END is nonzero if this is called to clean up after
 * the block's statements have been executed.
 *
 *  This also initializes files in
 *   1) components of structured types and
 *   2) parts of dynamic objects. (This should go elsewhere)
 *
 */
void
un_initialize_block (names, the_end)
tree names;
int the_end;
{
    tree decl;
    for (decl = names; decl; decl = TREE_CHAIN (decl))
	if (TREE_CODE (decl) == VAR_DECL)
	    init_any (decl, the_end);
}

void
init_std_files ()
{
    if (output_file_node)
	build_rts_call (p_REWRITE, 
			tree_cons (NULL_TREE,
				   output_file_node,
				   build_tree_list (NULL_TREE,
						    null_pointer_node)));
    if (input_file_node)
	build_rts_call (p_RESET, 
			tree_cons (NULL_TREE,
				   input_file_node,
				   build_tree_list (NULL_TREE,
						    null_pointer_node)));
}

/* @@@ range needs to be the correct type!!!! */
tree
convert_type_to_range (type)
tree type;
{
    enum tree_code code = TREE_CODE (type);
    tree min, max;
    tree itype;

    if (! ORDINAL_TYPE (code)) {
	error ("ordinal type expected");
	return error_mark_node;
    }
#define USE_RANGE_TYPE
#ifdef USE_RANGE_TYPE
    min = TYPE_MIN_VALUE (type);
    max = TYPE_MAX_VALUE (type);

    if (TREE_TYPE (min) != TREE_TYPE (max)) {
	error ("range limits are not of the same type");
	return error_mark_node;
    }
    itype = make_signed_range (TYPE_MIN_VALUE (type),
			       TYPE_MAX_VALUE (type));
    
    /* see comment in tree.def in INTEGER_TYPE */
    TREE_TYPE (itype) = type;
    return itype;
#else
    return build_index_2_type (TYPE_MIN_VALUE (type),
			       TYPE_MAX_VALUE (type));
#endif
}

/*
 * Enables/disables procedure/function evaluation when
 * the routines have no parameters. 
 * Reason:
 *   1) When parsing an IDENTIFIER it may be either a routine
 *	name or a variable name.
 *   2) When passing such an item to a routine, it may have to be
 *	evaluated, or not depending of the corresponding FORMAL
 *	parameter type; if this is a procedural parameter, the
 *	routine should be passed, if not, the routine should be
 *	evaluated, and the result passed.
 */
static int evaluate_function_calls = 1;

/*
 * disables the evaluation of procedure/function calls with no parameters
 * and returns the old value of the enable/disable flag.
 *
 * They need to be stacked, because routine calls can be nested :-)
 */
int
suspend_function_calls ()
{
    int old = evaluate_function_calls;
    evaluate_function_calls = 0;
    return old;
}

/*
 *  Resumes routine calls for routines with no parameters
 *  if STATE is nonzero.
 *
 *  STATE should be the value previously returned by
 *  suspend_function_calls ().
 */
void
resume_function_calls (state)
int state;
{
    evaluate_function_calls = state;
}

/*
 * Maybe call the function, or pass it as a routine parameter.
 */
tree
maybe_call_function (fun)
tree fun;
{
    tree temp;
    /* This is a quick test to do this.
       to do it properly requires knowledge what we are doing:
       1) this is a procedure statement without parameters
       2) this is an assignment from function with no parameters
          to some variable, or passing the RESULT of such function
	  to another routine.
       3) this procedure or function is being passed to a procedural
          parameter list.

       The problem is this:
         in 1 & 2 above the procedure/function should be CALLED;
	 in 3 it should NOT be called, but rather passed as a function
	 definition.
     */
    if (evaluate_function_calls)
	temp = build_function_call (fun, NULL_TREE);
    else
	temp = fun;
    return temp;
}


/*
 * Called from gpc-parse.y and gpc-decl.c (shadow_record_fields)
 */
int undeclared_variable_notice = 0; /* 1 if explained undeclared var errors. */

tree
check_identifier (decl, id)
tree decl;
tree id;
{
    /* First: Special case for WITH-shadowed names.
       Environment: We have a with statement inside which we find a
       name. That name may be shadowed by a component reference to the
       fields of the record variable, i.e. ID is an identifier
       whose IDENTIFIER_LOCAL_VALUE is a COMPONENT_REF
       and DECL is also that (found by lookup_name before).

       I admit this is an ugly way. Maybe someone cleans up this.
     */

  /* @@@ Should we mark the record variable TREE_USED? */
    if (decl && TREE_CODE (decl) == COMPONENT_REF &&
	id   && IDENTIFIER_LOCAL_VALUE (id) &&
	TREE_CODE (IDENTIFIER_LOCAL_VALUE (id)) == COMPONENT_REF)
      return decl;

    if (! decl || decl == error_mark_node) {
	if (IDENTIFIER_GLOBAL_VALUE (id) != error_mark_node
	    || IDENTIFIER_ERROR_LOCUS (id) != current_function_decl)
	    error ("undeclared identifier `%s' (first use this function)",
		   IDENTIFIER_POINTER (id));
	if (! undeclared_variable_notice)
	  {
	      error ("(Each undeclared identifier is reported only once");
	      error ("for each function it appears in.)");
	      undeclared_variable_notice = 1;
	  }
	/* Prevent repeated error messages */
	IDENTIFIER_GLOBAL_VALUE (id) = error_mark_node;
	IDENTIFIER_ERROR_LOCUS (id) = current_function_decl;
	return error_mark_node;
    } else if (! TREE_USED (decl))
      {
	if (DECL_EXTERNAL (decl))
	    assemble_external (decl);
	TREE_USED (decl) = 1;
      }

    /* Set TREE_NONLOCAL if var is inherited in nested fcn.  */
    if (DECL_CONTEXT (decl) != 0
	&& DECL_CONTEXT (decl) != current_function_decl)
      {
	DECL_NONLOCAL (decl) = 1;
	mark_addressable (decl);
      }

    /* @@@@ Check if the copy_node is really necessary here! */
    if (TREE_CODE (decl) == CONST_DECL)
	decl = copy_node (DECL_INITIAL (decl)); /* User defined constant */

    return decl;
}

/*
 * Return 1 if the type of the node P is a character array node
 * or it's a string constant.
 */
int
is_string_type (p)
tree p;
{

    if (TREE_CODE (p) == STRING_CST ||
	(TREE_CODE (TREE_TYPE (p)) == ARRAY_TYPE &&
	 TREE_TYPE (TREE_TYPE (p)) == char_type_node &&
	 tree_int_cst_equal (TYPE_MIN_VALUE (TYPE_DOMAIN (TREE_TYPE (p))),
			     integer_one_node)))
	return 1;
    return 0;
}

/*
 * Initialize misc. flags used in here.
 * size_of_file_type is also used in gpc-decl.c (when making TEXT file)
 */
init_util ()
{
  language_string = "GNU Pascal";

  size_of_file_type = size_int (FDR_Size * BITS_PER_UNIT);
  
  /* set the default integer set size, if not set with -fsetlimit:NNNNN option
   */
  integer_set_size = requested_set_size > 0
    			? size_int (requested_set_size)
			: size_int (DEFAULT_SET_SIZE);
  
  /* On integer sets of unknown size (i.e. constructors with integers in them)
   * we use the following default set
   */
  integer_set_type_node =
    build_set_type
         (build_tree_list
	       (integer_type_node,
		build_index_2_type
		     (size_zero_node,
		      size_int (TREE_INT_CST_LOW (integer_set_size) - 1))));

  /* @@@@ Try to get rid of this hack */
  get_empty_set_node ();
}

/* Get an "identifier" that starts with a BLANK character. */
tree
get_identifier_with_blank(str)
char *str;
{
  tree result;
  int len = strlen (str);
  char *newstr = (char *) xmalloc (len + 2);

  *newstr = ' ';
  strcpy (newstr + 1, str);
  *(newstr + len + 1) = '\0';
  result = get_identifier (newstr);
  free (newstr);
  return result;
}

int
set_empty_p (type)
     tree type;
{
  int is_empty_set;
  tree domain;
  
  /* @@@ Should abort? */
  if (TREE_CODE (type) != SET_TYPE)
    return 0;

  domain = TYPE_DOMAIN (type);

  is_empty_set =
         TREE_CODE (TYPE_MAX_VALUE (domain)) == INTEGER_CST
      && TREE_CODE (TYPE_MIN_VALUE (domain)) == INTEGER_CST
      && TREE_INT_CST_LOW (TYPE_MAX_VALUE (domain)) == -1
      && TREE_INT_CST_LOW (TYPE_MIN_VALUE (domain)) == 0;
    
  return is_empty_set;
}
