/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   File Descriptor Record definition.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/* This version does not do internal
 * buffering of its own
 *
 * no files are actually packed
 *
 * The compiler must pass the address of the FDR struct to the
 * RTS routines that needs files.
 */

/* Include the common definitions file for the compiler and
 * run time system
 */
#include "rts_types.h"

/* RtsSta bit definitions */

#define FiNOP 0		/* File has not been opened */
#define FiRONLY 1	/* File opened but is read only */
#define FiORE 2		/* File open for reading */
#define FiWRI 4		/* File open for writing */
#define FiRND 8		/* File open for random access */

/* RtsSta Device specific bits & Bit to tell that file buffer is junk */
#define FiTTY	16	/* /dev/tty* flush output before GET */
#define FiNUL   32	/* /dev/null */
#define FiPIPE  128	/* Output into pipe/Read from pipe (not implemented) */

#define FiANY (FiORE | FiWRI | FiRND)

/* some useful macro definitions */

#define m_FILBUF(file)	(*((file)->FilBuf))
#define m_FILBPTR(file)	((file)->FilBuf)
#define m_FILNUM(file)  ((file)->FilJfn)
#define m_FISIZE(file)  ((file)->FilElem)
#define m_NXTFDR(file)  ((file)->NxtFdr)
#define m_SIZ(file)	((file)->FilSiz)
#define m_NAM(file)	((file)->FilNam)
#define m_EXTNAM(file)	((file)->ExtNam)
#define m_FILSTA(file)	((file)->FilSta)
#define m_STATUS(file)	((file)->RtsSta)

#define tst_UND(file)	(!!(m_FILSTA(file) & FiUnd))
#define tst_LAZY(file)	(!!(m_FILSTA(file) & FiLzy))
#define tst_EOF(file)	(!!(m_FILSTA(file) & FiEof))
#define tst_EOLN(file)	(!!(m_FILSTA(file) & FiEln))
#define tst_TXT(file)	(!!(m_FILSTA(file) & FiTxt))
#define tst_EXT(file)	(!!(m_FILSTA(file) & FiExt))
#define tst_PCK(file)	(!!(m_FILSTA(file) & FiPck))
#define tst_EMPTY(file) (!!(m_FILSTA(file) & FiClr))

#define set_UND(file)	(m_FILSTA(file) |= FiUnd)
#define set_LAZY(file)	(m_FILSTA(file) |= FiLzy)
#define set_EOF(file)	(m_FILSTA(file) |= FiEof)
#define set_EOLN(file)	(m_FILSTA(file) |= FiEln)
#define set_TXT(file)	(m_FILSTA(file) |= FiTxt)
#define set_EXT(file)	(m_FILSTA(file) |= FiExt)
#define set_PCK(file)	(m_FILSTA(file) |= FiPck)
#define set_EMPTY(file) (m_FILSTA(file) |= FiClr)

#define clr_UND(file)	(m_FILSTA(file) &= ~FiUnd)
#define clr_LAZY(file)	(m_FILSTA(file) &= ~FiLzy)
#define clr_EOF(file)	(m_FILSTA(file) &= ~FiEof)
#define clr_EOLN(file)	(m_FILSTA(file) &= ~FiEln)
#define clr_TXT(file)	(m_FILSTA(file) &= ~FiTxt)
#define clr_EXT(file)	(m_FILSTA(file) &= ~FiExt)
#define clr_PCK(file)	(m_FILSTA(file) &= ~FiPck)
#define clr_EMPTY(file) (m_FILSTA(file) &= ~FiClr)

#define tst_TTY(file)	(m_STATUS(file) & FiTTY)
#define tst_NUL(file)	(m_STATUS(file) & FiNUL)

#define set_TTY(file)	(m_STATUS(file) |= FiTTY)
#define set_NUL(file)	(m_STATUS(file) |= FiNUL)

/* TTY & NUL are never cleared */

#define ok_EOF(file)   (!tst_EOF(file) || _p_generic(16))
#define ok_READ(file)  ((m_STATUS(file)  & (FiORE | FiRND)) || _p_generic(14))
#define ok_WRITE(file) (((m_STATUS(file) & (FiWRI | FiRND)) || _p_generic(9)) \
			&& !tst_NUL(file))

/* Clear Test and Set status bits in Rts Status word */
#define CLR_STATUS(file,bit)	(m_STATUS(file) &= ~(bit))
#define TST_STATUS(file,bit)	(m_STATUS(file) &   (bit))
#define SET_STATUS(file,bit)	(m_STATUS(file) |=  (bit))

/* BYTENUM calculates the byte where NumE'th element starts in file.
   First possible element is 1. */
#define BYTENUM(File, NumE)     (tst_PCK(File) ? ((NumE)-1) / (8 / m_SIZ(File)) : \
				 ((NumE) * m_SIZ(File)))

/* Opposite of BYTENUM. Calculates the number of Pascal file component
   the byte is in.
 */
#define NUMBYTE(File, NumBytes) (tst_PCK(File) ? ((8 / m_SIZ(File)) * NumBytes) : \
				 (NumBytes / m_SIZ(File)))

/* Pointers to the FDR chain */
extern FDR LastFdr;
extern FDR FirstFdr;
