/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Transcendential functions. Uses libm.a.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 * changes for gpc:
 *  everything receives and returns a double_type_node
 */

#include "gpcsym.h"
#include <math.h>

#ifdef p_ARCTAN
double
_p_arctan(x)
double	x;
{
    double val;

    val = atan(x);

    return(val);
}
#endif

#ifdef p_SQRT
double
_p_sqrt(x)
double	x;
{
    if (x < 0.0) {
	_p_generic(34);
	x = -x;
    }
    return(sqrt(x));
}
#endif

#ifdef p_LN
double
_p_ln(x)
double	x;
{
    if (x <= 0.0) {
	_p_generic(33);
	if (x == 0.0)
#ifdef DBL_MIN
	    return(DBL_MIN);
#else
	    return (0.0);
#endif
	x = -x;
    }
    return(log(x));
}
#endif

#ifdef p_EXP
double
_p_exp(x)
double	x;
{
    double ret;

    if ((ret = exp(x)) == HUGE)
	_p_generic(100);		 /* -JV- Floating overflow ?? */

    return(ret);
}
#endif

#ifdef p_SIN
double
_p_sin(x)
double	x;
{
    return(sin(x));
}
#endif

#ifdef p_COS
double
_p_cos(x)
double	x;
{
    return(cos(x));
}
#endif

#ifdef p_POW
#include <errno.h>
double
_p_pow (x,y)
double	x;
double	y;
{
  extern int errno;
  double rval;

  errno = 0;
  rval = pow (x, y);
  switch (errno)
    {
    case 0 :
      break;
    case ERANGE :
      _p_generic (621);
      break;
    case EDOM :
      _p_generic (622);
      break;
    default :
      _p_warning ("Undocumented errno value in libm.a(pow)");
    }

  return (rval);
}
#endif
