/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Top level entrypoint and RTS initialization. Interrupt handlers.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

# include "gpcsym.h"
# include <sys/stat.h>
# include "gpc_fdr.h"

# define SIGTOMASK(n) (1<<((n) - 1))
# define NASSOC	32

int     Gpc_signal    = 0; /* Use Gpc_sigcause if this signal appears */
int     Gpc_sigcause  = 0; /* Number of error message to give if Gpc_signal */
int	Gpc_debug     = 0;
int	Gpc_warn      = 0; /* if nonzero output runtime warnings */
int	Gpc_DEVNULL;
int	GPC_ARGC;
char	**GPC_ARGV,**GPC_ENVP;
char*	_p_GLOBAL[GLOBAL_SIZE];	/* this should be done otherwise */

/* FDR list head pointers.
   Add an FDR to the list when reset/rewritten, remove it on close.
   List is needeed to flush buffered output to terminal
    1) when program crashes, dump everything before giving error message
    2) when something is read from terminal opened as a TEXT file
 */

FDR LastFdr  = NULL;
FDR FirstFdr = NULL;

/* Prefix for the temporary file names */
char	*Gpc_tmpname = "/tmp/Gpc";

#ifdef BSD
static void	Gpc_handler();
#endif BSD

char	*index();
extern char	*optarg;

assoc	*Gpc_assoc;

FILE  *StdFile = 0;    /* -i option strings written here, and given as */
char  namebuf[ 100 ];  /*  standard input for the program. */

static void
clear_junk()
{
    _p_fflush(FALSE);
    if (StdFile) {
	fclose(StdFile);
	if (!Gpc_debug)
	    unlink(namebuf);
    }
}

main(argc,argv,envp)
int argc;
char **argv;
char **envp;
{
	int	c;
	int	eflag = 0;
	assoc	*ap;
	char	*p;
	
#if 0
	malloc_init (0, _p_malloc_warning);
#endif
	Gpc_assoc = (assoc*)malloc(NASSOC * sizeof(assoc));
	ap = Gpc_assoc;
	ap->int_name = NULL;
	while ((c = getopt(argc,argv,"dti:a:")) != EOF)
	    switch (c) {
	    case 'i':
		if (!StdFile) {
		    _p_makename(namebuf);
		    D(1, fprintf(stderr,"Opening file %s for options\n",
				 namebuf));
		    if (!(StdFile = fopen(namebuf, "w+"))) {
			_p_error(REPORT,
				"Can't open option file for writing (-i)");
			continue;
		    }
		}
		if (optarg) {
		    D(1,fprintf(stderr,"-i option line: '%s' %d\n", optarg));
		    fputs(optarg, StdFile);
		    fputs("\n", StdFile);
		}
		break;

	    case 'd':
		Gpc_debug++;
		break;
	    case 'w':
		Gpc_warn++;
		break;
	    case 'a':
		if (p = index(optarg,':')) {
		    char ch;

		    *p++ = '\0';
		    ap->int_name = optarg;
		    ap->ext_name = p;
		    D(1,fprintf(stderr,"Associated file: %s -> %s\n",
				ap->int_name, ap->ext_name));
		    ap++;
		    if (ap >= &Gpc_assoc[NASSOC])
			_p_error(REPORT, "Too many associated file names (-a)");
		    else
			ap->int_name = NULL;
		}
		break;
	    case '?':
		eflag++;
		break;
	    }
#ifdef BSD
	setlinebuf(stderr);
#endif BSD
	if (eflag) {
	 _p_error(REPORT,"Allowed Gpc program command line options:");
	 _p_error(REPORT," -d : Debug flag (one or more) Internal OTS reports");
	 _p_error(REPORT," -i : Each option makes one line to standard input");
	 _p_error(REPORT," -a : Associate file names. -a input:infile");
        }
	GPC_ARGC = argc;
	GPC_ARGV = argv;
	GPC_ENVP = envp;
	{
	    struct stat stinfo;
	    if (stat(NULL_DEVICE_NAME, &stinfo) == 0)
		Gpc_DEVNULL = stinfo.st_rdev;
	    else
		Gpc_DEVNULL = ILLINT;
	}
#ifdef BSD
	sigvecn(SIGTOMASK(SIGHUP)
		|SIGTOMASK(SIGINT)
		/* |SIGTOMASK(SIGQUIT) */
		|SIGTOMASK(SIGILL)
		|SIGTOMASK(SIGTRAP)
		|SIGTOMASK(SIGIOT)
		|SIGTOMASK(SIGEMT)
		|SIGTOMASK(SIGFPE)
		|SIGTOMASK(SIGBUS)
		|SIGTOMASK(SIGSEGV)
		|SIGTOMASK(SIGSYS)
		|SIGTOMASK(SIGPIPE)
		|SIGTOMASK(SIGTERM)
		|SIGTOMASK(SIGXCPU)
		|SIGTOMASK(SIGXFSZ),Gpc_handler);
#endif BSD

	if (StdFile) {
	    close(0);
	    rewind(StdFile);
	    D(1, fprintf(stderr,"Read stdin from file %s\n",namebuf));
	    if (dup(fileno(StdFile)))
		_p_error(REPORT, "Can't read stdin from -i option strings");
	}
	D(1, fprintf(stderr, "%s\n", gpc_rts_version));
	_p_setup ();
	gpc_main ();
	_p_final ();
	clear_junk();
	exit(0);
}

void
_p_fflush(only_ttys)
int only_ttys;	/* TRUE if flushing only terminals */
{
    FDR	scan;
    int i;

    /* Indices in range 0..2 used. */
    extern FILE _iob[];

    /* Flush stdout & stdin */
    /* DON't flush stdin; SunOS starts to behave strangely :-) */
    fflush(&_iob[1]);
    fflush(&_iob[2]);

    /* flush buffers to synchronize output messages */
    for (scan = FirstFdr; scan; scan = m_NXTFDR(scan)) {
	if (m_STATUS(scan) & FiWRI && (!only_ttys || tst_TTY(scan))) {
	    fflush(m_FILNUM(scan));
	    D(3, fprintf(stderr, "Flushed output file %s\n",m_NAM(scan)));
	}
    }
}

void
_p_makename(str)
char *str;
{
    static int counter = 0;
    sprintf(str,"%s%05d.%05d",Gpc_tmpname,getpid(),counter++);
}

#ifdef BSD
sigvecn(sigs,handler)

int	sigs;
void	(*handler)();

{
	register int	i;
	int	omask;
	struct sigvec	sv,osv;
	int	error = 0;

	omask = sigsetmask(-1);

	sv.sv_handler = handler;
	sv.sv_mask = sigs;
	sv.sv_onstack = 0;

	for (i=0;i<32;i++) {
		if (sigs >> (i - 1) & 1) {
			if (sigvec(i,&sv,&osv)) error++;
		}
	}
	sigsetmask(omask);
	return(error);
}

struct sig_expl
{
	int	sig,how,code,gpcn;
	char	*expl;
}	expls[] = 
{
	{SIGHUP,  ABORT, NONE,		  UND, "hangup"},
	{SIGINT,  ABORT, NONE,		 FAST, "interrupt"},
	{SIGQUIT, ABORT, NONE,		  UND, "quit"},
#ifdef vax
	{SIGILL,  ABORT, ILL_RESAD_FAULT,  UND, "reserved adressing"},
	{SIGILL,  ABORT, ILL_PRIVIN_FAULT, UND, "privileged instruction"},
	{SIGILL,  ABORT, ILL_RESOP_FAULT,  507, "reserved operand"},
#else
	{SIGILL,  ABORT, NONE,		  UND, "illegal instruction"},
#endif
	{SIGTRAP, ABORT, NONE,		  UND, "trap"},
	{SIGIOT,  ABORT, NONE,		  UND, "iot"},
	{SIGEMT,  ABORT, NONE,		  UND, "emt"},
#ifdef vax
	{SIGFPE,  ABORT, FPE_INTOVF_TRAP,   47, "integer overflow trap"},
	{SIGFPE,  ABORT, FPE_INTDIV_TRAP,   45, "integer divide by zero trap"},
	{SIGFPE,  ABORT, FPE_FLTOVF_TRAP,  100, "floating overflow trap"},
	{SIGFPE,  ABORT, FPE_FLTDIV_TRAP,   44, "floating divide by zero trap"},
	{SIGFPE, REPORT, FPE_FLTUND_TRAP,  101, "floating underflow trap"},
	{SIGFPE,  ABORT, FPE_DECOVF_TRAP,  UND, "decimal overflow"},
	{SIGFPE,  ABORT, FPE_SUBRNG_TRAP,  UND, "subscript"},
	{SIGFPE,  ABORT, FPE_FLTOVF_FAULT, 100, "floating overflow fault"},
	{SIGFPE,  ABORT, FPE_FLTDIV_FAULT,  44,"floating divide by zero fault"},
	{SIGFPE, REPORT, FPE_FLTUND_FAULT, 101, "floating underflow fault"},
#else
	{SIGFPE,  ABORT, NONE, 		  UND, "floating point trap"},
#endif
	{SIGKILL, ABORT, NONE,		  UND, "kill"},
	{SIGBUS,  ABORT, NONE,		  UND, "bus error"},
	{SIGSEGV, ABORT, NONE,		  UND, "memory fault"},
	{SIGSYS,  ABORT, NONE,		  UND, "bad system call"},
	{SIGPIPE, ABORT, NONE,		  UND, "broken pipe"},
	{SIGALRM, ABORT, NONE,		  UND, "alarm"},
	{SIGTERM, ABORT, NONE,		  UND, "termination"},
	{SIGURG,  ABORT, NONE,		  UND, "urgent"},
	{-1,      ABORT, NONE,		  0,   0}
};

static void
Gpc_handler(sig,code,scp)
int	sig,code;
struct sigcontext	*scp;
{
	struct sig_expl	*p;
	int Warning;
	char *msg;

	for(p = expls; p->sig >= 0; p++)
	{
		if (p->sig == sig && p->code == code)
			break;
	}
	if (p->sig < 0) p = 0;
	if (p && p->how == IGNORE) {
		return;
	}
	Warning = p && p->how == REPORT;
	D(1, fprintf(stderr, "Gpc_handler: Warning= %d\n", Warning));
	if (Gpc_signal == sig) {
	    msg = _p_errmsg(Gpc_sigcause); /* Preset message number */
	    Gpc_signal = 0;
	} else				/* Try to give something out */
	    msg = (p ? ((msg = _p_errmsg(p->gpcn))?msg:p->expl) : 0);
	_p_prmessage(msg,
		    (p ? p->gpcn : UND),
		    Warning);
	if (Warning) {
		return;
	}
	if (p && p->gpcn >= UND)
	    fprintf(stderr,"Exiting (with core dump)\n");
	else
	    fprintf(stderr,"Exiting\n");
	clear_junk();
	_cleanup(); /* In libc -> flsbuf.c */      
	sigvecn(SIGTOMASK(SIGILL),SIG_DFL);
        if (p && p->gpcn >= UND) {
	    sigsetmask(0);
	    kill(getpid(), SIGILL);
	}
	exit(1);
}
#endif BSD
