/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Definitions for SET type.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 *
 * Author: Jukka Virtanen <jtv@hut.fi>
 *
 */

#ifndef _GPC_SET_H_
#define _GPC_SET_H_
/*
 * This is the first test version, to test the algorithms.
 *
 * The sets are WORD ALIGNED. A pointer to the set bit vector, along
 * with the low and high element numbers are passed in strcut set_t.
 *
 * Code assumes that compiler has done type checking properly.
 * So, even if the set bounds are different, we must adjust the operation
 * to handle the difference. Also, the result set must be large
 * enough to hold the result.
 */

/* @@@ BUG: Does not work on negative bounds! */

/* for _p_set_single and _p_set_range */
#include "gvarargs.h"

/* Sets are returned as int pointer, since the size does not change here */
typedef unsigned int SET_VECTOR;

/* Sets are passed like pointers to this struct */
struct set_t {
  SET_VECTOR *Set;
  int low;
  int hi;
};

typedef struct set_t * set;

/* Get the address of the bit vector */
#define BITV(s) ((s)->Set)

/* Return low element ordinal number */
#define LO(s) ((s)->low)

/* Return high element ordinal number */
#define HI(s) ((s)->hi)

/* Return the first bit number of the first word.
   This bit does not necessarily belong to the set. */
#define RLO(s) (LO(s)-(LO(s)%BITS_PER_WORD))

/* Returns the size of the set as number of elements in it. */
#define SET_SIZE(s) (HI(s)-LO(s)+1)

/* Returns the word number from start of set in which the A'th member is in */
#define THE_WORD(a,s) (((a)-RLO(s)) / BITS_PER_WORD)

/* Returns the bit number in THE_WORD the A'th member is */
#define THE_BIT(a)  ((a) % BITS_PER_WORD)

/* returns the number of words in the set */
#define NUM_WORDS(s) ((HI(s)+BITS_PER_WORD-RLO(s)) / BITS_PER_WORD)

/* Return the number of words needeed to adjust set vector A to B */
#define VECTOR_ADJUST(a,b) ((RLO(b)-RLO(a))/BITS_PER_WORD)

/* Returns the number of bytes in the set */
/* Note that the "count" is word aligned */
#define NUM_BYTES(s) (NUM_WORDS(s) * UNITS_PER_WORD)

/* Return 1 if the set is the special node empty_set_node */
#define EMPTY_SET(s) ((HI(s) == -1 && LO(s) == 0) || !rts_card (s))

/* Return 1 if the set is a sparse set.
   This is special format:
   	HI(s) == -2,
	LO(s) specifies the number of range pairs that
	      specify the elements in the set. This is
	      only used when the number of elements in the
	      set would be HUGE, e.g. [ -maxint..maxint ]

   THIS IS NOT IMPLEMENTED YET.
 */   
#define SPARSE_SET(s) (HI(s) == -2 && LO(s) > 0)

#if (defined(__GNUC__) || defined(__STDC__)) && !defined(__PROTOTYPES__)
#define __PROTOTYPES__
#endif

#ifdef __PROTOTYPES__
int
_p_set_in (set setA, long val);

SET_VECTOR *
_p_set_copy (set setA, set setR);

int
_p_set_trap (set setA, set setB);

/* VARARGS */
SET_VECTOR *
_p_set_range ();

/* VARARGS */
SET_VECTOR *
_p_set_single ();

SET_VECTOR *
_p_set_union (set setA, set setB, set setR);

SET_VECTOR *
_p_set_diff (set setA, set setB, set setR);

SET_VECTOR *
_p_set_intersection (set setA, set setB, set setR);

int
_p_set_equal (set setA, set setB);

int
_p_set_le (set setA, set setB);

SET_VECTOR *
_p_set_symdiff (set setA, set setB, set setR);

long
_p_set_card (set setA);

#else

int _p_set_in ();

SET_VECTOR *_p_set_copy   ();
SET_VECTOR *_p_set_range  ();
SET_VECTOR *_p_set_single ();
SET_VECTOR *_p_set_union  ();
SET_VECTOR *_p_set_diff   ();
SET_VECTOR *_p_set_intersection ();

int _p_set_trap  ();
int _p_set_equal ();
int _p_set_le    ();

/* Extensions */
SET_VECTOR *_p_set_symdiff ();
long _p_set_card ();

#endif /* __PROTOTYPES__ */
#endif /* _GPC_SET_H_ */
