/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Time and date routines for Extended Pascal   

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

#if ! defined(BSD) || ! defined(USG)

/* Until someone implements time functions in your system */
#define GPC_NO_TIME

#endif

#include "gpcsym.h"
#ifdef USG
# include <string.h>
# include <time.h>
#else
# include <strings.h>
# include <sys/time.h>
#endif

#define BUFLEN 100

char *malloc ();

/* The record TIMESTAMP looks like this when compiling with gcc on vax.
 *
 * It does not mean that it looks like this after
 * PACKED does something to the pascal records...
 */
struct TIMESTAMP {
    char Datevalid;
    char Timevalid;
    int  Year;
    char Month;
    char Day;
    char Hour;
    char Minute;
    char Second;
};

/*
 * Extended Pascal routine:
 *
 * GetTimeStamp (VAR t);
 * where t is of required type TimeStamp
 * PACKED RECORD
 *   DateValid,
 *   TimeValid : Boolean;
 *   year      : integer;
 *   month     : 1 .. 12;
 *   day       : 1 .. 31;
 *   hour      : 0 .. 23;
 *   minute    : 0 .. 59;
 *   second    : 0 .. 59;
 * END;
 */
void
_p_gettimestamp (t)
struct TIMESTAMP *t;
{
#ifdef GPC_IMPLEMENTS_PACKED_STRUCTURES
    abort();
#endif
#ifdef GPC_NO_TIME
    t->Year  = 1;
    t->Month = 1;
    t->Day   = 1;
    t->Datevalid = 0; /* Date is invalid */
    t->Hour  = 0;
    t->Minute= 0;
    t->Second= 0;
    t->Timevalid = 0; /* So is the time */
#else
    struct tm *gnu; /* Is it a trademark after all? */

#ifdef BSD
    struct timeval gmt;

    if (gettimeofday (&gmt, 0))
	_p_error (ABORT, "Can not read system time");

    gnu = localtime (&gmt.tv_sec); /* Convert time zones, dst etc & return struct tm */
#else
    long gmt;

    gmt =  long time ((long *) 0);

    gnu = localtime (&gmt); /* Convert time zones, dst etc & return struct tm */
#endif

    t->Year  = gnu->tm_year;
    t->Month = gnu->tm_mon;
    t->Day   = gnu->tm_mday;
    t->Datevalid = 1;
    t->Hour  = gnu->tm_hour;
    t->Minute= gnu->tm_min;
    t->Second= gnu->tm_sec;
    t->Timevalid = 1;
#endif
}

/*
 * Return current date in a string. String length should not vary.
 * Change this.
 *
 * This is not complete. Should actually check that the day of the month
 * is also valid.
 */
static char *month_names[12] = {
    "January", "February", "March", "April", "May", "June", "July",
    "August", "September", "October", "November", "December" };

char *
_p_date (t)
struct TIMESTAMP t;
{
    char Date[ BUFLEN ];
    char *result;

    if (! t.Datevalid || t.Month < 1 || t.Month > 12 ||
	t.Day < 1 || t.Day > 31)
	_p_error (ABORT, "Invalid date supplied to library function `Date'");
    (void) sprintf (Date, "%s %d %d", month_names [ t.Month - 1 ], t.Day, t.Year);

    if ((result = malloc (strlen (Date) + 1)) == 0)
	_p_error ("Out of memory in libcall `Date'");

    return strcpy (result, Date);
}

/*
 * Return current time in a string
 * Length is implementation dependent. Should not vary. Change this.
 */
char *
_p_time (t)
struct TIMESTAMP t;
{
    char Time[ BUFLEN ];
    char *result;

    if (! t.Timevalid)
	_p_error (ABORT, "Invalid time supplied to library function `Time'");
    (void) sprintf (Time, "%d:%d:%d",  t.Hour, t.Minute, t.Second);
    if ((result = malloc (strlen (Time) + 1)) == 0)
	_p_error ("Out of memory in libcall `Time'");

    return strcpy (result, Time);
}

