/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Global run time system definitions.

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 * Author: Jukka Virtanen
 *	   jtv@hut.fi
 *
 *	   Computing Centre
 *	   Helsinki University of Technology
 *	   Finland
 */

/* edit history:
   -JTV-        15 June 85:    First iteration, slow&dummy version for Pax
   -ALO & JTV - 28 September 86: This version (2.x) has been working
                  half a year without problems. Minor changes made.
   -JTV-        30 September 86: But of course I had to change things.
                  Implemented the FDR chain and fixed the default real
		  output width. Let's hope it'll not blow up. Changed
		  the version to 3.0, added copyright notices.

   Gnu Pascal Compiler Notice:

   -jtv-	26 April 1988: I had a thought of writing a Pascal
   		  front end to GCC.
   -TIME-	Passes....
   -JTV-        7  March 1992 Hello again. GCC-2 is released and GPC
   		starts to be in somekind of shape. I've not been very
		active on this one, as you may see. Sigh.
 */

#ifndef _GPCSYM_
#define _GPCSYM_
# include <stdio.h>
# include <sys/types.h>
# include <signal.h>
# include "gpc_headers.h"

#include "config.h" /* GCC xm-*.h configuration file for system type */

#ifndef USG
# define BSD
#endif

/* If you don't have this file, you can generate it with
   the `enquire -f > float.h' command. Enquire comes
   with GCC distribution, and should have been made when
   you compiled GCC */
#include "float.h"	/* ANSI C float.h file */

#include "gpc_machine.h"

#ifdef USG
# define index strchr
# define rindex strrchr
#endif USG

# ifdef DEBUG
# define D(level, x)	{if (Gpc_debug >= level) x;}
# else
# define D(level, x)
# endif

/* Other magical stuff */

# define NEWPAGE "\f"		/* PAGE(File) string */
# define NEWLINE "\n"		/* WRITELN writes this */
# define FALSE_str "False"	/* For Write Boolean */
# define TRUE_str  "True"
# define REAL_OUT_WIDTH 14      /* digits in floating point output */
# define INT_OUT_WIDTH   5	/* digits in integer output */
# define TRUE	1
# define FALSE	0

#define NULL_DEVICE_NAME "/dev/null"
#define OUTPUT_FILE_NAME "Output"
#define INPUT_FILE_NAME  "Input"

/* Signal handling types */
# define NONE   (0)
# define UND    (-1)
# define FAST   (-2)

/* error message handling codes */
# define ABORT  (-1)
# define REPORT (-2)
# define IGNORE (-3)

/* Access the _p_GLOBAL[] array as a pointer and integer */
/* the run time system stores HEAP limits there for error checking. */
# define GLOBAL_P(x) (*((char **)(&_p_GLOBAL[x])))
# define GLOBAL_I(x) (*((int *)(&_p_GLOBAL[x])))

/* heap limits, in _p_GLOBAL */
# define G_NP	(0)
# define G_NPE	(4)
# define G_NPB	(8)
# define GLOBAL_SIZE (G_NPB + 4)  /* sizeof (char *) */

/* General purpose macros */
#define isdigit(ch) ((ch) >= '0' && (ch) <= '9')
#define isspace(ch) ((ch) == ' ' || (ch) == '\t')
#define islower(ch) ((ch) >= 'a' && (ch) <= 'z')
#define isupper(ch) ((ch) >= 'A' && (ch) <= 'Z')
#define tolower(ch) ((ch) + ' ')
#define toupper(ch) ((ch) - ' ')

typedef struct assoc
{
	char	*int_name;
	char	*ext_name;
} assoc;

extern char     *gpc_rts_version;
extern int	Gpc_signal;
extern int	Gpc_sigcause;
extern int	Gpc_debug;
extern int	Gpc_warn;
extern assoc	*Gpc_assoc;
extern int      GPC_ARGC;
extern int	Gpc_DEVNULL;
extern char     **GPC_ARGV, **GPC_ENVP;
extern char 	*_p_GLOBAL[];

#endif /* _GPCSYM_ */
