/* Copyright (C) 1991 Free Software Foundation, Inc.

   This file is part of GNU Pascal Library.

   Common defines for GNU Pascal compiler and the run time system

The GNU Pascal Library is free software; you can redistribute it and/or
modify it under the terms of the GNU Library General Public License as
published by the Free Software Foundation; either version 2 of the
License, or (at your option) any later version.

The GNU Pascal Library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Library General Public License for more details.

You should have received a copy of the GNU Library General Public
License along with the GNU Pascal Library; see the file COPYING.LIB.  If
not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA.  */

/*
 * The Pascal FILE_TYPE object is defined in rts/gpc_fdr.h
 *
 * The things the compiler needs to know about the file_type are
 * as follows:
 *  1) The size of the object. (Could be avoided if file buffers
 *				were allocated from HEAP, but...)
 *  2) The file buffer address is in the first word of the object.
 *  3) The file status bits are in the second word of the file object
 *     The compiler can make inline code of EOF and EOLN by checking
 *     these bits. (_p_eof and _p_eoln also exist, but are slower)
 */

/* This is also included in the compiler to keep the definition
 * unique (currently gpc-util.c file in the compiler Pascal front end)
 */

typedef struct Fdr* FDR;

struct Fdr {
    char *FilBuf;	  /* address of the file buffer in heap 	*/
    int   FilSta;	  /* status bits; see below 			*/
    			  /* THE ABOVE FIELDS ACCESSED BY COMPILER      */
    			  /* The fields below are used only by RTS	*/
    int   FilSiz;	  /* bufsize: if packed then in bits else bytes */
    char *FilNam;	  /* Internal name of the file			*/
    char *ExtNam;	  /* External name of the file			*/
    FILE *FilJfn;	  /* FILE pointer				*/
    int   FilElem;	  /* size of file in buf elements (random acc.) */
    int   RtsSta;	  /* run time system status bits		*/
    FDR	  NxtFdr; 	  /* FDR chain pointer				*/
};

#define FDR_Size  sizeof(struct Fdr)

/* FilSta bit definitions */

#define FiUnd 	1	/* File buffer is totally undefined */
#define FiCom   2	/* File buffer assigned to in pascal side */
#define FiEof	4	/* End of file is true */
#define FiEln	8	/* End of line is true. Textfiles only */
#define	FiTxt  16	/* It's a TEXT file */
#define FiExt  32	/* External file */
#define FiPck  64	/* Packed file */
#define FiClr 128	/* Empty file */
#define FiLzy 256	/* Must do a get before buffer reference (Lazy I/O) */

/* The following defines specify the argument types of the run time system
 * routines _p_read and _p_write that get variable number and type
 * of arguments.
 * The first parameter to these routines is always a GPC file of type FDR
 * and the second is an integer that tells how many of the following
 * types follow. Note that the types require different number of arguments.
 */

#define P_INT		1	/* READ: (int *)    WRITE: int */
#define P_CHAR		2	/* READ: (char *)   WRITE: int */
#define P_REAL		3	/* READ: (double *) WRITE: double */
#define P_LINE		4	/* none */
#define P_FIX_INT	5	/* WRITE: int    & int */
#define P_FIX1_REAL	6	/* WRITE: double & int */
#define P_FIX2_REAL	7	/* WRITE: double & int & int */
#define P_STRING	8	/* WRITE: char * & int */
#define P_FIX_STRING	9	/* WRITE: char * & int & int */
#define P_FIX_CHAR	10	/* WRITE: int   & int */
#define P_BOOL		11	/* WRITE: int */
#define P_FIX_BOOL	12	/* WRITE: int & int */
